/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
	These functions are an interface between the parameters passed
	from HyperCard and the string list library. The functions are
	meant to be called once the appropriate command has been identified,
	and, in most cases, after a string list has been extracted from the
	parameter list. There is a one-to-one correspondence between these
	functions and the features of the XFCN.
*/

#include <MyHeaders>
#include "ListMain.h"

/*----------------------------------------------------------------*/
/* Getting info about XFCN */
/*----------------------------------------------------------------*/

OSErr ExecUsage(GlobalData gdata, short arg)
{
	CStr255 str;
	GetIndString(str, GetRsrcID(gdata, RSRC_INFO_ID), RSRC_USAGE_INDEX);
	result2string(GetParam(gdata), PtoCstr(str));
	return(noErr);
}

OSErr ExecVersion(GlobalData gdata, short arg)
{
	CStr255	str;
	GetVerString(str, GetRsrcID(gdata, RSRC_INFO_ID), RSRC_VERSION_INDEX,
					__DATE__, __TIME__);
	result2string(GetParam(gdata), str);
	return(noErr);
}

/*----------------------------------------------------------------*/
/* Checking for errors */
/*----------------------------------------------------------------*/

OSErr ExecError(GlobalData gdata, short arg)
{
	OSErr	err;
	
	err = GetError(gdata);
	if (err != noErr)
		result2error(GetParam(gdata), err);
	return(noErr);
}

/*----------------------------------------------------------------*/
/* Creating and disposing */
/*----------------------------------------------------------------*/

/* create the new array named in argument number 'arg'; the array's
	dimensions are set to those specified by the subscript, unless
	the setdimensions flag is false, in which case the array is left
	with empty dimensions.
*/
OSErr ExecNew(GlobalData gdata, short arg, Boolean setdimensions)
{
	String		name;			/* name of new array */
	CStr255		subscript;	/* string containing subscripts */
	ArraySub		sub;			/* the parsed subscript */
	ArraySub		dim;			/* empty dimensions */
	ArrayType	array;		/* the new array */
	OSErr			err = noErr;
	
	name = NULL;
	dim.lo = dim.hi = 0; dim.type = SUBSCRIPT0;
	
	/* get a copy of array's name */
	err = param2arrayname(gdata, arg, &name);
	
	/* parse subscripts to get array's dimensions */
	if (err == noErr) {
		err = param2subscript(gdata, arg, subscript);
		if (err == noErr) {
			if (! parse_subscript(subscript, &dim, &sub))
				err = ERR_ARRAY_SUBSCRIPT;
			if (err == noErr) {
				/* create new array */
				err = ltable_insert(GetTable(gdata), name);
				
				if (err == noErr && setdimensions) {
					/* get the new array */
					err = ltable_get(GetTable(gdata), name, &array);
					assert(err == noErr);
				
					/* set new array's dimensions */
					err = array_setdim(array, &sub);
				}
			}
		}
		/* dispose of copy of array's name */
		strfree(name);
	}
	return(err);
}

OSErr ExecDispose(GlobalData gdata, short arg)
{
	String	name;
	ArraySub	sub;
	OSErr		err = noErr;
	
	name = NULL;
	if ((err=param2arrayname(gdata, arg++, &name)) == noErr) {
		err = ltable_delete(GetTable(gdata), name);
		strfree(name);
	}
	return(err);
}

/*----------------------------------------------------------------*/
/* Set operations */
/*----------------------------------------------------------------*/

#if LIST_SETS

static OSErr setoperator(GlobalData gdata, short arg, FPTR operation);

OSErr ExecUnion(GlobalData gdata, short arg)
{
	return(setoperator(gdata, arg, array_union));
}

OSErr ExecDiff(GlobalData gdata, short arg)
{
	return(setoperator(gdata, arg, array_diff));
}

OSErr ExecInter(GlobalData gdata, short arg)
{
	return(setoperator(gdata, arg, array_inter));
}

/* execute the specified set operation on the arrays specified by
	the parameters; the results of the operation are placed into
	a new array.
*/
static OSErr setoperator(GlobalData gdata, short arg, FPTR operation)
{
	ArrayType array1, array2, array3;
	ArraySub sub1, sub2, sub3;
	OSErr err = noErr;

	/* get first and second arrays */
	err = param2array(gdata, arg++, &array1, &sub1);
	if (! err) {
		err = param2array(gdata, arg++, &array2, &sub2);
		if (! err) {
			/* create third array */
			err = ExecNew(gdata, arg, false);
			if (! err) {
				/* get third array */
				err = param2array(gdata, arg, &array3, &sub3);
				if (! err) {
					/* execute operation */
					err = operation(array1, array2, array3, &sub1, &sub2);
				}
				if (err) /* dispose of array we just created */
					ExecDispose(gdata, arg);
			}
		}
	}
}

#endif /* LIST_SETS */

/*----------------------------------------------------------------*/
/* Getting and setting dimensions */
/*----------------------------------------------------------------*/

OSErr ExecSize(GlobalData gdata, short arg, ArrayType array, ArraySub *sub)
{
	result2num(GetParam(gdata), array_size(array));
	return(noErr);
}

OSErr	ExecGetDim(GlobalData gdata, short arg, ArrayType array, ArraySub *sub)
{
	return(ERR_BASIC_NOTIMPLEMENTED);
}

OSErr	ExecSetDim(GlobalData gdata, short arg, ArrayType array, ArraySub *sub)
{
	return(array_setdim(array, sub));
}

/*----------------------------------------------------------------*/
/* inserting and deleting items */
/*----------------------------------------------------------------*/

OSErr ExecInsert(GlobalData gdata, short arg, ArrayType array, ArraySub *sub)
{
	String		items;
	char			separator;
	OSErr			err = noErr;
	
	items = NULL; separator = 0;
	if ((err = param2handle(GetParam(gdata), arg++, &items)) == noErr) {
		if (param2separator(GetParam(gdata), arg++, &separator) == noErr)
			/* ok */ ;
		if (! err)
			err = array_insert(array, sub, items, separator);
	}
	return(err);
}

OSErr ExecDelete(GlobalData gdata, short arg, ArrayType array, ArraySub *sub)
{
	return(array_delete(array, sub));
}

/*----------------------------------------------------------------*/
/* adding items */
/*----------------------------------------------------------------*/

OSErr ExecAdd(GlobalData gdata, short arg, ArrayType array, ArraySub *sub)
{
	String	items;
	char		separator;
	OSErr		err = noErr;
	
	items = NULL; separator = 0;
	if ((err = param2handle(GetParam(gdata), arg++, &items)) == noErr) {
		if (param2separator(GetParam(gdata), arg++, &separator) == noErr)
			/* ok */ ;
		if (! err)
			err = array_add(array, items, separator);
	}
	return(err);
}

/*----------------------------------------------------------------*/
/* setting and getting items */
/*----------------------------------------------------------------*/

OSErr ExecSet(GlobalData gdata, short arg, ArrayType array, ArraySub *sub)
{
	String	items;
	char		separator;
	OSErr		err = noErr;
	
	items = NULL; separator = 0;
	if ((err = param2handle(GetParam(gdata), arg++, &items)) == noErr) {
		if (param2separator(GetParam(gdata), arg++, &separator) == noErr)
			/* ok */ ;
		if (! err)
			err = array_set(array, sub, items, separator);
	}
	return(err);
}

OSErr ExecGet(GlobalData gdata, short arg, ArrayType array, ArraySub *sub)
{
	char		separator;
	String	items;
	OSErr		err = noErr;
	
	separator = 0; items = NULL;
	(void) param2separator(GetParam(gdata), arg++, &separator);
	if (! err) {
		err = array_get(array, sub, separator, &items);
		if (! err)
			result2handle(GetParam(gdata), items);
	}
	return(err);
}

/*----------------------------------------------------------------*/
/* Searching */
/*----------------------------------------------------------------*/

#if LIST_SEARCHES

OSErr ExecSearch(GlobalData gdata, short arg, ArrayType array, ArraySub *sub)
{
	String		item;
	ListIndex	index;
	OSErr			err = noErr;
	
	item = NULL; index = 0;
	if ((err = param2handle(GetParam(gdata), arg++, &item)) == noErr) {
		if ((err = array_search(array, sub, item, &index)) == noErr) {
			if (index != ListIllegalIndex)
				result2num(GetParam(gdata), index+1);
		}
	}
	return(err);
}

OSErr ExecBinSearch(GlobalData gdata, short arg, ArrayType array, ArraySub *sub)
{
	String		item;
	ListIndex	index;
	OSErr			err = noErr;
	
	item = NULL; index = 0;
	if ((err = param2handle(GetParam(gdata), arg++, &item)) == noErr) {
		if ((err = array_binsearch(array, sub, item, &index)) == noErr) {
			if (index != ListIllegalIndex)
				result2num(GetParam(gdata), index+1);
		}
	}
	return(err);
}

OSErr ExecSeqSearch(GlobalData gdata, short arg, ArrayType array, ArraySub *sub)
{
	String		item;
	ListIndex	index;
	OSErr			err = noErr;
	
	item = NULL; index = 0;
	if ((err = param2handle(GetParam(gdata), arg++, &item)) == noErr) {
		if ((err = array_seqsearch(array, sub, item, &index)) == noErr) {
			if (index != ListIllegalIndex)
				result2num(GetParam(gdata), index+1);
		}
	}
	return(err);
}

#endif /* LIST_SEARCHES */

/*----------------------------------------------------------------*/
/* Sorting */
/*----------------------------------------------------------------*/

#if LIST_SORTS

OSErr ExecSort(GlobalData gdata, short arg, ArrayType array, ArraySub *sub)
{
	return(array_sort(array, sub));
}

OSErr ExecQkSort(GlobalData gdata, short arg, ArrayType array, ArraySub *sub)
{
	return(array_qksort(array, sub));
}

OSErr ExecShSort(GlobalData gdata, short arg, ArrayType array, ArraySub *sub)
{
	return(array_shsort(array, sub));
}

#endif /* LIST_SORTS */

/*----------------------------------------------------------------*/
/* Stack operations */
/*----------------------------------------------------------------*/

OSErr ExecPush(GlobalData gdata, short arg, ArrayType array, ArraySub *sub)
{
	String	items;
	char		separator;
	OSErr		err = noErr;
	
	items = NULL; separator = 0;
	if ((err = param2handle(GetParam(gdata), arg++, &items)) == noErr) {
		if (param2separator(GetParam(gdata), arg++, &separator) == noErr)
			/* ok */ ;
		if (! err)
			err = array_push(array, items, separator);
	}
	return(err);
}

OSErr ExecPop(GlobalData gdata, short arg, ArrayType array, ArraySub *sub)
{
	String	items;
	ListIndex count;
	char		separator;
	OSErr		err = noErr;
	
	count = 1; separator = 0;
	if (param2long(GetParam(gdata), arg++, &count) == noErr) {
		if (param2separator(GetParam(gdata), arg++, &separator) == noErr)
			/* ok */ ;
	}
	if (! err) {
		err = array_pop(array, count, separator, &items);
		if (! err)
			result2handle(GetParam(gdata), items);
	}
	return(err);
}

OSErr ExecTop(GlobalData gdata, short arg, ArrayType array, ArraySub *sub)
{
	String		items;
	ListIndex	count;
	char			separator;
	OSErr			err = noErr;
	
	count = 1; separator = 0;
	if (param2long(GetParam(gdata), arg++, &count) == noErr) {
		if (param2separator(GetParam(gdata), arg++, &separator) == noErr)
			/* ok */ ;
	}
	if (! err) {
		err = array_top(array, count, separator, &items);
		if (! err)
			result2handle(GetParam(gdata), items);
	}
	return(err);
}

/*----------------------------------------------------------------*/
/* Getting and setting attributes */
/*----------------------------------------------------------------*/

OSErr ExecGetAttr(GlobalData gdata, short arg, ArrayType array, ArraySub *sub)
{
	OSErr		err;
	int	 	value;
	StrListAttr attr;
	StrListType	strlist;
	
	strlist = array_getlist(array);
	if ((err = param2attr(gdata, arg++, &attr)) == noErr) {
		if ((err = str_lgetattr(strlist, attr, &value)) == noErr) {
			switch (attr) {
			case AttrSorted:
				result2bool(GetParam(gdata), value);
				break;
			case AttrCompare:
				result2cmpr(gdata, value);
				break;
			}
		}
	}
	return(err);
}

OSErr ExecSetAttr(GlobalData gdata, short arg, ArrayType array, ArraySub *sub)
{
	OSErr			err;
	StrListAttr	attr;
	Boolean		bool;
	StrCompareType cmpr;
	StrListType	strlist;
	int			value;
	
	strlist = array_getlist(array);
	if ((err = param2attr(gdata, arg++, &attr)) == noErr) {
		switch (attr) {
		case AttrCompare:
			err = param2cmpr(gdata, arg++, &cmpr);
			value = cmpr;
			break;
		case AttrSorted:
			err = param2bool(GetParam(gdata), arg++, &bool);
			value = bool;
			break;
		}
		if (! err)
			err = str_lsetattr(strlist, attr, value);
	}
	return(err);
}

