/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

#include <MyHeaders>
#include "arraylib.h"

#define getlist(array)	((**(array)).list)
#define getndim(array)	((**(array)).ndim)

static ArraySub *getdim(ArrayType array, short i);
static void setdim(ArrayType array, short i, ArraySub *sub);
static Boolean validarray(ArrayType array);

/*--------------------------------------------------------------------*/

#if ASSERTIONS
static Boolean validarray(ArrayType array)
{
	if (! array) return(false);
	assert(validhandle((Handle) array));
	assert(getndim(array) == 1);
	return(true);
}
#endif

/*--------------------------------------------------------------------*/

/* return the i'th dimension */
static ArraySub *getdim(ArrayType array, short i)
{
	assert(validarray(array));
	assert(i == 0); /* only one-dimensional arrays are currently supported */
	return((**array).dim + i);
}

/* set the i'th dimension */
static void setdim(ArrayType array, short i, ArraySub *sub)
{
	assert(validarray(array));
	assert(i == 0);
	(**array).dim[i] = *sub;
}

/*--------------------------------------------------------------------*/

StrListType array_getlist(ArrayType array)
{
	return(getlist(array));
}

/*--------------------------------------------------------------------*/
/* creating and disposing */
/*--------------------------------------------------------------------*/

OSErr array_new(ArrayType *arrayp)
{
	ArrayType	array;		/* the new array */
	StrListType	list;			/* the new string list */
	ArraySub		emptydim;	/* used to clear dimensions */
	OSErr			err = noErr;
	
	/* initialize */
	emptydim.lo = emptydim.hi = 0;
	emptydim.type = SUBSCRIPT0;
	*arrayp = NULL;
	
	/* create array */
	array = (ArrayType) halloc(sizeof(ArrayStruct));
	if ((err=MemError()) == noErr) {
		/* create string list */
		err = str_lnew(&list);
		if (err == noErr) {
			/* initialize array */
			getlist(array) = list;
			getndim(array) = 1;
			setdim(array, 0, &emptydim);
			*arrayp = array;
			return(noErr);
		}
		hfree(array);
	}
	return(err);
}

/* updates volatile fields in array, such as function pointers, since
	the Macintosh can load function pointers into different locations
	in memory when a code resource is loaded; call this function before
	using the array.
*/
OSErr array_update(ArrayType array)
{
	str_lupdate(getlist(array));
	assert(validarray(array));
}

OSErr array_dispose(ArrayType array)
{
	assert(validarray(array));
	str_ldispose(getlist(array));
	hfree(array);
}

/*--------------------------------------------------------------------*/
/* getting and setting dimensions */
/*--------------------------------------------------------------------*/

ListIndex array_size(ArrayType array)
{
	assert(validarray(array));
	return(str_lsize(getlist(array)));
}

OSErr array_getdim(ArrayType array, ArraySub *sub)
{
	assert(validarray(array));
	*sub = *getdim(array, 0);
	sub->hi = sub->lo + array_size(array);
	return(noErr);
}

OSErr array_setdim(ArrayType array, ArraySub *sub)
{
	ListIndex size;
	
	assert(validarray(array));
	size = str_lsize(getlist(array));
	switch (sub->type) {
	case SUBSCRIPT0: /* [] */
	case SUBSCRIPT5: /* [] */
		sub->lo = 0; sub->hi = size;
		break;
	case SUBSCRIPT1: /* [n1] */
		sub->lo = sub->hi;
		if (sub->lo < 0) sub->hi = 0;
		else sub->lo = 0;
		break;
	case SUBSCRIPT2: /* [n1] */
		sub->hi = sub->lo + size;
		break;
	case SUBSCRIPT3: /* [n1n2] */
		break;
	case SUBSCRIPT4: /* [n2] */
		if (sub->hi < 0)
			return(ERR_ARRAY_SUBSCRIPT);
		sub->lo = 0;
		break;
	}
	size = labs(sub->hi - sub->lo);
	setdim(array, 0, sub);
	assert(validarray(array));
	return(str_lsetsize(getlist(array), size));
}

/*--------------------------------------------------------------------*/
/* inserting and deleting */
/*--------------------------------------------------------------------*/

OSErr array_insert(ArrayType array, ArraySub *sub, String str, char sep)
{
	assert(validarray(array));
	return( str_linsert(
				getlist(array),
				labs(sub->lo - getdim(array, 0)->lo),	/* index */
				str, sep)
			);
}

OSErr array_delete(ArrayType array, ArraySub *sub)
{
	assert(validarray(array));
	return( str_ldelete(
				getlist(array),
				labs(sub->lo - getdim(array, 0)->lo),	/* low index */
				labs(sub->hi - getdim(array, 0)->lo))	/* high index */
			);
}

/*--------------------------------------------------------------------*/
/* getting and setting items */
/*--------------------------------------------------------------------*/

OSErr array_get(ArrayType array, ArraySub *sub, char sep, String *strp)
{
	assert(validarray(array));
	return(str_lget(
				getlist(array),
				labs(sub->lo - getdim(array, 0)->lo),	/* low index */
				labs(sub->hi - getdim(array, 0)->lo),	/* high index */
				sep, strp)
			);
}

OSErr array_set(ArrayType array, ArraySub *sub, String str, char sep)
{
	assert(validarray(array));
	return( str_lset(
				getlist(array),
				labs(sub->lo - getdim(array, 0)->lo),	/* low index */
				labs(sub->hi - getdim(array, 0)->lo),	/* high index */
				str, sep)
			);
}

/*--------------------------------------------------------------------*/
/* searching */
/*--------------------------------------------------------------------*/

#if LIST_SEARCHES

OSErr array_search(ArrayType array, ArraySub *sub, String str,
							ListIndex *index)
{
	assert(validarray(array));
	return( str_lsearch(
				getlist(array),
				labs(sub->lo - getdim(array, 0)->lo),	/* low index */
				labs(sub->hi - getdim(array, 0)->lo),	/* high index */
				str, index)
			);
}

OSErr array_binsearch(ArrayType array, ArraySub *sub, String str,
								ListIndex *index)
{
	assert(validarray(array));
	return( str_lbinsearch(
				getlist(array),
				labs(sub->lo - getdim(array, 0)->lo),	/* low index */
				labs(sub->hi - getdim(array, 0)->lo),	/* high index */
				str, index)
			);
}

OSErr array_seqsearch(ArrayType array, ArraySub *sub, String str,
								ListIndex *index)
{
	assert(validarray(array));
	return( str_lseqsearch(
				getlist(array),
				labs(sub->lo - getdim(array, 0)->lo),	/* low index */
				labs(sub->hi - getdim(array, 0)->lo),	/* high index */
				str, index)
			);
}

#endif /* LIST_SEARCHES */

/*--------------------------------------------------------------------*/
/* sorting */
/*--------------------------------------------------------------------*/

#if LIST_SORTS

OSErr array_sort(ArrayType array, ArraySub *sub)
{
	assert(validarray(array));
	return(str_lsort(
			getlist(array),
			labs(sub->lo - getdim(array, 0)->lo),	/* low index */
			labs(sub->hi - getdim(array, 0)->lo)));	/* high index */
}

OSErr array_qksort(ArrayType array, ArraySub *sub)
{
	assert(validarray(array));
	return(str_lqksort(
			getlist(array),
			labs(sub->lo - getdim(array, 0)->lo),	/* low index */
			labs(sub->hi - getdim(array, 0)->lo)));	/* high index */
}

OSErr array_shsort(ArrayType array, ArraySub *sub)
{
	assert(validarray(array));
	return(str_lshsort(
			getlist(array),
			labs(sub->lo - getdim(array, 0)->lo),	/* low index */
			labs(sub->hi - getdim(array, 0)->lo)));	/* high index */
}

#endif /* LIST_SORTS */

/*--------------------------------------------------------------------*/
/* Set operations */
/*--------------------------------------------------------------------*/

#if LIST_SETS

OSErr array_union(ArrayType array1, ArrayType array2, ArrayType array3,
						ArraySub *sub1, ArraySub *sub2)
{
	return(str_lunion(getlist(array1), getlist(array2), getlist(array3),
			labs(sub1->lo - getdim(array1, 0)->lo),
			labs(sub1->hi - getdim(array1, 0)->lo),
			labs(sub2->lo - getdim(array2, 0)->lo),
			labs(sub2->hi - getdim(array2, 0)->lo)));
}

OSErr array_diff(ArrayType array1, ArrayType array2, ArrayType array3,
						ArraySub *sub1, ArraySub *sub2)
{
	return(str_ldiff(getlist(array1), getlist(array2), getlist(array3),
			labs(sub1->lo - getdim(array1, 0)->lo),
			labs(sub1->hi - getdim(array1, 0)->lo),
			labs(sub2->lo - getdim(array2, 0)->lo),
			labs(sub2->hi - getdim(array2, 0)->lo)));
}

OSErr array_inter(ArrayType array1, ArrayType array2, ArrayType array3,
						ArraySub *sub1, ArraySub *sub2)
{
	return(str_linter(getlist(array1), getlist(array2), getlist(array3),
			labs(sub1->lo - getdim(array1, 0)->lo),
			labs(sub1->hi - getdim(array1, 0)->lo),
			labs(sub2->lo - getdim(array2, 0)->lo),
			labs(sub2->hi - getdim(array2, 0)->lo)));
}

#endif /* LIST_SETS */

/*--------------------------------------------------------------------*/
/* Operations on one-dimensional arrays only */
/*--------------------------------------------------------------------*/

/*--------------------------------------------------------------------*/
/* adding items to a sorted list */
/*--------------------------------------------------------------------*/

OSErr array_add(ArrayType array, String str, char sep)
{
	assert(validarray(array));
	return(str_ladd(getlist(array), str, sep));
}

/*--------------------------------------------------------------------*/
/* stack operations */
/*--------------------------------------------------------------------*/

OSErr array_push(ArrayType array, String str, char sep)
{
	assert(validarray(array));
	return(str_lpush(getlist(array), str, sep));
}

OSErr array_pop(ArrayType array, ListIndex cnt, char sep, String *strp)
{
	assert(validarray(array));
	return(str_lpop(getlist(array), cnt, sep, strp));
}

OSErr array_top(ArrayType array, ListIndex cnt, char sep, String *strp)
{
	assert(validarray(array));
	return(str_ltop(getlist(array), cnt, sep, strp));
}

