/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
	A library implementing a list stored in an array. See header
	for more details.
*/

#include <MyHeaders>
#include "listlib.h"
#include "sortlib.h"

#define ALLOC_CHUNK	(16)	/* chunk to allocate at a time */

static void	lsetsize(ListType list, ListIndex newsize);
static ListIndex	locate(ListType list, ListIndex low, ListIndex high,
								ListItem item, FPTR compare, Boolean *found);

/*----------------------------------------------------------------*/
/* Validate operations */
/*----------------------------------------------------------------*/

Boolean lvaliditem(ListItem item)
{
	if (item == NULL) return(false);
	assert(validhandle((Handle) item));
	return(true);
}

Boolean lvalidlist(ListType list)
{
	if (list == NULL) return(false);
	assert(validhandle((Handle) list));
	assert(validhandle((Handle) (**list).array));
	assert(hsize((**list).array) / sizeof(ListItem)
				== (**list).abs_size);
	assert((**list).size <= (**list).abs_size);
	return(true);
}

Boolean lvalidindex(ListType list, ListIndex index)
{
	assert(lvalidlist(list));
	return(index != ListIllegalIndex &&
			index >= lfirst(list) && index < lsize(list));
}

Boolean lvalidcmp(ListType list)
{
	assert(lvalidlist(list));
	return(lgetcmp(list) != NULL);
}

Boolean lvalidorder(ListType list)
{
	assert(lvalidlist(list));
	return(lgetorder(list) != NULL);
}

/*----------------------------------------------------------------*/
/* Creating and disposing of a list */
/*----------------------------------------------------------------*/

ListType lnew(void)
{
	ListType		list;
	ListArray	array;
	
	list = (ListType) halloc(sizeof(struct ListType));
	if (list) {
		array = (ListArray) halloc(0L);
		if (array) {
			(**list).array = array;
			(**list).size = 0;
			(**list).abs_size = 0;
			lsetcmp(list, NULL);
			lsetorder(list, NULL);
			assert(lvalidlist(list));
			return(list);
		}
		hfree(list);
	}
	return(NULL);
}

void ldispose(ListType list)
{
	register ListIndex i;
	
	assert(lvalidlist(list));
	hfree((**list).array);
	hfree(list);
}

/*----------------------------------------------------------------*/
/* Increasing and decreasing number of items in list */
/*----------------------------------------------------------------*/

void lincrease(ListType list, ListIndex n)
{
	register ListIndex	newsize;
	register ListIndex	oldsize;
	
	assert(lvalidlist(list));
	oldsize = lsize(list);
	lsetsize(list, oldsize + n);
	if (MemError() == noErr) {
		newsize = lsize(list);
		while (oldsize < newsize)
			lset(list, oldsize++, NULL);
	}
}

void ldecrease(ListType list, ListIndex n)
{
	register ListIndex i;
	
	assert(lvalidlist(list));
	assert(n >= 0);
	assert(lsize(list)-n >= 0);
	lsetsize(list, lsize(list) - n);
}

/*----------------------------------------------------------------*/
/* Shifting items right and left */
/*----------------------------------------------------------------*/

void lrshift(ListType list, ListIndex index, ListIndex n)
{
	register ListIndex	i;
	
	assert(lvalidlist(list));
	assert(lsize(list) > lfirst(list));
	assert(lvalidindex(list, index));
	if (n <= 0) return;
	if (lsize(list)-n <= index) return;
	assert(lvalidindex(list, index + n));
	for (i = llast(list); i >= index+n; i--)
		lset(list, i, lget(list, i-n));
}

void llshift(ListType list, ListIndex index, ListIndex n)
{
	register ListIndex	i;
	register ListIndex	last;
	
	assert(lvalidlist(list));
	assert(lvalidindex(list, index));
	assert(index+n <= lsize(list));
	last = lsize(list)-n;
	for (i = index; i < last; i++)
		lset(list, i, lget(list, i+n));
}

/*----------------------------------------------------------------*/
/* Inserting and deleting items */
/*----------------------------------------------------------------*/

void linsert(ListType list, ListIndex index, ListItem item)
{
	assert(lvalidlist(list));
	lincrease(list, 1);
	if (! MemError()) {
		assert(lvalidindex(list, index));
		lrshift(list, index, 1);
		lset(list, index, item);
	}
}

void ldelete(ListType list, ListIndex index)
{
	assert(lvalidlist(list));
	assert(lvalidindex(list, index));
	llshift(list, index, 1);
	ldecrease(list, 1);
}				

/*----------------------------------------------------------------*/
/* Functions for changing order of items */
/*----------------------------------------------------------------*/

void lreverse(ListType list, ListIndex lo, ListIndex hi)
{
	assert(lvalidlist(list));
	assert(0 <= lo && lo <= hi && hi <= lsize(list));
	while (lo < --hi)
		lswap(list, lo++, hi);
}

void lswap(ListType list, ListIndex i, ListIndex j)
{
	ListItem tmp = lget(list, i);
	lset(list, i, lget(list, j));
	lset(list, j, tmp);
}

/*----------------------------------------------------------------*/
/* Sorting */
/*----------------------------------------------------------------*/

#if LIST_SORTS

void lqksort(ListType list, ListIndex low, ListIndex high)
{
	HandleKey key;
	
	assert(lvalidlist(list));
	assert(0 <= low && low <= high && high <= lsize(list));
	
	key = hlock((Handle) (**list).array, false);

	qksort((void*) (*(**list).array + low * sizeof(ListItem)),
			high - low, sizeof(ListItem), lgetorder(list));

	hunlock((Handle) (**list).array, key);
}

void linsort(ListType list, ListIndex low, ListIndex high)
{
	HandleKey key;

	assert(lvalidlist(list));
	assert(0 <= low && low <= high && high <= lsize(list));

	key = hlock((Handle) (**list).array, false);

	insort((void*) (*(**list).array + low * sizeof(ListItem)),
			high - low, sizeof(ListItem), lgetorder(list));

	hunlock((Handle) (**list).array, key);
}

void lshsort(ListType list, ListIndex low, ListIndex high)
{
	HandleKey key;

	assert(lvalidlist(list));
	assert(0 <= low && low <= high && high <= lsize(list));

	key = hlock((Handle) (**list).array, false);

	shsort((void*) (*(**list).array + low * sizeof(ListItem)),
			high - low, sizeof(ListItem), lgetorder(list));

	hunlock((Handle) (**list).array, key);
}

#endif /* LIST_SORTS */

/*----------------------------------------------------------------*/
/* Merge functions */
/*----------------------------------------------------------------*/

ListType lmerge(ListType list1, ListType list2)
{
	ListType		list3;
	ListIndex	index1, index2, index3;
	ListIndex	size1, size2;
	int cond;
	
	/* create resulting list */
	list3 = lnew();
	if (! list3) return(NULL);
	/* copy list1 parameters to resulting list */
	lsetcmp(list3, lgetcmp(list1));
	lsetorder(list3, lgetorder(list1));

	/* get sizes and allocate space for total number of items */
	size1 = lsize(list1);
	size2 = lsize(list2);
	lincrease(list3, size1+size2);
	if (MemError()) {
		ldispose(list3);
		return(NULL);
	}

	/* merge lists */
	index1 = index2 = index3 = 0;
	while (index1 < size1 && index2 < size2) {
		cond = lorder(list1, lget(list1, index1), lget(list2, index2));
		if (cond <= 0)
			lset(list3, index3++, lget(list1, index1++));
		else
			lset(list3, index3++, lget(list2, index2++));
	}

	/* merge any remaining items from each list */
	while (index1 < size1)
		lset(list3, index3++, lget(list1, index1++));
	while (index2 < size2)
		lset(list3, index3++, lget(list2, index2++));

	assert(index1 == lsize(list1));
	assert(index2 == lsize(list2));
	assert(index3 == lsize(list3));
	assert(index3 == (index1 + index2));

	return(list3);
}

/*----------------------------------------------------------------*/
/* Searching */
/*----------------------------------------------------------------*/

#if LIST_SEARCHES
ListIndex lseqsrch(register ListType list,
						register ListIndex low,
						register ListIndex high,
						register ListItem item)
{
	assert(lvalidlist(list));
	assert(lvaliditem(item));
	assert(lvalidindex(list, low));
	assert(high >= low && high <= lsize(list));

	for (; low < high; low++)
		if (lcmp(list, item, lget(list, low)) == 0)
			return(low);

	return(ListIllegalIndex);
}

ListIndex lbinsrch(ListType list,
						ListIndex low,
						ListIndex high,
						ListItem item)
{
	Boolean found;
	ListIndex index;

	index = locate(list, low, high, item, lcmp, &found);
	return(found ? index : ListIllegalIndex);
}

ListIndex llocate(ListType list,
						ListIndex low,
						ListIndex high,
						ListItem item,
						Boolean *found)
{
	ListIndex	index;
	
	/* try to find item */
	index = locate(list, low, high, item, lcmp, found);

	if (! *found && lgetcmp(list) != lgetorder(list)) {
		/* get index before which item should be inserted */
		index = locate(list, low, high, item, lorder, found);
		assert(! *found);
	}

	return(index);
}

/* binary search on list for item using 'compare' function for comparing
	items; returns index to item if found, otherwise index before which
	item should be inserted so as to maintain the sorted order of the
	list. */
static ListIndex locate(register ListType list,
								register ListIndex lo,
								register ListIndex hi,
								register ListItem item,
								register FPTR compare,
								Boolean *found)
{
	register ListIndex mid;
	register int cond;
	
	assert(lvalidlist(list));
	assert(lvaliditem(item));
	assert(lvalidindex(list, lo));
	assert(lo <= hi && hi <= lsize(list));
	
	while (lo < hi) {
		mid = (lo + hi) >> 1;
		if ((cond = (*compare)(list, item, lget(list, mid))) == 0) {
			*found = true;
			return(mid);
		}
		else if (cond < 0)
			hi = mid;
		else
			lo = mid + 1;
	}
	
	*found = false;
	return(lo);
}

#endif /* LIST_SEARCHES */

/*----------------------------------------------------------------*/
/*
	Set operations: union, difference, and intersection
	Union(a,b)			Items in a, or in b, or in both
	Difference(a,b)	Items in a that are not in b
	Intersection(a,b)	Items both in a and in b
	The following functions, which implement the above operations,
	operate on items lo1 to hi1 of the first list (list1)
	and items lo2 to hi2 of the second list (list2). Both lists
	1 and 2 should be sorted according to the same rules, and
	should share the same rules for comparing items. These functions
	may be used on unsorted lists, but their output will be
	unpredictable. The results are placed into list 3, which should
	initially be empty. List 3 is not guaranteed to be in sorted
	order when the operation is finished, for the simple reason
	that the compare function is used and not the order function.
*/
/*----------------------------------------------------------------*/

#if LIST_SETS

/* append item to list, return true if succesful */
static Boolean append(ListType list, ListItem item)
{
	lincrease(list, 1);
	if (MemError()) return(false);
	lset(list, lsize(list)-1, item);
	return(true);
}

/* Place union of list1 with list2 into list3. The
	ranges of items specified lo1, hi1 and lo2, hi2
	should be sorted, or this function won't work
	correctly. */
void lunion(ListType list1, ListType list2, ListType list3,
				ListIndex lo1, ListIndex hi1,
				ListIndex lo2, ListIndex hi2)
{
	ListIndex i1, i2;
	int cmp;
	
	assert(lvalidlist(list1));
	assert(lvalidlist(list2));
	assert(lvalidlist(list3));
	assert(0 <= lo1 && lo1 <= hi1 && hi1 <= lsize(list1));
	assert(0 <= lo2 && lo2 <= hi2 && hi2 <= lsize(list2));

	/* append all differing items and one copy of every matching item */
	i1 = lo1; i2 = lo2;
	while (i1 < hi1 && i2 < hi2) {
		do {
			cmp = lcmp(list1, lget(list1, i1), lget(list2, i2));
			if (! cmp) break;
			if (cmp < 0 && ! append(list3, lget(list1, i1++)))
				return;
			else if (! append(list3, lget(list2, i2++)))
				return;
		} while (i1 < hi1 && i2 < hi2);
		if (i1 < hi1 && i2 < hi2) {
			if (! append(list3, lget(list1, i1))) return;
			i1++; i2++;
		}
	}
	/* append remaining items */
	while (i1 < hi1 && append(list3, lget(list1, i1++)))
		;
	while (i2 < hi2 && append(list3, lget(list2, i2++)))
		;
}

/* Place difference of list1 with list2 into list3. The
	ranges of items specified lo1, hi1 and lo2, hi2
	should be sorted, or this function won't work
	correctly. */
void ldiff(ListType list1, ListType list2, ListType list3,
				ListIndex lo1, ListIndex hi1,
				ListIndex lo2, ListIndex hi2)
{
	ListIndex i1, i2;
	int cmp;
	
	assert(lvalidlist(list1));
	assert(lvalidlist(list2));
	assert(lvalidlist(list3));
	assert(0 <= lo1 && lo1 <= hi1 && hi1 <= lsize(list1));
	assert(0 <= lo2 && lo2 <= hi2 && hi2 <= lsize(list2));

	/* append all items not in list 2 */
	i1 = lo1; i2 = lo2;
	while (i1 < hi1 && i2 < hi2) {
		cmp = lcmp(list1, lget(list1, i1), lget(list2, i2));
		if (cmp < 0) {
			if (! append(list3, lget(list1, i1++)))
				return;
		}
		else if (cmp == 0) {
			i1++; i2++;
		}
		else
			i2++;
	}
	/* append remaining items in list 1 */
	while (i1 < hi1 && append(list3, lget(list1, i1++)))
		;
}

/* Place intersection of list1 with list2 into list3.
	The ranges of items specified lo1, hi1 and lo2, hi2
	should be sorted, or this function won't work
	correctly. */
void linter(ListType list1, ListType list2, ListType list3,
				ListIndex lo1, ListIndex hi1,
				ListIndex lo2, ListIndex hi2)
{
	ListIndex i1, i2;
	int cmp;
	
	assert(lvalidlist(list1));
	assert(lvalidlist(list2));
	assert(lvalidlist(list3));
	assert(0 <= lo1 && lo1 <= hi1 && hi1 <= lsize(list1));
	assert(0 <= lo2 && lo2 <= hi2 && hi2 <= lsize(list2));

	/* append all items both in list 1 and in list 2 */
	i1 = lo1; i2 = lo2;
	while (i1 < hi1 && i2 < hi2) {
		cmp = lcmp(list1, lget(list1, i1), lget(list2, i2));
		if (cmp == 0) {
			if (! append(list3, lget(list1, i1)))
				return;
			i1++; i2++;
		}
		else if (cmp < 0)
			i1++;
		else
			i2++;
	}
}

#endif /* LIST_SETS */

/*----------------------------------------------------------------*/

/* Expand list's size to newsize. The handle to the
	list's memory is reallocated if newsize is greater
	than the size of the handle. Otherwise, we have
	some spare space, so we don't need to do any
	memory allocation. */
static void lsetsize(ListType list, ListIndex newsize)
{
	ListIndex abs_size;
	
	assert(lvalidlist(list));
	assert(newsize >= 0);
	
	abs_size = (newsize / ALLOC_CHUNK + 1) * ALLOC_CHUNK;
	if (abs_size != (**list).abs_size) {
		hrealloc((**list).array, abs_size * sizeof(ListItem));
		if (MemError() == noErr) {
			(**list).abs_size = abs_size;
			(**list).size = newsize;
		}
	}
	else
		(**list).size = newsize;
}

/*----------------------------------------------------------------*/
/* Functions for comparing items */
/*----------------------------------------------------------------*/

int lcmp(ListType list, ListItem item1, ListItem item2)
{
	assert(lvalidlist(list));
	assert(lvaliditem(item1));
	assert(lvaliditem(item2));
	assert(lvalidcmp(list));
	return((*lgetcmp(list))((void*)&item1, (void*)&item2));
}

int lorder(ListType list, ListItem item1, ListItem item2)
{
	assert(lvalidlist(list));
	assert(lvaliditem(item1));
	assert(lvaliditem(item2));
	assert(lvalidorder(list));
	return((*lgetorder(list))((void*)&item1, (void*)&item2));
}

#if ASSERTIONS /* following functions are replaced with macros */

/*-------------------------------------------------------------------*/
/* Getting and setting number of items and beginning and end of list */
/*-------------------------------------------------------------------*/

ListIndex lsize(ListType list)
{
	assert(lvalidlist(list));
	return((**list).size);
}

ListIndex lfirst(ListType list)
{
	assert(lvalidlist(list));
	return(0);
}

ListIndex llast(ListType list)
{
	assert(lvalidlist(list));
	assert(lsize(list) > lfirst(list));
	return(lsize(list)-1);
}

/*----------------------------------------------------------------*/
/* Getting and setting items */
/*----------------------------------------------------------------*/

ListItem lget(ListType list, ListIndex index)
{
	assert(lvalidlist(list));
	assert(lvalidindex(list, index));
	return((*(**(list)).array)[index]);
}	

void lset(ListType list, ListIndex index, ListItem item)
{
	assert(lvalidlist(list));
	assert(lvalidindex(list, index));
	(*(**(list)).array)[index] = item;
}

/*----------------------------------------------------------------*/
/* Functions for getting and setting comparison and order fptrs */
/*----------------------------------------------------------------*/

void lsetcmp(ListType list, FPTR compare)
{
	assert(lvalidlist(list));
	(**list).cmp = compare;
}

FPTR lgetcmp(ListType list)
{
	assert(lvalidlist(list));
	return((**list).cmp);
}

void lsetorder(ListType list, FPTR order)
{
	assert(lvalidlist(list));
	(**list).order = order;
}

FPTR lgetorder(ListType list)
{
	assert(lvalidlist(list));
	return((**list).order);
}

#endif /* ASSERTIONS */

