/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
DESCRIPTION

A data structure and operations implementing a simple array list.
Operations supported include insertion, deletion, several methods
of sorting and searching as well as set operations. Many commands
may be restricted to operate on a range of items, instead of
the entire list.

The list is appropriate for the Macintosh environment
since it uses handles for all of its operations. Unfortunately, use
of handles instead of void pointers makes the code slightly less
portable; appropriate changes of the main type definitions and
changes to handle dereferrencing should make this unit more general.

Several preprocessor symbols may be defined to control what portions
of the list library are compiled. The default is to compile the
entire library.

__________________
MEMORY ALLOCATION

The program using these routines is reponsible for passing valid
data for the items and for disposing of such data.  Items are by
default filled with NULL pointers: use lset to put something meaningful
into an item*. Ldispose simply discards the handle to the items in the
list and the handle to the list itself; it doesn't dispose of any items.
Therefore, be sure to dispose of such items individually before calling
ldispose. The same rule applies when using ldelete: you should
dispose of the contents of an item before deleting it.
None of the functions will duplicate an item, they
simply pass its value around (for instance, lmerge merges handles, not
actual data, while lget returns a handle to the data, not a handle to
a copy).

A list may have more elements allocated for it than are actually used,
but the size of the list always reflects the number of elements used.
This speeds insertions and deletions from the list since calls to
memory manager routines are made less frequent.

*Warning:
		Only lincrease sets the new items at the end of the list to NULL;
		operations like shifting (lrshift or llshift) do not clear
		the data. This warning is here because I had a very elusive bug
		which I didn't catch till near the end of development! I'm not
		adding loops to clear the data in the shift functions since most
		operations do not require that the data be cleared (clearing the
		data in these cases would only serve to slow down the program).

_______________________
SORTING and SEARCHING

__________________________________
Order and Compare Rules may Differ

To use the sorting and searching operations you must supply both a
comparison routine and a separate routine to be used for sorting,
called the "order" routine. Having two different routines is useful
for lists that must ignore case of string items but must still place
upper case letters before lower case letters. For simpler lists the
comparison and ordering operations may use the same function. Both the
comparison and ordering functions should return values as shown below:
	First parameter is smaller than second:	result < 0
	First parameter is equal to second:			result == 0
	First parameter is greater than second:	result > 0
The compare and order functions should be declared as follows:
	int function(ListItem *, ListItem *);
Note: Actually, a void * is passed to the compare functions,
but all pointers have the same size on the Macintosh, so
it doesn't matter that the function isn't declared as taking
a void *.

In general, the ordering rules should be either as permissive as the
comparison rules or more restrictive, but never more lenient.
As an example of differing order and comparison rules, suppose a list
contains text that is to be sorted alphabetically. The ordering rules
might specify that upper case letters come before lower case versions
of the same letter, while the comparison rules could specify that
upper and lower case letters are to be considered equivalent. This is
an example where the ordering rules are more restrictive than the
comparison rules, that is, fewer instances of similar items are
considered equal. It would be an error to reverse the order and
comparison rules in this example.

_________________
Sorting

Several sorting algorithms are implemented: quick sort, shell sort,
insertion sort, and merge (not a complete sort).

_________________
ERRORS

Assertions check that preconditions hold, while multiple redundant
assertions check for consistency within a list. Checks are made for memory
manager errors. Though no error codes are returned, for those operations
that access the memory manager, the caller should check the MemError()
function to ensure that the call executed succesfully. Specifically,
lnew, lincrease, ldecrease, linsert, and ldelete may allocate
or deallocate memory (ldispose deallocates memory, but should
never encounter an error) [note: don't take this list as exhaustive].

Unfortunately, assertions slow operation to a crawl. Compile with
ASSERTIONS defined as 0 (or undefined), which will also replace
some function calls with macro's.

______________________________________________________________________________

_________________
TYPES

ListItem			An item in a list
ListIndex		An index to an item in a list
ListType			A list of items

_________________
OPERATIONS

Validation operations
lvaliditem		True if item is valid (not NULL)
lvalidlist		True if list is valid
lvalidindex		True if index is valid (between ListFirst and ListLast)
lvalidcmp		True if comparison routine is valid (not NULL)
lvalidorder		True if ordering routine is valid (not NULL)

______________________
Creating and disposing
lnew			Create a new empty list
ldispose		Dispose of the list
_____________________________________
Getting size and first and last items
lsize			Return the size of a list
lfirst		Return index to first item
llast			Return index to last item
_________________________
Getting and setting items
lget			Return the indexed item
lset			Set the value of the indexed item
______________________________
Increasing and decreasing size

Notes:
	1. lincrease fills the new items with NULL pointers.

lincrease	Increase the size of a list by n
ldecrease	Decrease the size of a list by n
_____________________________
Shifting items left and right
lrshift	Shift n items right starting from indexed item.
			Size of list is unchanged.
llshift	Shift n items left starting from indexed item
			Size of list is unchanged.
____________________________
Inserting and deleting items

Notes:	1. These functions are provided as a convinience when inserting
				or deleting only one item. For larger chunks it is more
				efficient to use the primitive operations
				lincrease, ldecrease, lrshift, and llshift.
				
linsert	Insert one item
ldelete	Delete one item
______________________________________________________
Setting and getting comparison and ordering operations
lsetcmp		Set the function used for comparing items when searching
lgetcmp		Return pointer to the function used for comparing items
				when searching
lsetorder	Set the function used for comparing items when sorting
lgetorder	Return pointer to the function used for comparing items
				when sorting
____________________
Searching operations

Notes:
	1. You must set the compare function before calling any of these
		routines.
	2. The binary search routine expects the list to be sorted
		according to rules compatible with the comparison routine; failure to
		maintain this condition could result in a fatal and undetectable
		error.
	3. An illegal index is returned if an item isn't found.

lcmp			Compare two items when searching list
lseqsrch	 	Do a sequential search in a list (useful for unsorted lists)
lbinsrch		Return index to item in a list or invalid index if not found
				using binary search. List must be sorted.
llocate		Return index before which item should be inserted for a
				sorted list.
___________________
Sorting operations

Note:
	1. You must set the order function before calling any of these routines.
	2. lmerge uses the order routine from the list1 parameter; the new list
		resulting from the merge uses the compare and order routines from list1.
	
lorder		Order two items when sorting
lswap			Swap two items
lqksort		Sort a list using quick sort
lshsort		Sort a list using shell sort
linsort		Sort a list using insertion sort
lmerge		Merges two lists to produce a third. The lists are expected
				to be sorted, so that the new list will also be sorted,
				though the function will work on unsorted lists as well.
				Items from list1 will tend to come before items in
				list2, subject to their order as determined by lorder.
________________
Reordering list
lreverse		Reverse count items after the indexed item
________________
Set operations
lunion		Produces union of two lists
ldiff			Produces difference of two lists
linter		Produces intersection of two lists
_______________________________________________________________________________
*/

#ifndef _H_listlib
#define _H_listlib

#ifndef _H_mystdlib
#include "mystdlib.h"
#endif

/* conditional compilation flags */
#define LIST_SETS			(1)	/* compile set functions */
#define LIST_SORTS		(1)	/* compile sort functions */
#define LIST_SEARCHES	(1)	/* compile search functions */

#define ListIllegalIndex	(-1L)	/* guarenteed never to address an item */

/* warning: don't make ListIndex an unsigned value, since not being
	able to go to -1 causes some annoying bugs which are hard to see
	when writing the code!
*/
typedef long		ListIndex;		/* index to an item in a list */
typedef Handle		ListItem;		/* item in a list */
typedef ListItem	**ListArray;	/* handle containing array of items */
typedef struct ListType {
	ListArray	array;		/* array of items */
	ListIndex	abs_size;	/* absolute size (number of items) of list */
	ListIndex	size;			/* number of items actually in use */
	FPTR			cmp;			/* comparison function for searching */
	FPTR			order;		/* comparison function for sorting */
} **ListType;

/* validate operations */
Boolean	lvaliditem(ListItem item);
Boolean	lvalidlist(ListType list);
Boolean	lvalidindex(ListType list, ListIndex index);
Boolean	lvalidcmp(ListType list);
Boolean	lvalidorder(ListType list);

/* getting info about list */
ListIndex	lsize(ListType list);
ListIndex	lfirst(ListType list);
ListIndex	llast(ListType list);
ListItem		lget(ListType list, ListIndex index);
void			lset(ListType list, ListIndex index, ListItem item);

/* creating and disposing */
ListType	lnew(void);
void		ldispose(ListType list);

/* changing number of items */
void	lincrease( ListType list, ListIndex n);
void	ldecrease(ListType list, ListIndex n);

/* shifting items left and right */
void	lrshift(ListType list, ListIndex index, ListIndex n);
void	llshift(ListType list, ListIndex index, ListIndex n);

/* inserting and deleting items */
void	linsert(ListType list, ListIndex index, ListItem item);
void	ldelete(ListType list, ListIndex index);

void	lreverse(ListType list, ListIndex lo, ListIndex hi);

/* setting and getting comparison and order function pointers */
void	lsetcmp(ListType list, FPTR compare);
FPTR	lgetcmp(ListType list);
void	lsetorder(ListType list, FPTR order);
FPTR	lgetorder(ListType list);

/* comparing items */
int	lcmp(ListType list, ListItem item1, ListItem item2);
int	lorder(ListType list, ListItem item1, ListItem item2);

/* sorting */
void	lswap(ListType list, ListIndex i, ListIndex j);
void	lqksort(ListType list, ListIndex low, ListIndex high);
void	linsort(ListType list, ListIndex low, ListIndex high);
void	lshsort(ListType list, ListIndex low, ListIndex high);
ListType	lmerge(ListType list1, ListType list2);

/* searching */
ListIndex	lseqsrch(ListType list, ListIndex low, ListIndex high, ListItem item);
ListIndex	lbinsrch(ListType list, ListIndex low, ListIndex high, ListItem item);
ListIndex	llocate(ListType list, ListIndex low, ListIndex high, ListItem item, Boolean *found);

/* set operations */
void lunion(ListType list1, ListType list2, ListType list3,
				ListIndex lo1, ListIndex hi1,
				ListIndex lo2, ListIndex hi2);
void ldiff(ListType list1, ListType list2, ListType list3,
				ListIndex lo1, ListIndex hi1,
				ListIndex lo2, ListIndex hi2);
void linter(ListType list1, ListType list2, ListType list3,
				ListIndex lo1, ListIndex hi1,
				ListIndex lo2, ListIndex hi2);

#if ! (ASSERTIONS)

#define lgetcmp(l)			((**(l)).cmp)
#define lgetorder(l)			((**(l)).order)
#define lsetcmp(l, fp)		((void) (lgetcmp(l) = (fp)))
#define lsetorder(l, fp)	((void) (lgetorder(l) = (fp)))

#define lsize(l)				((**(l)).size)
#define lfirst(l)				(0)
#define llast(l)				(lsize(l)-1)

#define lget(l, i)			((*(**(l)).array)[(i)])
#define lset(l, i, v)		((void) (lget((l), (i)) = (v)))

#endif /* ASSERTIONS */

#endif