/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
	Please see header for more details.
*/

#include <MyHeaders>
#include "listtable.h"
#include "itemutil.h"
#include "strlib.h"

#define getnames(table)		((**(table)).names)
#define gethandles(table)	((**(table)).handles)

static ListIndex ltable_find(ListTableType table, String name,
										ArrayType *arrayp);

#if ASSERTIONS
static Boolean	validtable(ListTableType table);

static Boolean	validtable(ListTableType table)
{
	assert(validhandle((Handle) table));
	assert(lgetcmp(getnames(table)) == icompareitems);
	assert(lgetorder(getnames(table)) == iorderitems);
	return(true);
}
#endif

/*--------------------------------------------------------------------*/
/* creating and disposing */
/*--------------------------------------------------------------------*/

/* allocate a new empty table */
OSErr	ltable_new(ListTableType *tablep)
{
	ListTableType	table;
	ListType	names, handles;
	OSErr 	err = noErr;
	
	*tablep = NULL;
	table = (ListTableType) halloc(sizeof(ListTableStruct));
	if ((err=MemError()) == noErr) {
		names = lnew();
		if ((err=MemError()) == noErr) {
			handles = lnew();
			if ((err=MemError()) == noErr) {
				(**table).names = names;
				(**table).handles = handles;
				ltable_update(table);
				*tablep = table;
				return(noErr);
			}
			ldispose(names);
		}
		hfree(table);
	}
	return(err);
}

/* dispose of the table */
OSErr ltable_dispose(ListTableType table)
{
	assert(validtable(table));
	ldispose(getnames(table));
	ldispose(gethandles(table));
	hfree(table);
}

/* update volatile fields in table, such as function pointers, since
	the Macintosh can load function pointers into different locations
	in memory when a code resource is loaded; call this function before
	using the table.
*/
OSErr ltable_update(ListTableType table)
{
	lsetcmp(getnames(table), icompareitems);
	lsetorder(getnames(table), iorderitems);
	assert(validtable(table));
}

/*--------------------------------------------------------------------*/
/* inserting and deleting */
/*--------------------------------------------------------------------*/

OSErr ltable_insert(ListTableType table, String name)
{
	ArrayType	array;
	ListIndex	index;
	Boolean		found;
	OSErr			err=noErr;

	assert(validtable(table));

	/* find location to insert */
	index = 0; found = false;
	if (lsize(getnames(table)) > 0)
		index = llocate(getnames(table), lfirst(getnames(table)),
							lsize(getnames(table)), name, &found);
	if (found)
		return(ERR_LTABLE_EXISTS);
		
	/* duplicate name */
	name = strdup(name);
	if ((err=MemError()) == noErr) {
		/* create new string list */
		err = array_new(&array);
		if (err == noErr) {
			/* insert name into names index */
			linsert(getnames(table), index, name);
			if ((err=MemError()) == noErr) {
				/* insert list into handles index */
				linsert(gethandles(table), index, (ListItem) array);
				/* return if ok, otherwise fall through to cleanup code */
				if ((err=MemError()) == noErr)
					return(noErr);
				/* falling through to cleanup code... */
				ldelete(getnames(table), index);
			}
			array_dispose(array);
		}
		strfree(name);
	}
	assert(validtable(table));
	return(err);
}

/* delete and dispose of the named array */
OSErr ltable_delete(ListTableType table, String name)
{
	ListIndex	index;	/* index of array in table */
	ArrayType	array;	/* array to dispose of */
	OSErr			err = noErr;

	assert(validtable(table));
	index = ltable_find(table, name, &array);
	if (lvalidindex(getnames(table), index)) {
		
		/* dispose of name */
		strfree(lget(getnames(table), index));
		ldelete(getnames(table), index);

		/* dispose of array */
		array_dispose(array);
		ldelete(gethandles(table), index);
	}
	else
		err = ERR_LTABLE_NOTFOUND;
	assert(validtable(table));
	return(err);
}

/* get the named array */
OSErr ltable_get(ListTableType table, String name, ArrayType *arrayp)
{
	ListIndex	index;
	OSErr			err = noErr;
	
	assert(validtable(table));
	index = ltable_find(table, name, arrayp);
	if (! lvalidindex(getnames(table), index))
		return(ERR_LTABLE_NOTFOUND);
	return(noErr);
}

static ListIndex ltable_find(ListTableType table, String name,
										ArrayType *arrayp)
{
	ListIndex index = ListIllegalIndex;

	assert(validtable(table));
	index = lbinsrch(getnames(table), lfirst(getnames(table)),
							lsize(getnames(table)), name);
	if (lvalidindex(getnames(table), index)) {
		*arrayp = (ArrayType) lget(gethandles(table), index);
		array_update(*arrayp);
	}
	return(index);
}

