/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
	See header ("strlib.h") for more details.
*/

#include <MyHeaders>
#include "strlistlib.h"
#include "itemutil.h"
#include "strlib.h"

#define getlist(l)	((**(l)).list)
#define getsorted(l)	((**(l)).sorted)
#define getcmpr(l)	((**(l)).cmpr)

/*----------------------------------------------------------------*/
/* Validate operations */
/*----------------------------------------------------------------*/

static Boolean validstrlist(StrListType strlist);
static Boolean validlohi(StrListType strlist, ListIndex lo, ListIndex hi);

static Boolean validstrlist(StrListType strlist)
{
	
	if (strlist == NULL) return(false);
	assert(validhandle((Handle)strlist));
	assert(lvalidlist(getlist(strlist)));
	switch (getcmpr(strlist)) {
	case STR_COMPARE_EXACT:
		assert(lgetcmp(getlist(strlist)) == compareitems);
		assert(lgetorder(getlist(strlist)) == compareitems);
		break;
	case STR_COMPARE_IGNORECASE:
		assert(lgetcmp(getlist(strlist)) == icompareitems);
		assert(lgetorder(getlist(strlist)) == iorderitems);
		break;
	case STR_COMPARE_INTERNATIONAL:
		assert(lgetcmp(getlist(strlist)) == IUcompareitems);
		assert(lgetorder(getlist(strlist)) == IUorderitems);
		break;
	case STR_COMPARE_FLOAT:
		assert(lgetcmp(getlist(strlist)) == numcmpitems);
		assert(lgetorder(getlist(strlist)) == numorderitems);
		break;
	default:
		assert(false);
	}
	return(true);
}

static Boolean validlohi(StrListType strlist, ListIndex lo, ListIndex hi)
{
	if (lo < 0 || lo >= str_lsize(strlist))
		return(false);
	if (hi < lo || hi > str_lsize(strlist))
		return(false);
	return(true);
}

/*----------------------------------------------------------------*/
/* Creating and disposing */
/*----------------------------------------------------------------*/

OSErr str_lnew(StrListType *strlistPtr)
{
	OSErr	err;
	StrListType strlist;
	ListType	list;
	
	strlist = *strlistPtr = (StrListType) halloc(sizeof(StrListStruct));
	if ((err=MemError()) == noErr) {
		list = lnew();
		if ((err=MemError()) == noErr) {
			getlist(strlist) = list;
			getsorted(strlist) = false;
			getcmpr(strlist) = STR_COMPARE_IGNORECASE;
			str_lupdate(strlist);
			return(noErr);
		}
		hfree(strlist);
	}
	return(err);
}

OSErr str_ldispose(StrListType strlist)
{
	ListType		list;
	ListIndex	index;
	
	assert(validstrlist(strlist));
	list = getlist(strlist);
	
	/* free all items in the list */
	if (lsize(list) > 0) {
		for (index = lfirst(list); index < lsize(list); ++index)
			strfree(lget(list, index));
	}
	
	/* free the list */
	ldispose(list);
	
	/* free the string list */
	hfree(strlist);
	
	return(noErr);
}

/* set the list's function pointers to their correct values; should be
	called before list is used.
*/
void str_lupdate(StrListType strlist)
{
	switch (getcmpr(strlist)) {
	case STR_COMPARE_EXACT:
		lsetcmp(getlist(strlist), compareitems);
		lsetorder(getlist(strlist), compareitems);
		break;
	case STR_COMPARE_IGNORECASE:
		lsetcmp(getlist(strlist), icompareitems);
		lsetorder(getlist(strlist), iorderitems);
		break;
	case STR_COMPARE_INTERNATIONAL:
		lsetcmp(getlist(strlist), IUcompareitems);
		lsetorder(getlist(strlist), IUorderitems);
		break;
	case STR_COMPARE_FLOAT:
		lsetcmp(getlist(strlist), numcmpitems);
		lsetorder(getlist(strlist), numorderitems);
		break;
	default:
		assert(false);
	}
	assert(validstrlist(strlist));
}

/*----------------------------------------------------------------*/
/* Getting and setting size of list */
/*----------------------------------------------------------------*/

ListIndex str_lsize(StrListType strlist)
{
	assert(validstrlist(strlist));
	return(lsize(getlist(strlist)));
}

OSErr str_lsetsize(StrListType strlist, ListIndex newsize)
{
	ListIndex	i;			/* for making or getting rid of excess items */
	String		empty;	/* for creating empty items */
	ListIndex	oldsize;	/* the old size of the list */
	OSErr			err = noErr;
	
	assert(validstrlist(strlist));
	assert(newsize >= 0);
	oldsize = lsize(getlist(strlist));
	if (newsize < oldsize) {
		/* dispose of excess items */
		for (i = newsize; i < oldsize; i++)
			strfree(lget(getlist(strlist), i));
		err = str_ldelete(strlist, newsize, oldsize - newsize);
	}
	else if (newsize > oldsize) {
		/* increase size of list */
		lincrease(getlist(strlist), newsize - oldsize);
		if ((err=MemError()) == noErr) {
			/* create new empty items */
			for (i = oldsize; i < newsize; i++) {
				empty = stralloc("");
				if (empty)
					lset(getlist(strlist), i, empty);
				else {
					/* clean up */
					err = MemError();
					ldecrease(getlist(strlist), newsize - i);
					break;
				}
			}
		}
	}
	return(err);
}

/*----------------------------------------------------------------*/
/* Getting and setting items */
/*----------------------------------------------------------------*/

/* Get items lo to hi, separate each item with separator character,
	or with a comma if no separator is given. Result handle is NULL
	if lo==hi.
*/
OSErr str_lget(StrListType strlist, ListIndex lo, ListIndex hi,
					char separator, String *result)
{
	char			*p;		/* pointer to result */
	ListType		list;		/* the list */
	size_t		length;	/* total length of output string */
	ListIndex	i;			/* index to current item */
	
	*result = NULL;

	/* check for illegal conditions */
	assert(validstrlist(strlist));
	list = getlist(strlist);
	if (! validlohi(strlist, lo, hi)) return(ERR_STRLIST_INDEX);
	if (hi == lo) return(noErr);

	/* use default separator if necessary */
	if (! separator) separator = ',';

	/* calculate total size of string including separators */
	length = MAX(1, hi-lo);	/* start with separators and terminating NULL */
	for (i = lo; i < hi; i++)
		length += strlen(*lget(list, i));
	
	/* allocate handle big enough to hold result */
	*result = halloc(length);
	if (! *result) return(MemError());
	
	/* append items to output string */
	HLock(*result);
	p = **result; *p = 0;
	for (i = lo; i < hi; i++) {
		/* append current item to end of result */
		p += strlen(strcpy(p, *lget(list, i)));
		
		/* append separator (if needed) to end of result */
		if (hi-lo > 1 && i < hi-1) {
			*p++ = separator; *p = 0;
		}
	}
	HUnlock(*result);
	return(noErr);
}

/* set items lo to hi to the substrings in the source string */
OSErr str_lset(StrListType strlist, ListIndex lo, ListIndex hi,
					String str, char separator)
{
	HandleKey	key;		/* for locking/unlocking string handle */
	Boolean		sorted;	/* for list's sorted attribute */
	ListIndex	dummy;
	OSErr			err;
	
	/* check for illegal conditions */
	assert(validstrlist(strlist));
	if (! validlohi(strlist, lo, hi))
		return(ERR_STRLIST_INDEX);
	if (hi == lo)
		return(noErr);

	/* lock string and set items from it */

	key = hlock(str, true);
	
	sorted = getsorted(strlist);
	err = substrset(getlist(strlist), lo, *str, separator, &dummy, &sorted);
	getsorted(strlist) = sorted;

	hunlock(str, key);

	return(err);
}

/*----------------------------------------------------------------*/
/* Inserting and deleting items */
/*----------------------------------------------------------------*/

/* insert substrings of input string into strlist before indexed item. */
OSErr str_linsert(StrListType strlist, ListIndex index, String str, char separator)
{
	ListType		list;			/* the list */
	ListIndex	count;		/* number of strins to insert */
	ListIndex	successful;	/* number of string succesfully inserted */
	Boolean		sorted;		/* true if still in sorted order after insertion */
	HandleKey	key;			/* for locking/unlocking string handle */
	OSErr			err;
	
	/* check for illegal conditions */
	assert(validstrlist(strlist));
	assert(validhandle(str));
	list = getlist(strlist);
	if (index < 0 || index > lsize(list))
		return(ERR_STRLIST_INDEX);
	
	/* increase size of list to accommodate new items */
	count = substrcnt(*str, separator);
	lincrease(list, count);
	if (MemError()) return(MemError());
	
	assert(lsize(list) >= index+count);
	
	/* shift items over to accommodate new items */
	if (lsize(list) > 1) {
		register ListIndex	i;
		lrshift(list, index, count);
		/* clear shifted items */
		for (i = index; i < index+count; i++)
			lset(list, i, NULL);
	}
	else if (lsize(list) == 1)
		lset(list, lfirst(list), NULL);
	
	/* fill new items with strings */

	key = hlock(str, true);

	sorted = getsorted(strlist);
	err = substrset(list, index, *str, separator, &successful, &sorted);
	getsorted(strlist) = sorted;

	hunlock(str, key);
	
	if (err) {
		/* error occured, so get rid of items that were not filled */
		llshift(list, index+successful, count-successful);
		ldecrease(list, count-successful);
	}
	
	return(err);
}

/* delete items lo to hi */
OSErr str_ldelete(StrListType strlist, ListIndex lo, ListIndex hi)
{
	ListIndex i;
	
	/* check for illegal conditions */
	assert(validstrlist(strlist));
	if (! validlohi(strlist, lo, hi)) return(ERR_STRLIST_INDEX);
	if (hi == lo) return(noErr);
	
	/* dispose of each item individually */
	for (i = lo; i < hi; i++)
		strfree(lget(getlist(strlist), i));

	/* close the gap and decrease size of list */
	llshift(getlist(strlist), lo, hi-lo);
	ldecrease(getlist(strlist), hi-lo);
	return(noErr);
}

/*----------------------------------------------------------------*/
/* Searching */
/*----------------------------------------------------------------*/

#if LIST_SEARCHES

/* return index to item containing string; uses most efficient algorithm */
OSErr str_lsearch(StrListType strlist, ListIndex lo, ListIndex hi,
							String str, ListIndex *index)
{
	if (getsorted(strlist))
		return(str_lbinsearch(strlist, lo, hi, str, index));
	return(str_lseqsearch(strlist, lo, hi, str, index));
}

/* return index to item containing string; uses sequential search */
OSErr str_lseqsearch(StrListType strlist, ListIndex lo, ListIndex hi,
							String str, ListIndex *index)
{
	assert(validstrlist(strlist));
	assert(lvaliditem(str));
	if (! validlohi(strlist, lo, hi)) return(ERR_STRLIST_INDEX);
	*index = lseqsrch(getlist(strlist), lo, hi, str);
	return(noErr);
}

/* return index to item containing string; uses binary search */
OSErr str_lbinsearch(StrListType strlist, ListIndex lo, ListIndex hi,
							String str, ListIndex *index)
{	
	assert(validstrlist(strlist));
	assert(lvaliditem(str));
	if (! validlohi(strlist, lo, hi)) return(ERR_STRLIST_INDEX);
	*index = lbinsrch(getlist(strlist), lo, hi, str);
	return(noErr);
}

#endif /* LIST_SEARCHES */

/*----------------------------------------------------------------*/
/* Sorting */
/*----------------------------------------------------------------*/

#if LIST_SORTS

/* sort the list using most efficient algorithm */
OSErr str_lsort(StrListType strlist, ListIndex lo, ListIndex hi)
{
	if (getsorted(strlist))
		return(str_lshsort(strlist, lo, hi));
	return(str_lqksort(strlist, lo, hi));
}

/* sort the list using quick sort */
OSErr str_lqksort(StrListType strlist, ListIndex lo, ListIndex hi)
{
	assert(validstrlist(strlist));
	if (! validlohi(strlist, lo, hi)) return(ERR_STRLIST_INDEX);
	lqksort(getlist(strlist), lo, hi);
	getsorted(strlist) = (lo == 0 && hi == str_lsize(strlist));
	return(noErr);
}

/* sort the list using shell sort */
OSErr str_lshsort(StrListType strlist, ListIndex lo, ListIndex hi)
{
	assert(validstrlist(strlist));
	if (! validlohi(strlist, lo, hi)) return(ERR_STRLIST_INDEX);
	lshsort(getlist(strlist), lo, hi);
	getsorted(strlist) = (lo == 0 && hi == str_lsize(strlist));
	return(noErr);
}

#endif /* LIST_SORTS */

/*----------------------------------------------------------------*/
/* Getting and setting attributes */
/*----------------------------------------------------------------*/

/* set value of an attribute of the strlist */
OSErr	str_lsetattr(StrListType strlist, StrListAttr attr, int value)
{
	switch (attr) {
	case AttrCompare:
		if (getcmpr(strlist) != value)
			getsorted(strlist) = false;
		getcmpr(strlist) = value;
		break;
	case AttrSorted:
		getsorted(strlist) = value;
		break;
	default:
		return(ERR_STRLIST_ATTR);
	}
	return(noErr);
}

/* return the value of the strlist's attribute */
OSErr	str_lgetattr(StrListType strlist, StrListAttr attr, int *value)
{
	switch (attr) {
	case AttrCompare:
		*value = getcmpr(strlist);
		break;
	case AttrSorted:
		*value = getsorted(strlist);
		break;
	default:
		return(ERR_STRLIST_ATTR);
	}
	return(noErr);
}

/*----------------------------------------------------------------*/
/* Union, difference, and intersection */
/*----------------------------------------------------------------*/

#if LIST_SETS

static OSErr setoperator(
			StrListType strlist1, StrListType strlist2, StrListType strlist3,
			ListIndex lo1, ListIndex hi1,
			ListIndex lo2, ListIndex hi2,
			VFPTR operator);
			
/* place union of list1 and list2 into list3 */
str_lunion(StrListType strlist1, StrListType strlist2, StrListType strlist3,
				ListIndex lo1, ListIndex hi1,
				ListIndex lo2, ListIndex hi2)
{
	return(setoperator(strlist1, strlist2, strlist3,
							lo1, hi1, lo2, hi2, lunion));
}

/* place difference of list1 and list2 into list3 */
str_ldiff(StrListType strlist1, StrListType strlist2, StrListType strlist3,
				ListIndex lo1, ListIndex hi1,
				ListIndex lo2, ListIndex hi2)
{
	return(setoperator(strlist1, strlist2, strlist3,
							lo1, hi1, lo2, hi2, ldiff));
}

/* place intersection of list1 and list2 into list3 */
str_linter(StrListType strlist1, StrListType strlist2, StrListType strlist3,
				ListIndex lo1, ListIndex hi1,
				ListIndex lo2, ListIndex hi2)
{
	return(setoperator(strlist1, strlist2, strlist3,
							lo1, hi1, lo2, hi2, linter));
}

/* Execute the specified binary list operator, and place the
	results into strlist3. Strlist3 should be a new, empty, list.
	Copies are made of each of the items, so that strlist3 is completely
	independent of strlist1 or strlist2.
*/
static OSErr setoperator(
			StrListType strlist1, StrListType strlist2, StrListType strlist3,
			ListIndex lo1, ListIndex hi1,
			ListIndex lo2, ListIndex hi2,
			VFPTR operator)
{
	ListType	list1, list2, list3;
	ListIndex i;
	String str;
	OSErr err = noErr;
	
	assert(validlohi(strlist1, lo1, hi1));
	assert(validlohi(strlist2, lo2, hi2));
	assert(str_lsize(strlist3) == 0);
	
	list1 = getlist(strlist1);
	list2 = getlist(strlist2);
	list3 = getlist(strlist3);
	/* execute operation, placing results into list3 */
	operator(list1, list2, list3, lo1, hi1, lo2, hi2);
	if ((err=MemError()) == noErr) {
		/* duplicate strings in list3 */
		for (i = 0; !err && i < lsize(list3); i++) {
			str = strdup(lget(list3, i));
			if (str) lset(list3, i, str);
			else err = MemError();
		}
		/* dispose of duplicated strings if error */
		if (err) {
			while (--i > 0)
				strfree(lget(list3, i));
			ldecrease(list3, lsize(list3));
		}
	}
	return(err);
}

#endif /* LIST_SETS */

/*----------------------------------------------------------------*/
/* Adding items */
/*----------------------------------------------------------------*/

/* add substrings of input string to strlist; if strlist is sorted then
	a temporary sorted list of the new substrings is created and then
	merged into the strlist. If the strlist isn't sorted then the substrings
	are simply appended to it.
*/
OSErr str_ladd(StrListType strlist, String str, char separator)
{
	ListType		list1;	/* current list */
	ListType 	list2;	/* list made from string */
	ListType		list3;	/* merged lists */
	HandleKey	key;		/* for locking/unlocking string handle */
	OSErr	err;
	
	assert(validstrlist(strlist));
	assert(str != NULL);
	list1 = getlist(strlist);
	
	/* if list isn't sorted then simply append to end of list */
	if (! getsorted(strlist))
		return(str_linsert(strlist, str_lsize(strlist), str, separator));
	
	/* else, create a new sorted list from the substrings */
	/* (ignore any errors for now since we wan't to do as much as possible) */

	key = hlock(str, true);

	err = substr2list(*str, separator, &list2);
	
	hunlock(str, key);
	
	if (! list2)
		return(err);
	if (lsize(list2) == 0) {
		/* an empty list is totally useless, so may as well return */
		ldispose(list2);
		return(err);
	}
	
	/* sort the list */
	lsetcmp(list2, lgetcmp(list1));
	lsetorder(list2, lgetorder(list1));
	lqksort(list2, lfirst(list2), lsize(list2));

	/* merge the new sorted list with the original list to get a third list */
	list3 = lmerge(list1, list2);
	if (! list3) {
		ListIndex	i;
		/* oops, an error: dispose of the new sorted list */
		if (! err) err = MemError(); /* remember error */
		for (i = lsize(list2)-1; i >= 0; i--)
			strfree(lget(list2, i));
		ldispose(list2);
		return(err);
	}
	
	/* Dispose of the original and new sorted lists.
		(But not of the data they contain, which are now in list3.)
	*/
	ldispose(list1);
	ldispose(list2);
	
	/* replace with merged list */
	getlist(strlist) = list3;
	return(err);
}

/*----------------------------------------------------------------*/
/* Stack operations */
/*----------------------------------------------------------------*/

/* push the substrings in the input string */
OSErr str_lpush(StrListType strlist, String str, char separator)
{
	return(str_linsert(strlist, str_lsize(strlist), str, separator));
}

/* pop count items off of the stack */
OSErr str_lpop(StrListType strlist, ListIndex count, char separator, String *str)
{
	OSErr	err = noErr;

	*str = NULL;
	
	/* get top items */
	err = str_ltop(strlist, count, separator, str);
	if (err) return(err);
	
	/* delete top items */
	err = str_ldelete(strlist, str_lsize(strlist)-count, str_lsize(strlist));
	
	if (err && *str)
		strfree(*str); /* an error occured, so dispose of results */
	return(err);
}

/* return the top count items of the stack; if count is greater than
	size of stack then it is reduced; multiple items are separated
	with the separator character, just like in the function str_lget.
	Notice that the returned string handle may be NULL if no items
	are retrieved (i.e., count is zero).
*/
OSErr str_ltop(StrListType strlist, ListIndex count, char separator, String *str)
{
	OSErr	err = noErr;
	ListType list;

	*str = NULL;
	assert(validstrlist(strlist));
	list = getlist(strlist);
	
	/* make sure count is in correct range */
	if (str_lsize(strlist) == 0)	return(ERR_STRLIST_EMPTY);
	if (count > str_lsize(strlist))return(ERR_STRLIST_EMPTY);
	if (count < 0)						return(ERR_STRLIST_NEGCNT);
	if (count == 0)					return(noErr);
	
	/* reverse top items; get top items; put items back in order */
	lreverse(list, lsize(list)-count, lsize(list));
	err = str_lget(strlist, lsize(list)-count, lsize(list), separator, str);
	lreverse(list, lsize(list)-count, lsize(list));
	return(err);
}

