/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*				
__________________
DESCRIPTION

Data type and operations implementing a list storing handles to strings,
with many operations inherited and enhanced from the basic list package.
This library was written to be used by XCMD's and XFCN's called from
HyperCard, and it therefore contains various enhancements which facilitate
more fluid interaction between this package and HyperCard. For instance,
searching and sorting can be adaptive, that  is, they remember the state
of the list (whether it's sorted or not), and use more efficient
algorithms (qsort won't be used on a presorted list). If a list
is already sorted then functions try to maintain such an order,
or at least to detect a violation of sorted order by setting
or clearing a sorted flag.

Most functions can operate on more than one item at a time. For instance,
when adding items to a list, more than one item can be specified
by delimiting each item in the input string with a special
separator character. This separator character delimits the string
into substrings, each of which is then operate on separately.
Thus, to add all the lines in a HyperCard container
as individual items, the separator character would be the return
character and the input string would be the contents of the HyperCard
container. Similarly, to extract more than one item from the list,
specify a count of the number of items to extract (to the str_lget
function); the function will build one "superstring" containing
as substrings the individual items.

Each list has associated with it certain attributes. One of these
is the sorted attribute mentioned above. Another is the compare
attribute, which controls the rules for searching and sorting.

__________________
ERRORS

A set of error codes are defined in addition to the standard Macintosh
system errors. The error codes are positive integers, so as not to
conflict with system errors which appear to be negative.

__________________
FUNCTIONS

Creating and Disposing
str_lnew			allocate a new strlist
str_ldispose	deallocate an strlist

Getting and Setting List's Size
str_lsize		return number of items in strlist
str_lsetsize	set new size for list; if list shrinks then extra items
					are discarded, but if list grows then empty items are
					appended

Getting and Setting items
str_lget			return the contents of requested items following an
					indexed item; if more than one item is requested then items
					are separated by a separator character, which defaults to
					a comma (,).
str_lset			set each item from the indexed item on to the corresponding
					substring; list is expanded to accommodate extra substrings.

Inserting and Deleting Items
str_linsert		insert substrings before indexed item
str_ldelete		delete count substrings starting at indexed item

Adding Items
str_ladd			adds each substring to the list; if list is sorted then
					substrings are merged into the list, otherwise they're appended.

Searching
str_lsearch		return index to requested item; uses binary search if list
					is sorted, otherwise sequential search
str_lseqsearch	search for item using sequential search
str_lbinsearch	search for item using binary search

Sorting
str_lsort		sort the list; uses shell sort if list is sorted, otherwise
					uses quick sort
str_lqksort		sort the list using quick sort
str_lshsort		sort the list using shell sort

Set Operations
str_lunion		places union of two lists into a third list
str_ldiff		places difference of two lists into a third list
str_linter		places intersection of two lists into a third list

Stack Operations
str_lpush		push each substring onto the list
str_lpop			pop requested number of items off of list; if more than one
					item is requested then items are separated with separator
					character.
str_ltop			same as str_lpop except no items are removed from list.

Setting and Getting Attributes
str_lsetattr	set value of an attribute
str_lgetattr	get value of an attribute
*/

/*
DESCRIPTION

Functions for converting a list represented as substrings into
an array list.

OPERATIONS

substr2list	Convert list of substrings into a newly created array list.
substrset	Sets items in a list to substrings. The list's size is increased
				if substrings would extend beyond end. The succesful parameter
				is set to the number of strings which were set successfully. If
				some strings couldn't be set and the list had to be expanded then
				the list is shrunk back down by the number of remaining
				substrings. If the sorted parameter is true then each substring
				added is checked for being in sorted order; the parameter is
				set to false if a violation of sorted order is found.
*/

#ifndef _H_strlistlib
#define _H_strlistlib

#ifndef _H_listlib
#include "listlib.h"
#endif

#ifndef _H_strlib
#include "strlib.h"
#endif

#ifndef _H_errorlib
#include "errorlib.h"
#endif

/* list attributes */
typedef enum {
	AttrNULL,
	AttrCompare,	/* method for comparing strings */
	AttrSorted,		/* true/false depending on whether list is sorted */
	AttrLast
} StrListAttr;

/* The term StrListType stands for "External Hypercard Array List" */
typedef struct {
	ListType			list;		/* the list */
	Boolean			sorted;	/* true if list is in sorted order */
	StrCompareType	cmpr;		/* method for comparing strings */
} StrListStruct, **StrListType;

OSErr substr2list(char *str, char separator, ListType *listPtr);
OSErr substrset(ListType list, ListIndex index, char *str,
					char separator, ListIndex *successful, Boolean *sorted);

/* creating and disposing */
OSErr	str_lnew(StrListType *strlistPtr);
OSErr	str_ldispose(StrListType strlist);
void	str_lupdate(StrListType strlist);

/* getting and setting size */
ListIndex	str_lsize(StrListType strlist);			
OSErr 		str_lsetsize(StrListType strlist, ListIndex newsize);
	
/* getting and setting items */
OSErr	str_lset(StrListType strlist, ListIndex lo, ListIndex hi,
					String items, char separator);
OSErr	str_lget(StrListType strlist, ListIndex lo, ListIndex hi,
					char separator, String *result);

/* inserting and deleting items */
OSErr	str_ldelete(StrListType strlist, ListIndex lo, ListIndex hi);
OSErr str_linsert(StrListType strlist, ListIndex index, String items,
						char separator);

/* searching */
OSErr str_lsearch(StrListType strlist, ListIndex lo, ListIndex hi,
						String str, ListIndex *index);
OSErr str_lseqsearch(StrListType strlist, ListIndex lo, ListIndex hi,
						String str, ListIndex *index);
OSErr str_lbinsearch(StrListType strlist, ListIndex lo, ListIndex hi,
						String str, ListIndex *index);

/* sorting */
OSErr str_lsort(StrListType strlist, ListIndex lo, ListIndex hi);
OSErr str_lqksort(StrListType strlist, ListIndex lo, ListIndex hi);
OSErr str_lshsort(StrListType strlist, ListIndex lo, ListIndex hi);

/* set operations */
str_lunion(StrListType strlist1, StrListType strlist2, StrListType strlist3,
				ListIndex lo1, ListIndex hi1,
				ListIndex lo2, ListIndex hi2);
str_ldiff(StrListType strlist1, StrListType strlist2, StrListType strlist3,
				ListIndex lo1, ListIndex hi1,
				ListIndex lo2, ListIndex hi2);
str_linter(StrListType strlist1, StrListType strlist2, StrListType strlist3,
				ListIndex lo1, ListIndex hi1,
				ListIndex lo2, ListIndex hi2);

/* adding items */
OSErr str_ladd(StrListType strlist, String items, char separator);

/* stack operations */
OSErr	str_lpush(StrListType strlist, String items, char separator);
OSErr	str_lpop(StrListType strlist, ListIndex count, char separator, String *str);
OSErr	str_ltop(StrListType strlist, ListIndex count, char separator, String *str);

/* setting and getting attributes */
OSErr	str_lsetattr(StrListType strlist, StrListAttr attr, int value);
OSErr	str_lgetattr(StrListType strlist, StrListAttr attr, int *value);

#endif
