/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
	These functions are an interface between the parameters passed
	from HyperCard and the strtree library. The functions are
	meant to be called once the appropriate command has been identified,
	and, in most cases, after the specified tree has been extracted from the
	parameter list. There is a one-to-one correspondence between these
	functions and the features of the XFCN.
*/

#include <MyHeaders>
#include "TreeMain.h"
#include "lookuptabl.h"

static OSErr traverse(GlobalData gdata, short arg, StrTreeType tree, FPTR func);
static OSErr param2type(GlobalData gdata, short arg, TreeOutputType *type);

/*----------------------------------------------------------------*/
/* Getting info about XFCN */
/*----------------------------------------------------------------*/

OSErr ExecUsage(GlobalData gdata, short arg)
{
	CStr255 str;
	GetIndString(str, GetRsrcID(gdata, RSRC_INFO_ID), RSRC_USAGE_INDEX);
	result2string(GetParam(gdata), str);
	return(noErr);
}

OSErr ExecVersion(GlobalData gdata, short arg)
{
	CStr255	str;
	GetVerString(str, GetRsrcID(gdata, RSRC_INFO_ID), RSRC_VERSION_INDEX,
					__DATE__, __TIME__);
	result2string(GetParam(gdata), str);
	return(noErr);
}

/*----------------------------------------------------------------*/
/* Checking for errors */
/*----------------------------------------------------------------*/

OSErr ExecError(GlobalData gdata, short arg)
{
	return(GetError(gdata));
}

/*----------------------------------------------------------------*/
/* Creating and disposing */
/*----------------------------------------------------------------*/

OSErr	ExecNew(GlobalData gdata, short arg)
{
	String	tree_name;
	StrCompareType cmpr;
	OSErr		err = noErr;
	
	tree_name = NULL;
	cmpr = STR_COMPARE_IGNORECASE;
	if ((err=param2handle(GetParam(gdata), arg++, &tree_name)) == noErr) {
		if (arg < GetParamCount(gdata))
			err = param2cmpr(gdata, arg++, &cmpr);
		if (err == noErr)
			err = str_forestinsert(GetForest(gdata), tree_name, cmpr);
	}
	return(err);
}

OSErr	ExecDispose(GlobalData gdata, short arg)
{
	String	tree_name;
	OSErr		err = noErr;
	
	tree_name = NULL;
	if ((err=param2handle(GetParam(gdata), arg++, &tree_name)) == noErr)
		err = str_forestdelete(GetForest(gdata), tree_name);
	return(err);
}

/*----------------------------------------------------------------*/
/* Inserting and deleting */
/*----------------------------------------------------------------*/

OSErr ExecInsert(GlobalData gdata, short arg, StrTreeType tree)
{
	String	keys;
	String	data;
	char		keys_separator;
	char		data_separator;
	OSErr		err = noErr;
	
	keys = data = NULL;
	keys_separator = data_separator = 0;
	if ((err = param2handle(GetParam(gdata), arg++, &keys)) == noErr) {
		if (param2handle(GetParam(gdata), arg++, &data) == noErr)
			if (param2separator(GetParam(gdata), arg++, &keys_separator) == noErr)
				if (param2separator(GetParam(gdata), arg++, &data_separator) == noErr)
					/* ok */ ;
		err = str_tinsert_many(tree, keys, data,
										keys_separator, data_separator);
	}
	return(err);
}

OSErr ExecDelete(GlobalData gdata, short arg, StrTreeType tree)
{
	String	keys;
	char		keys_separator;
	OSErr		err = noErr;
	
	keys = NULL;
	keys_separator = 0;
	if ((err = param2handle(GetParam(gdata), arg++, &keys)) == noErr) {
		if (param2separator(GetParam(gdata), arg++, &keys_separator) == noErr)
			/* ok */ ;
		err = str_tdelete_many(tree, keys, keys_separator);
	} 
	return(err);
}

/*----------------------------------------------------------------*/
/* Setting and getting data */
/*----------------------------------------------------------------*/

OSErr ExecSet(GlobalData gdata, short arg, StrTreeType tree)
{
	String	keys;
	String	data;
	char		keys_separator;
	char		data_separator;
	OSErr		err = noErr;
	
	keys = data = NULL;
	keys_separator = data_separator = 0;
	if ((err = param2handle(GetParam(gdata), arg++, &keys)) == noErr) {
		if ((err = param2handle(GetParam(gdata), arg++, &data)) == noErr) {
			if (param2separator(GetParam(gdata), arg++, &keys_separator) == noErr)
				if (param2separator(GetParam(gdata), arg++, &data_separator) == noErr)
					/* ok */ ;
			err = str_tset_many(tree, keys, data, keys_separator, data_separator);
		}
	} 
	return(err);
}

OSErr ExecGet(GlobalData gdata, short arg, StrTreeType tree)
{
	String	keys;
	String	data;
	char		keys_separator;
	char		data_separator;
	OSErr		err = noErr;
	
	keys = data = NULL;
	keys_separator = data_separator = 0;
	if ((err = param2handle(GetParam(gdata), arg++, &keys)) == noErr) {
		if (param2separator(GetParam(gdata), arg++, &keys_separator) == noErr)
			if (param2separator(GetParam(gdata), arg++, &data_separator) == noErr)
				/* ok */ ;
		err = str_tget_many(tree, keys, keys_separator, data_separator, &data);
		result2handle(GetParam(gdata), data);
	}
	return(err);
}

/*----------------------------------------------------------------*/
/* Accessing nodes */
/*----------------------------------------------------------------*/

OSErr ExecLeft(GlobalData gdata, short arg, StrTreeType tree)
{
	String	key, left;
	OSErr		err = noErr;
	
	key = left = NULL;
	if ((err = param2handle(GetParam(gdata), arg++, &key)) == noErr) {
		err = str_tleft(tree, key, &left);
		if (!err) result2handle(GetParam(gdata), left);
	}
	return(err);
}

OSErr ExecRight(GlobalData gdata, short arg, StrTreeType tree)
{
	String	key, right;
	OSErr		err = noErr;
	
	key = right = NULL;
	if ((err = param2handle(GetParam(gdata), arg++, &key)) == noErr) {
		err = str_tright(tree, key, &right);
		if (!err) result2handle(GetParam(gdata), right);
	}
	return(err);
}

OSErr ExecParent(GlobalData gdata, short arg, StrTreeType tree)
{
	String	key, parent;
	OSErr		err = noErr;
	
	key = parent = NULL;
	if ((err = param2handle(GetParam(gdata), arg++, &key)) == noErr) {
		err = str_tparent(tree, key, &parent);
		if (!err) result2handle(GetParam(gdata), parent);
	}
	return(err);
}

OSErr ExecRoot(GlobalData gdata, short arg, StrTreeType tree)
{
	String	root = NULL;
	OSErr		err = noErr;
	
	err = str_troot(tree, &root);
	if (!err) result2handle(GetParam(gdata), root);
	return(err);
}

/*----------------------------------------------------------------*/
/* Getting height of tree */
/*----------------------------------------------------------------*/

OSErr ExecHeight(GlobalData gdata, short arg, StrTreeType tree)
{
	String	key;
	TreeIndex height;
	CStr31	str;
	
	key = NULL;
	(void) param2handle(GetParam(gdata), arg, &key);
	height = str_theight(tree, key);
	result2handle(GetParam(gdata), stralloc(ltoa(height, str)));
	return(MemError());
}

/*----------------------------------------------------------------*/
/* Range searching */
/*----------------------------------------------------------------*/

OSErr	ExecRange(GlobalData gdata, short arg, StrTreeType tree)
{
	String	low_key;
	String	high_key;
	String	root_key;
	char		output_separator;
	String	range_output;
	TreeOutputType type;
	OSErr		err = noErr;
	
	output_separator = 0;
	root_key = low_key = high_key = NULL;
	type = TREE_OUTPUT_KEYS;
	range_output = NULL;
	
	if ((err = param2handle(GetParam(gdata), arg++, &low_key)) == noErr) {
		if ((err = param2handle(GetParam(gdata), arg++, &high_key)) == noErr) {
			if (param2type(gdata, arg++, &type) == noErr)
				if (param2separator(GetParam(gdata), arg++, &output_separator) == noErr)
					if (param2handle(GetParam(gdata), arg++, &root_key) == noErr)
						/* ok */ ;
			err = str_trange(tree, root_key, low_key, high_key,
									output_separator, type, &range_output);
			result2handle(GetParam(gdata), range_output);
		}
	}
	return(err);
}

/*----------------------------------------------------------------*/
/* Traversing */
/*----------------------------------------------------------------*/

OSErr ExecInorder(GlobalData gdata, short arg, StrTreeType tree)
{
	return(traverse(gdata, arg, tree, str_tinorder));
}

OSErr ExecPreorder(GlobalData gdata, short arg, StrTreeType tree)
{
	return(traverse(gdata, arg, tree, str_tpreorder));
}

OSErr ExecPostorder(GlobalData gdata, short arg, StrTreeType tree)
{
	return(traverse(gdata, arg, tree, str_tpostorder));
}

OSErr ExecLevelorder(GlobalData gdata, short arg, StrTreeType tree)
{
	return(traverse(gdata, arg, tree, str_tlevelorder));
}

static OSErr traverse(GlobalData gdata, short arg, StrTreeType tree, FPTR func)
{
	String	root_key;
	String	output;
	char		separator;
	TreeOutputType type;
	OSErr		err = noErr;
	
	separator = 0;
	type = TREE_OUTPUT_KEYS;
	root_key = output = NULL;
	if (param2type(gdata, arg++, &type) == noErr)
		if (param2separator(GetParam(gdata), arg++, &separator) == noErr)
			if (param2handle(GetParam(gdata), arg++, &root_key) == noErr)
				/* ok */ ;
	err = func(tree, root_key, separator, type, &output);
	result2handle(GetParam(gdata), output);
	return(err);
}

static OSErr param2type(GlobalData gdata, short arg, TreeOutputType *type)
{
	String	str;
	OSErr		err;
	short		index;

	if ((err = param2handle(GetParam(gdata), arg++, &str)) == noErr)
		err = hlookupTabl(GetRsrcID(gdata, RSRC_OUTPUTS_ID), str, &index);
	if (err) return(err);
	switch (index) {
	case TREE_OUTPUT_KEYS:
	case TREE_OUTPUT_DATA:
	case TREE_OUTPUT_INDEXES:
		*type = index;
		return(noErr);
	}
	return(ERR_BASIC_USAGE);
}

/*----------------------------------------------------------------*/
/* Setting and getting attributes */
/*----------------------------------------------------------------*/

OSErr ExecSetAttr(GlobalData gdata, short arg, StrTreeType tree)
{
	TreeAttrType	attr;
	StrCompareType	cmpr;
	Boolean			bool;
	int				value;
	OSErr 			err = noErr;
	
	value = false;
	if ((err=param2attr(gdata, arg++, &attr)) == noErr) {
		switch (attr) {
		case TREE_ATTR_COMPARE:
			err = param2cmpr(gdata, arg++, &cmpr);
			value = cmpr;
			break;
		case TREE_ATTR_UNIQUE:
			err = param2bool(GetParam(gdata), arg++, &bool);
			value = bool;
			break;
		}
		if (! err)
			err = str_tsetattr(tree, attr, value);
	}
	return(err);
}

OSErr ExecGetAttr(GlobalData gdata, short arg, StrTreeType tree)
{
	TreeAttrType	attr;
	int				value;
	OSErr 			err = noErr;
	
	value = false;
	if ((err=param2attr(gdata, arg++, &attr)) == noErr) {
		if ((err = str_tgetattr(tree, attr, &value)) == noErr) {
			switch (attr) {
			case TREE_ATTR_COMPARE:
				result2cmpr(gdata, value);
				break;
			case TREE_ATTR_UNIQUE:
				result2bool(GetParam(gdata), value);
				break;
			}
		}
	}
	return(err);
}
