/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
	Main file for the BinaryTree XFCN. See manual for more details
	on program, its use, and its organization.
*/

#include <MyHeaders>
#include "TreeMain.h"
#include "strlib.h"
#include "codelib.h"
#include "lookuptabl.h"

/*--------------------------------------------------------------------*/
/* Prototypes
/*--------------------------------------------------------------------*/

static OSErr Initialize(XCmdBlockPtr param, Handle rsrc,
									Boolean has128k, GlobalData *gdataPtr);
static void  Execute(GlobalData gdata);
static OSErr DoExecute(GlobalData gdata, short arg, CommandType command);

static OSErr param2strtree(GlobalData gdata, short arg,
									StrTreeType *strtreePtr);
static OSErr param2command(GlobalData gdata, short arg, CommandType *command);

/*--------------------------------------------------------------------*/
/* The main routine
/*--------------------------------------------------------------------*/

int codersrc_main(XCmdBlockPtr param, Handle rsrc, Boolean has128k)
{	
	GlobalData gdata; /* global data used by the program */
	OSErr err;
	
	err = Initialize(param, rsrc, has128k, &gdata);
	if (! err)
		Execute(gdata);
	return(0);
}

static OSErr Initialize(XCmdBlockPtr param,
								Handle rsrc,
								Boolean has128k,
								GlobalData *gdataPtr)
{
	if (! has128k)
		return(ERR_BASIC_ROMVER);
	result2handle(param, NULL);
	return(InitGlobalData(param, rsrc, gdataPtr));
}

/*--------------------------------------------------------------------*/
/* Functions for executing the specified command
/*--------------------------------------------------------------------*/

static void Execute(GlobalData gdata)
{
	StrTreeType	strtree;	/* strtree to operate on */
	CommandType	command;	/* command to execute */
	OSErr			err = noErr;

	/* get and then execute the command */
	err = param2command(gdata, 0, &command);
	if (! err)
		err = DoExecute(gdata, 1, command);
	
	/* set error code (if command was ok then error code is cleared) */
	if (command != COMMAND_ERROR)
		SetError(gdata, err);

	/* The following commands set the return value to the error code if
		there was an error. This makes it simpler to check for errors
		from the script since the script doesn't need to make an
		extra call with the Error command to find out if there was
		an error.
	*/
	if (err) {
		switch (command) {
		case COMMAND_NULL:
		case COMMAND_DELETE:
		case COMMAND_DISPOSE:
		case COMMAND_ERROR:
		case COMMAND_INSERT:
		case COMMAND_NEW:
		case COMMAND_SET:
		case COMMAND_SETATTR:
			result2error(GetParam(gdata), err);
			break;
		}
	}
}

/* execute the specified command */
static OSErr DoExecute(GlobalData gdata, short arg, CommandType command)
{
	StrTreeType	strtree = NULL;
	OSErr			err = noErr;
	
	/* get the strtree if needed for this command */
	switch (command) {
	case COMMAND_ERROR:
	case COMMAND_VERSION:
	case COMMAND_USAGE:
	case COMMAND_NEW:
	case COMMAND_DISPOSE:
		break; /* do nothing */
	default:
		err = param2strtree(gdata, arg++, &strtree);
		if (err)
			return(err);
		break;
	}
	
	/* execute the command */
	switch (command) {
	case COMMAND_VERSION:	return(ExecVersion(gdata, arg));
	case COMMAND_USAGE:		return(ExecUsage(gdata, arg));
	case COMMAND_ERROR:		return(ExecError(gdata, arg));
	
	case COMMAND_NEW:			return(ExecNew(gdata, arg));
	case COMMAND_DISPOSE:	return(ExecDispose(gdata, arg));

	case COMMAND_INSERT:		return(ExecInsert(gdata, arg, strtree));
	case COMMAND_DELETE:		return(ExecDelete(gdata, arg, strtree));

	case COMMAND_GET:			return(ExecGet(gdata, arg, strtree));
	case COMMAND_SET:			return(ExecSet(gdata, arg, strtree));
	
	case COMMAND_GETATTR:	return(ExecGetAttr(gdata, arg,strtree));
	case COMMAND_SETATTR:	return(ExecSetAttr(gdata,arg,strtree));

	case COMMAND_INORDER:	return(ExecInorder(gdata, arg, strtree));
	case COMMAND_PREORDER:	return(ExecPreorder(gdata, arg, strtree));
	case COMMAND_POSTORDER:	return(ExecPostorder(gdata, arg, strtree));
	case COMMAND_LEVELORDER:return(ExecLevelorder(gdata, arg, strtree));

	case COMMAND_RANGE:		return(ExecRange(gdata, arg, strtree));

	case COMMAND_LEFT:		return(ExecLeft(gdata, arg, strtree));
	case COMMAND_PARENT:		return(ExecParent(gdata, arg, strtree));
	case COMMAND_RIGHT:		return(ExecRight(gdata, arg, strtree));
	case COMMAND_ROOT:		return(ExecRoot(gdata, arg, strtree));

	case COMMAND_HEIGHT:		return(ExecHeight(gdata, arg, strtree));
	}
	return(ERR_BASIC_COMMAND);
}

/*--------------------------------------------------------------------*/
/* Parameter extraction functions.
/*--------------------------------------------------------------------*/

/* convert parameter into a string tree */
static OSErr param2strtree(GlobalData gdata, short arg,
									StrTreeType *strtreePtr)
{
	OSErr		err;
	String	name;
	
	err = param2handle(GetParam(gdata), arg++, &name);
	if (err) return(err);
	return(str_forestfind(GetForest(gdata), name, strtreePtr));
}

/* convert a parameter into a command number */
static OSErr param2command(GlobalData gdata, short arg, CommandType *command)
{
	String	str;
	OSErr		err;
	short		index;

	*command = COMMAND_NULL;
	if ((err = param2handle(GetParam(gdata), arg++, &str)) == noErr)
		err = hlookupTabl(GetRsrcID(gdata, RSRC_COMMANDS_ID), str, &index);
	if (err) return(err);
	if (index <= COMMAND_NULL || COMMAND_LAST <= index)
		return(ERR_BASIC_COMMAND);
	*command = index;
	return(noErr);
}

/* convert parameter into an attribute number */
OSErr	param2attr(GlobalData gdata, short arg, TreeAttrType *attr)
{
	String	str;
	OSErr		err;
	short		index;

	if ((err = param2handle(GetParam(gdata), arg++, &str)) == noErr)
		err = hlookupTabl(GetRsrcID(gdata, RSRC_ATTRIBUTES_ID), str, &index);
	if (err) return(err);
	if (index <= TREE_ATTR_NULL || TREE_ATTR_LAST <= index)
		return(ERR_STRTREE_ATTR);
	*attr = index;
	return(noErr);
}

/* convert parameter into a string comparison attribute number */
OSErr param2cmpr(GlobalData gdata, short arg, StrCompareType *cmpr)
{
	String	str;
	OSErr		err;
	short		index;

	if ((err = param2handle(GetParam(gdata), arg++, &str)) == noErr)
		err = hlookupTabl(GetRsrcID(gdata, RSRC_COMPARES_ID), str, &index);
	if (err) return(err);
	switch (index) {
	case STR_COMPARE_EXACT:
	case STR_COMPARE_IGNORECASE:
	case STR_COMPARE_INTERNATIONAL:
	case STR_COMPARE_FLOAT:
		*cmpr = index;
		return(noErr);
	}
	return(ERR_STRTREE_ATTR);
}

/* set result two a string comparison number */
void result2cmpr(GlobalData gdata, StrCompareType cmpr)
{
	CStr255	str;
	OSErr		err;
	
	err = lookupTablNum(GetRsrcID(gdata, RSRC_COMPARES_ID), cmpr, str);
	if (! err)
		result2string(GetParam(gdata), str);
}
