/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

#include <MyHeaders>
#include "str_tprivate.h"

/*
	This file contains functions for acting on many nodes at a time.
	These are:
		str_tinsert_many	Inserts any number of key and datum pairs.
		str_tdelete_many	Deletes any number of nodes given their keys.
		str_tget_many		Gets any number of datum fields of the nodes.
								given the nodes' keys. This function makes
								use of the string tree's buffer output functions.
		str_tset_many		Sets any number of datum fields of the nodes
								given the nodes' keys.
*/

/*----------------------------------------------------------------*/
/* Inserting and deleting many nodes at a time */
/*----------------------------------------------------------------*/

/*	Insert the given keys into the tree, and set every key to its
	corresponding datum. Each key is separated by the key separator
	character. Each datum is separated by the datum separator
	character, unless the datum separator is NULL, in which case the
	key separator is used instead. The handle to the data
	may be NULL, in which case NULL's are used instead of actual data.
	If there are fewer items in the data than in the keys then only
	the extra keys are paired with empty data (so that the datum fields
	of new nodes are cleared). If there are mode data items than
	there are keys then the extra items are ignored.
	
	There are two basic approaches that can be taken to inserting
	M nodes into a tree with N existing nodes:
	
	1. Build a tree out of the new keys and then join it with
		the original tree. 
	2. Insert each key succesively into the original tree.

	Since a tree may have the attribute that every node be unique,
	it is simplest to insert a node at a time, and have the
	insert function check that the node doesn't exist before
	inserting. I already have a function which will insert
	a single key into a tree, and I know the function works, so I
	will use the second method (insert each item succesively).
*/
OSErr	str_tinsert_many(StrTreeType strtree, String keys, String data,
						char key_separator, char datum_separator)
{
	OSErr err = noErr;
	char	*keys_list;	/* pointer to list of keys */
	char	*data_list;	/* pointer to list of data */
	SignedByte keys_state;	/* saved state of locked keys handle */
	SignedByte data_state;	/* saved state of locked data handle */
	
	assert(str_tvalid(strtree));
	assert(validhandle(keys));

	if (! datum_separator)
		datum_separator = key_separator;
	
	/* lock and dereference keys and data lists */
	keys_state = hlock(keys, true);
	keys_list = *keys;
	if (data) {
		data_state = hlock(data, true);
		data_list = *data;
	}
		
	/* get each key and datum and insert them into the tree */
	while (*keys_list && ! err) {
		size_t	key_len;		/* length of current key */
		size_t	datum_len;	/* length of current datum */
		char		key_delim;	/* saved delimiter of current key */
		char		datum_delim;/* saved delimiter of current datum */
		Handle	key_handle;		/* handle to copy of current key */
		Handle	datum_handle;	/* handle to copy of current datum */
		
		/* terminate current key and datum with a NULL */
		key_len = substrlen(keys_list, key_separator);
		key_delim = keys_list[key_len];
		keys_list[key_len] = 0;
		if (data) {
			datum_len = substrlen(data_list, datum_separator);
			datum_delim = data_list[datum_len];
			data_list[datum_len] = 0;
		}
		
		/* duplicate and insert key and datum */
		key_handle = stralloc(keys_list);
		if (key_handle) {
			/* duplicate datum */
			if (data && *data_list) { 
				datum_handle = stralloc(data_list);
				if (! datum_handle)
					err = MemError();
			}
			else
				datum_handle = NULL;
			
			/* Insert. Since the copy flag is false, insert won't
				make of copy of the key and datum, which is good,
				since we've already done that. */
			if (! err)
				err = str_tinsert(strtree, key_handle, datum_handle, false);
			
			/* cleanup if error */
			if (err) {
				if (datum_handle)
					strfree(datum_handle);
				strfree(key_handle);
			}
		}
		else
			err = MemError();
		
		/* restore current key and datum delimiters */
		keys_list[key_len] = key_delim;
		if (data)
			data_list[datum_len] = datum_delim;

		/* advance to next key and datum */
		keys_list += key_len;
		if (key_delim) ++keys_list;
		if (data) {
			data_list += datum_len;
			if (datum_delim) ++data_list;
		}
	}
	
	/* restore states of handles to keys and data */
	hunlock(keys, keys_state);
	if (data)
		hunlock(data, data_state);
		
	return(err);
}

/* Delete the nodes with the given keys from the tree. Each key
	is delimited by the key_separator character.
*/
OSErr str_tdelete_many(StrTreeType strtree, String keys, char key_separator)
{
	char *keys_list;			/* pointer to list of keys */
	SignedByte keys_state;	/* saved state of locked keys handle */
	OSErr err = noErr;

	assert(str_tvalid(strtree));
	assert(str_tvalidkey(keys));
	
	/* lock and dereference keys list */
	keys_state = hlock(keys, true);
	keys_list = *keys;
	
	/* get each key and insert it into the tree */
	while (*keys_list) {
		size_t	key_len;		/* length of current key */
		char		key_delim;	/* saved delimiter of current key */
		String	tmp_key;		/* temporary copy of current key */
		
		/* terminate current key with a NULL */
		key_len = substrlen(keys_list, key_separator);
		key_delim = keys_list[key_len];
		keys_list[key_len] = 0;
		
		/* make a temporary copy of the key and then delete it */
		tmp_key = stralloc(keys_list);
		if (tmp_key) {
			err = str_tdelete(strtree, tmp_key);
			strfree(tmp_key);
		}
		else
			err = MemError();
		
		/* restore current key's delimiter */
		keys_list[key_len] = key_delim;

		if (err) break;

		/* advance to next key  */
		keys_list += key_len;
		if (key_delim) ++keys_list;
	}
	
	/* restore state of keys list handle */
	hunlock(keys, keys_state);

	return(err);
}

/*----------------------------------------------------------------*/
/* Setting and getting many nodes at once */
/*----------------------------------------------------------------*/

/* set the datum fields of the nodes with the given keys to copies
	of the given data. Each key is separated by the key_separator
	character, and each corresponding datum is separated by the
	datum_separator character. If the datum_separator is NULL
	then the key_separator is used instead.
	
	This function requires M searches in the binary tree, where
	M is the number of keys. This function takes around O(MlogN),
	where N is the number of nodes in the tree.
*/
OSErr	str_tset_many(StrTreeType strtree, String keys, String data,
						char key_separator, char datum_separator)
{
	OSErr err = noErr;
	char	*keys_list;	/* pointer to list of keys */
	char	*data_list;	/* pointer to list of data */
	SignedByte keys_state;	/* saved state of locked keys handle */
	SignedByte data_state;	/* saved state of locked data handle */
	
	assert(str_tvalid(strtree));
	assert(validhandle(keys));
	assert(validhandle(data));

	if (! datum_separator)
		datum_separator = key_separator;
	
	/* lock and dereference keys and data lists */
	keys_state = hlock(keys, true);
	data_state = hlock(data, true);
	keys_list = *keys;
	data_list = *data;
	
	/* get each key */
	while (*keys_list) {
		size_t	key_len;		/* length of current key */
		size_t	datum_len;	/* length of current datum */
		char		key_delim;	/* saved delimiter of current key */
		char		datum_delim;/* saved delimiter of current datum */
		Handle	datum;		/* handle to current datum */
		
		/* terminate current key and datum with a NULL */
		key_len = substrlen(keys_list, key_separator);
		key_delim = keys_list[key_len];
		keys_list[key_len] = 0;
		datum_len = substrlen(data_list, datum_separator);
		datum_delim = data_list[datum_len];
		data_list[datum_len] = 0;
		
		/* duplicate and set datum */
		datum = stralloc(data_list);
		if (datum) {
			/* unfortunately, we must make a temporary copy of
				the key, since all keys are stored as handles */
			String tmp_key = stralloc(keys_list);
			if (tmp_key) {
				/* since the copy flag is false, the set function
					won't make a copy of the datum, which is good
					since we've already done that. */
				err = str_tset(strtree, tmp_key, datum, false);
				strfree(tmp_key);
			}
			else
				err = MemError();
			if (err)
				strfree(datum);
		}
		else
			err = MemError();
		
		/* restore current key and datum delimiters */
		keys_list[key_len] = key_delim;
		data_list[datum_len] = datum_delim;

		if (err) break;

		/* advance to next key and datum */
		keys_list += key_len;
		if (key_delim) ++keys_list;
		data_list += datum_len;
		if (datum_delim) ++data_list;
	}
	
	/* restore states of keys and data list handles */
	hunlock(keys, keys_state);
	hunlock(data, data_state);
		
	return(err);
}

/* return a string containing all of the datum fields of the nodes
	with the given keys. Each key is separated by the key_separator
	character, and each corresponding datum in the output is separated
	by the datum_separator character. If the datum_separator is NULL then
	then the key_separator is used instead.
	
	This function requires M searches in the binary tree, where
	M is the number of requested keys. Therefore, this function
	takes O(MlogN) on average (N is number of nodes in tree)
*/
OSErr	str_tget_many(StrTreeType strtree, String keys,
						char key_separator, char datum_separator,
						String *output)
{
	OSErr err = noErr;
	char	*keys_list;	/* pointer to list of keys */
	SignedByte keys_state;	/* saved state of locked keys list */
	StrTreeBufType stbuf;	/* output buffer */
	
	assert(str_tvalid(strtree));
	assert(validhandle(keys));

	*output = NULL;
	
	/* initialize output buffer */
	err = str_tinitbuf(&stbuf, datum_separator);
	if (err) return(err);

	if (! datum_separator)
		datum_separator = key_separator;
	
	/* lock and dereference keys list */
	keys_state = hlock(keys, true);
	keys_list = *keys;
	
	/* get each datum */
	while (*keys_list) {
		size_t	key_len;		/* length of current key */
		char		key_delim;	/* saved delimiter of current key */
		String	tmp_key;		/* temporary handle to current key */
		String	datum;		/* current datum */
		
		/* terminate current key with a NULL */
		key_len = substrlen(keys_list, key_separator);
		key_delim = keys_list[key_len];
		keys_list[key_len] = 0;
		
		/* unfortunately, we must make a temporary copy of
			the key, since all keys are stored as handles */
		tmp_key = stralloc(keys_list);
		if (tmp_key) {
			/* Get the datum. Notice that the copy flag is false,
				so that the get function returns the actual handle to
				the datum, not a copy. This is good since the buffer
				output function will make a copy of the datum when
				it appends it to the output. */
			err = str_tget(strtree, tmp_key, &datum, false);
			strfree(tmp_key);
		}
		else
			err = MemError();
		
		/* restore current key's delimiter */
		keys_list[key_len] = key_delim;
		
		if (err) break;
		
		/* append datum to output buffer */
		err = str_toutputstr(&stbuf, datum);
		if (err) break;
		
		/* advance to next key */
		keys_list += key_len;
		if (key_delim) ++keys_list;

	}
	
	/* restore state of locked keys list */
	hunlock(keys, keys_state);
				
	/* end use of buffer, placing its contents into the output pointer */
	if (! err)
		err = str_tendbuf(&stbuf, output);
	else
		(void) str_tendbuf(&stbuf, output);
	
	return(err);
}

