/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

#include <MyHeaders>
#include "str_tprivate.h"

/*
	This file contains function functions which recursively visit
	nodes in the string tree, and collect data about the nodes
	encountered. Also included is a range searching function.
	The functions are:
		str_tinorder		Inorder traversal
		str_tpreorer		Preorder traversal
		str_tpostorder		Postorder traversal
		str_tlevelorder	Level order traversal
		str_trange			Range searching
	Each function makes use of the string tree's output buffer
	(in "str_tbuf.c"). Also, the functions may output different
	data about each node, such as the nodes' keys, data fields,
	or indexes.
	
	These functions use the binary tree library's traversal
	functions, which call an action function for each node
	encountered. The data field of this action function may
	contain any data needed by the application, so we use
	it here to pass a pointer to the output buffer.
*/

/* these are the action functions calls by the binary tree library */
static int outputdatum(TREE_ACTION_PROTOTYPE);
static int outputkey(TREE_ACTION_PROTOTYPE);
static int outputindex(TREE_ACTION_PROTOTYPE);

static OSErr traverse(StrTreeType strtree, String rootkey, char separator,
							TreeOutputType type, String *output, FPTR func);

/*----------------------------------------------------------------*/
/* Traversing tree */
/*----------------------------------------------------------------*/

/* visit nodes in in-order; see description of traverse function */
OSErr	str_tinorder(StrTreeType strtree,
						String rootkey,
						char separator,
						TreeOutputType type,
						String *output)
{
	return(traverse(strtree, rootkey, separator, type, output, tinorder));
}

/* visit nodes in pre-order; see description of traverse function */
OSErr	str_tpreorder(StrTreeType strtree,
						String rootkey,
						char separator,
						TreeOutputType type,
						String *output)
{
	return(traverse(strtree, rootkey, separator, type, output, tpreorder));
}

/* visit nodes in post-order; see description of traverse function */
OSErr str_tpostorder(StrTreeType strtree,
						String rootkey,
						char separator,
						TreeOutputType type,
						String *output)
{
	return(traverse(strtree, rootkey, separator, type, output, tpostorder));
}

/* visit nodes in level-order; see description of traverse function */
OSErr str_tlevelorder(StrTreeType strtree,
						String rootkey,
						char separator,
						TreeOutputType type,
						String *output)
{
	return(traverse(strtree, rootkey, separator, type, output, tlevelorder));
}

/* traverse the tree, and concatenate nodes encountered
	to a string, separating each node with the separator character
	given (if no separator is given then a default separator is used).
	The choice of data to output depends on the value of the output
	type. Traversal starts from the rootkey given
	(if root is NULL traversal starts from the root of the tree).
	The tree is traversed using the function in the func parameter.
*/
static OSErr traverse(StrTreeType strtree,
						String rootkey,
						char separator,
						TreeOutputType type,
						String *output,
						FPTR func)
{
	FPTR action;			/* action function to be called at each node */
	TreeNode	rootnode;	/* node to start from */
	StrTreeBufType stbuf;/* output buffer */
	OSErr	err = noErr;
	
	assert(str_tvalid(strtree));

	*output = NULL;
	
	/* initialize output buffer */
	err = str_tinitbuf(&stbuf, separator);
	if (err) return(err);

	/* find node containing root key */
	if (rootkey)
		rootnode = tfind(gettree(strtree), (TreeKey) rootkey, NULL);
	else
		rootnode = troot(gettree(strtree));
	if (rootnode) {
		/* traverse tree and use requested output function */
		switch (type) {
		case TREE_OUTPUT_KEYS:	action = outputkey; break;
		case TREE_OUTPUT_DATA:	action = outputdatum; break;
		case TREE_OUTPUT_INDEXES:action = outputindex; break;
		default: assert(false);
		}
		err = func(gettree(strtree), rootnode, action, (void *) &stbuf);
	}
	
	/* end use of buffer, placing its contents into the output pointer */
	if (! err)
		err = str_tendbuf(&stbuf, output);
	else
		(void) str_tendbuf(&stbuf, output);
	
	return(err);
}

/*----------------------------------------------------------------*/
/* range searching */
/*----------------------------------------------------------------*/

/* do a one dimensional range search for keys between high and low.
	Each node encountered is concatenated to a string,
	and each node is separated with the separator character given (if no
	separator is given then a default separator is used).
	The data output from each node depend on the value of the type
	parameter.
	The search starts from the root key given, unless root is
	NULL, in which case search starts from the root of the tree.
*/
OSErr str_trange(StrTreeType strtree,
						String rootkey,
						String low,
						String high,
						char separator,
						TreeOutputType type,
						String *output)
{
	FPTR action;			/* action function to be called at each node */
	TreeNode	rootnode;	/* node to start from */
	StrTreeBufType stbuf;/* output buffer */
	OSErr err = noErr;
	
	assert(str_tvalid(strtree));
	
	*output = NULL;

	/* initialize output buffer */
	err = str_tinitbuf(&stbuf, separator);
	if (err) return(err);

	/* find node containing root key */
	if (rootkey)
		rootnode = tfind(gettree(strtree), (TreeKey) rootkey, NULL);
	else
		rootnode = troot(gettree(strtree));
	
	if (rootnode) {
		/* traverse tree and use requested output function */
		switch (type) {
		case TREE_OUTPUT_KEYS:	action = outputkey; break;
		case TREE_OUTPUT_DATA:	action = outputdatum; break;
		case TREE_OUTPUT_INDEXES:action = outputindex; break;
		default: assert(false);
		}
		err = trange(gettree(strtree), rootnode, low, high, action, (void*)&stbuf);
	}
	
	/* end use of buffer, placing its contents into the output pointer */
	if (! err)
		err = str_tendbuf(&stbuf, output);
	else
		(void) str_tendbuf(&stbuf, output);
	
	return(err);
}

/*----------------------------------------------------------------*/
/* action functions called from tree traversal functions */
/*
	These functions are passed to tree library functions 
	tinorder, tpreorder, tpostorder, and trange  which then
	call these functions for each node encountered. The choice
	of the action function depends on whether you want
	to get the nodes' keys, data, or indexes. The data parameter
	to the action functions must contain a pointert to the
	output buffer.
*/
/*----------------------------------------------------------------*/

static int outputdatum(TREE_ACTION_PROTOTYPE)
{
	StrTreeBufType *stbuf;
		
	/* recover the string tree output buffer */
	stbuf = (StrTreeBufType *) data;
	
	/* append the datum to the output buffer */
	return(str_toutputstr(stbuf, (String) tdatum(node)));
}

static int outputkey(TREE_ACTION_PROTOTYPE)
{
	StrTreeBufType *stbuf;
		
	/* recover the string tree output buffer */
	stbuf = (StrTreeBufType *) data;
	
	/* append the key to the output buffer */
	return(str_toutputstr(stbuf, (String) tkey(node)));
}

static int outputindex(TREE_ACTION_PROTOTYPE)
{
	StrTreeBufType *stbuf;
	CStr31 str;
	
	/* recover the string tree output buffer */
	stbuf = (StrTreeBufType *) data;
		
	/* append the index to the output buffer */
	return(str_toutputptr(stbuf, ltoa(index, str)));
}

