/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

#include <MyHeaders>
#include "str_tprivate.h"

/*
	This is the "main" file of the string tree library. The header
	has more details about this library.
*/

#if ASSERTIONS
/* make sure the string tree data structure contains valid data */
Boolean str_tvalid(StrTreeType strtree)
{
	if (! strtree) return(false);
	assert(validhandle((Handle) strtree));
	switch (getcmptype(strtree)) {
	case STR_COMPARE_EXACT:
		assert(tgetcmp(gettree(strtree)) == comparekeys);
		assert(tgetorder(gettree(strtree)) == comparekeys);
		break;
	case STR_COMPARE_IGNORECASE:
		assert(tgetcmp(gettree(strtree)) == compareikeys);
		assert(tgetorder(gettree(strtree)) == orderikeys);
		break;
	case STR_COMPARE_INTERNATIONAL:
		assert(tgetcmp(gettree(strtree)) == IUcomparekeys);
		assert(tgetorder(gettree(strtree)) == IUorderkeys);
		break;
	case STR_COMPARE_FLOAT:
		assert(tgetcmp(gettree(strtree)) == numcmpkeys);
		assert(tgetorder(gettree(strtree)) == numorderkeys);
		break;
	default:
		assert(false);
	}
	assert(tgetdisposekey(gettree(strtree)) == disposekey);
	assert(tgetdisposedatum(gettree(strtree)) == disposedatum);
	return(true);
}

/* make sure the key contains valid data */
Boolean str_tvalidkey(String key)
{
	if (! key) return(false);
	assert(validhandle((Handle) key));
	return(true);
}
#endif /* ASSERTIONS */

/*----------------------------------------------------------------*/
/* Tree creating and disposing */
/*----------------------------------------------------------------*/

/* create a new empty strtree */
OSErr	str_tnew(StrTreeType *strtreePtr, StrCompareType cmpr)
{
	StrTreeType	strtree;
	TreeType		tree;
	OSErr			err=noErr;
	
	*strtreePtr = NULL;
	strtree = (StrTreeType) halloc(sizeof(StrTreeStruct));
	if ((err=MemError()) == noErr) {
		tree = tnew(comparekeys, comparekeys);
		if ((err=MemError()) == noErr) {
			uniq(strtree) = false;
			getcmptype(strtree) = cmpr;
			gettree(strtree) = tree;
			str_tupdate(strtree);
			*strtreePtr = strtree;
			return(noErr);
		}
		hfree(strtree);
	}
	return(err);
}

/* dispose of the tree */
OSErr str_tdispose(StrTreeType strtree)
{
	OSErr err = noErr;
	
	assert(str_tvalid(strtree));
	err = tdispose(gettree(strtree));
	if (!err)
		hfree(strtree);
	return(err);
}

/* update any volatile fields of the strtree, such as function pointers,
	which can be loaded into different locations in memory.
*/
void str_tupdate(StrTreeType strtree)
{
	/* set key comparison functions */
	switch (getcmptype(strtree)) {
	case STR_COMPARE_EXACT:
		tsetcmp(gettree(strtree), comparekeys);
		tsetorder(gettree(strtree), comparekeys);
		break;
	case STR_COMPARE_IGNORECASE:
		tsetcmp(gettree(strtree), compareikeys);
		tsetorder(gettree(strtree), orderikeys);
		break;
	case STR_COMPARE_INTERNATIONAL:
		tsetcmp(gettree(strtree), IUcomparekeys);
		tsetorder(gettree(strtree), IUorderkeys);
		break;
	case STR_COMPARE_FLOAT:
		tsetcmp(gettree(strtree), numcmpkeys);
		tsetorder(gettree(strtree), numorderkeys);
		break;
	default:
		assert(false);
	}
	
	/* set dispose functions */
	tsetdisposekey(gettree(strtree), disposekey);
	tsetdisposedatum(gettree(strtree), disposedatum);
	
	assert(str_tvalid(strtree));
}

/*----------------------------------------------------------------*/
/* Inserting and deleting */
/*----------------------------------------------------------------*/

/* insert a copy of the the key and datum into the tree, unless
	the copy flag is false, in which case the given handles are inserted.
	if the tree's unique attribute is true then the key is inserted only if
	it doesn't already exist.
*/
OSErr	str_tinsert(StrTreeType strtree, String key, String datum, Boolean copy)
{
	OSErr	err = noErr;
	TreeNode node;		/* the new node */
	
	assert(str_tvalid(strtree));
	assert(str_tvalidkey(key));

	if (**key == 0)
		return(ERR_STRTREE_EMPTYKEY);
	
	/* if unique attribute, then first search for key */
	if (uniq(strtree) && tfind(gettree(strtree), key, NULL))
		return(ERR_STRTREE_EXISTS);
	
	/* make a copy of the key and then insert it into the tree */
	if (copy) {
		key = strdup(key);
		if (! key)
			return(MemError());
		if (datum) {
			datum = strdup(datum);
			if (! datum) {
				strfree(key);
				return(MemError());
			}
		}
	}
	
	/* insert key and set datum */
	node = tinsert(gettree(strtree), (TreeKey) key);
	if (node)
		tsetdatum(gettree(strtree), node, datum);
	else {
		if (copy) {
			if (key)
				strfree(key);
			if (datum)
				strfree(datum);
		}
		err = MemError();
	}
	return(err);
}

/* delete the node with the given key from the tree */
OSErr	str_tdelete(StrTreeType strtree, String key)
{
	assert(str_tvalid(strtree));
	assert(str_tvalidkey(key));

	tdelete(gettree(strtree), (TreeKey) key);
	return(noErr);
}

/*----------------------------------------------------------------*/
/* setting and getting datum fields */
/*----------------------------------------------------------------*/

/* set the datum field of the node with the specified key to a copy
	of the given datum (if the copy flag is true, otherwise just set
	the datum without copying it).
*/
OSErr	str_tset(StrTreeType strtree, String key, String datum, Boolean copy)
{
	TreeNode	node;
	OSErr	err = noErr;
	
	assert(str_tvalid(strtree));
	assert(str_tvalidkey(key));
	
	node = tfind(gettree(strtree), (TreeKey) key, NULL);
	if (node) {
		if (datum && copy) {
			HandToHand(&datum);
			if ((err=MemError()) != noErr)
				return(err);
		}
		/* set the datum (disposing of old datum is handled automatically) */
		tsetdatum(gettree(strtree), node, (TreeDatum) datum);
	}
	else
		err = ERR_STRTREE_NOTFOUND;
	return(err);
}

/* return the datum of the node with the specified key; if copy
	is true then a copy of the datum is returned, instead of
	the actual datum.
*/
OSErr	str_tget(StrTreeType strtree, String key, String *datumptr,
					Boolean copy)
{
	TreeNode	node;
	OSErr	err = noErr;
	
	assert(str_tvalid(strtree));
	assert(str_tvalidkey(key));

	*datumptr = NULL;
	node = tfind(gettree(strtree), key, NULL);
	if (node) {
		*datumptr = tdatum(node);
		if (*datumptr && copy) {
			HandToHand(datumptr);
			err = MemError();
		}
	}
	else
		err = ERR_STRTREE_NOTFOUND;
	return(err);
}

/*----------------------------------------------------------------*/
/* getting height of a tree */
/*----------------------------------------------------------------*/

/* Return height of tree rooted at the node with the given key,
	or if rootkey is NULL then the height of the entire tree. */
TreeIndex str_theight(StrTreeType strtree, String rootkey)
{
	TreeNode rootnode;
	
	/* find node containing root key */
	if (rootkey) rootnode = tfind(gettree(strtree), (TreeKey) rootkey, NULL);
	else rootnode = troot(gettree(strtree));
	/* get height of tree at rootnode */
	if (rootnode)
		return(theight(rootnode, 0));
	return(0);
}

/*----------------------------------------------------------------*/
/* setting and getting strtree's attributes */
/*----------------------------------------------------------------*/

/* set the value of the given attribute */
OSErr	str_tsetattr(StrTreeType strtree, TreeAttrType attr, int value)
{
	assert(str_tvalid(strtree));

	switch (attr) {
	case TREE_ATTR_UNIQUE:
		uniq(strtree) = value;
		break;
	default:
		return(ERR_STRTREE_ATTR);
	}
	str_tupdate(strtree);
	return(noErr);
}

/* get the value of the given attribute */
OSErr	str_tgetattr(StrTreeType strtree, TreeAttrType attr, int *valueptr)
{
	assert(str_tvalid(strtree));

	switch (attr) {
	case TREE_ATTR_COMPARE:
		*valueptr = getcmptype(strtree);
		break;
	case TREE_ATTR_UNIQUE:
		*valueptr = uniq(strtree);
		break;
	default:
		return(ERR_STRTREE_ATTR);
	}
	return(noErr);
}
	