/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
An "string tree" is a binary tree whose keys and datum fields contain
character strings. Operations on the string tree are designed to make
building trees from strings (and extracting data from the tree) simple
and fast. This library was designed for use by a HyperCard XFCN,
but could be used for other purposes as well.

A string tree uses the more generic binary tree library, which implements
all the basic operations on trees. An string tree extends the features
provided by the binary tree library by implementing the specific
"string" data type for the tree's nodes. Since strings are used
as the tree's keys, there are several attributes associated with
a strtree which modify the comparison rules for insertion into the
tree; these are: exact, ignore case, international, and numeric.
In addition, a unique attribute prevents the insertion of multiple
copies of a key into the tree.

A string tree installs several functions into the binary tree it uses.
These are the dispose node and dispose key functions, in addition
to the required compare key and order key functions. The data field

Finally, a string tree uses the buffer library while constructing large
output strings from the data in the binary tree. For instance,
the tree traversal functions must append the key or datum from
each node to a dynamically growing output string, exactly
the purpose for which the buffer library was written.

Remember that the addresses of function pointers may change on 
the Macintosh due to unloading and subsequent uploading of code
resources. A string tree installs function pointers into
the binary tree, which must be updated _before_ the string tree is
used. This is done by calling the function str_tupdate. Failure
to do so may cause some of the strangest bugs.

Brief descriptions of the functions in this library follow.

Creating and Disposing
str_tnew			Creates a new string tree
str_tdispose	Dispose of a string tree
str_tupdate		Updates the function pointer fields. Should be called
					before a string tree is used, since addresses of
					function pointers may change during execution of a
					program on the Macintosh (ya, the Mac is totally brain dead).

Inserting and Deleting Nodes
str_tinsert		Insert a key into the tree; if unique attribute is
					true then checks if key already exists before inserting.
str_tdelete		Delete the node with the given key from the tree.
str_tinsert_many	Inserts many keys into the tree. Each key and datum is
						delimited by a delimiter character.
str_tdelete_many	Deletes many keys; similar to str_tinsert_many.

Getting and Setting Data
str_tset			Sets the datum field of the node with the given key.
str_tget			Returns a copy of the datum field of the node with the
					given key.
str_tset_many	Sets many datum fields; similar to str_tinsert_many.
str_tget_many	Sets many datum fields; similar to str_tinsert_many.
str_trange		Does a one-dimensional range search.

Accessing Nodes
str_tleft		Gets key of left child
str_tright		Gets key of right child
str_parent		Gets parent
str_troot		Gets root of tree

Traversing Tree
str_tinorder	Traverses tree in in-order, and returns either a list
					of the keys encountered or a list of the datum fields
					encountered; each output item is separated by a separator
					character.
str_tpreorder	Similar to str_tinorder
str_tpostorder	Similar to str_tinorder
str_tlevelorder Similar to str_tinorder

Getting Height of Tree
str_theight		Returns height of tree, starting from the node containing
					the specified key, or from the root if no key is given.
					
Getting and Setting Attributes
str_tgetattr	Gets the value of an attribute
str_tsetattr	Sets the value of an attribute
*/
	
#ifndef _H_strtreelib
#define _H_strtreelib

#ifndef _H_nodeutil
#include "nodeutil.h"
#endif

#ifndef _H_buflib
#include "buflib.h"
#endif

#ifndef _H_hyperlib
#include "hyperlib.h"	/* for error codes */
#endif

/* data types */
typedef struct {
	TreeType	tree;			/* the binary tree */
	
	/* the unique flag is checked when keys are inserted into the tree */
	Boolean	unique;		/* only unique keys are allowed */
	
	/* the cmptype flag is used to set the correct key comparison
		functions by str_tupdate */
	StrCompareType	cmptype;		/* string comparison method */

} StrTreeStruct, **StrTreeType;

/* tree attributes */
typedef enum {
	TREE_ATTR_NULL,
	TREE_ATTR_COMPARE,
	TREE_ATTR_UNIQUE,
	TREE_ATTR_LAST
} TreeAttrType;

/* data types to output from tree traversal functions */
typedef enum {
	TREE_OUTPUT_NULL,
	TREE_OUTPUT_KEYS,
	TREE_OUTPUT_DATA,
	TREE_OUTPUT_INDEXES,
	TREE_OUTPUT_LAST
} TreeOutputType;

/* strtree functions */

/* tree creating and disposing */
OSErr	str_tnew(StrTreeType *strtreeptr, StrCompareType cmpr);
OSErr	str_tdispose(StrTreeType strtree);
void	str_tupdate(StrTreeType strtree);

/* inserting and deleting */
OSErr	str_tinsert(StrTreeType tree, String key, String datum, Boolean copy);
OSErr	str_tdelete(StrTreeType tree, String key);

OSErr str_tinsert_many(StrTreeType tree, String keys, String data,
								char key_separator, char datum_separator);
OSErr str_tdelete_many(StrTreeType tree, String keys, char key_separator);

/* setting and getting nodes' datum fields */
OSErr	str_tset(StrTreeType tree, String key, String datum, Boolean copy);
OSErr	str_tget(StrTreeType tree, String key, String *datumptr, Boolean copy);

OSErr	str_tset_many(StrTreeType strtree, String keys, String data,
						char key_separator, char datum_separator);
OSErr	str_tget_many(StrTreeType strtree, String keys,
						char key_separator, char datum_separator,
						String *output);

/* range searching */
OSErr str_trange(StrTreeType strtree,
						String rootkey, String low, String high,
						char separator, TreeOutputType type, String *output);

/* accessing nodes */
OSErr	str_tleft(StrTreeType tree, String key, String *leftkey);
OSErr	str_tright(StrTreeType tree, String key, String *rightkey);
OSErr	str_tparent(StrTreeType tree, String key, String *parentkey);
OSErr	str_troot(StrTreeType tree, String *rootkey);

/* tree traversal */
OSErr	str_tinorder(StrTreeType tree, String rootkey, char separator,
					TreeOutputType type, String *output);
OSErr	str_tpreorder(StrTreeType tree, String rootkey, char separator,
					TreeOutputType type, String *output);
OSErr	str_tpostorder(StrTreeType tree, String rootkey, char separator,
					TreeOutputType type, String *output);
OSErr	str_tlevelorder(StrTreeType tree, String rootkey, char separator,
					TreeOutputType type, String *output);
					
/* getting height of tree */
TreeIndex str_theight(StrTreeType strtree, String rootkey);

/* setting and getting strtree's attributes */
OSErr	str_tsetattr(StrTreeType tree, TreeAttrType attr, int value);
OSErr	str_tgetattr(StrTreeType tree, TreeAttrType attr, int *valueptr);

#endif