/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

#include <MyHeaders>
#include "treelib_private.h"

static int	disposekey(TreeType tree, TreeKey key);
static int	disposedatum(TreeType tree, TreeDatum datum);

static TreeNode newnode(TreeKey key);
static int		 disposenode(TreeType tree, TreeNode node);
static TreeNode find(TreeType tree,
							TreeKey key,
							FPTR cmpfunc,
							Boolean first,
							TreeNode *parentptr,
							int *cmpptr);

/*----------------------------------------------------------------*/
/* Functions for checking validity of various things */
/*----------------------------------------------------------------*/

#if ASSERTIONS
Boolean	validtree(TreeType tree)
{
	if (! tree) return(false);
	assert(validhandle((Handle) tree));
	assert(ordr(tree) != NULL);
	assert(cmpr(tree) != NULL);
	if (root(tree)) assert(validnode(root(tree)));
	return(true);
}

Boolean	validnode(TreeNode node)
{
	if (node == NULL) return(false);
	assert(validhandle((Handle) node));
	return(true);
}

Boolean validkey(TreeKey key)
{
	return(key != NULL);
}
#endif

/*----------------------------------------------------------------*/
/* Functions for creating and disposing of tree */
/*----------------------------------------------------------------*/

/* create a new empty node */
static TreeNode newnode(TreeKey key)
{
	TreeNode	node;
	
	assert(validkey(key));
	node = (TreeNode) halloc(sizeof(TreeNodeStruct));
	if (node) {
		lchild(node) = rchild(node) = NULL;
		getkey(node) = key;
		getdatum(node) = NULL;
	}
	return(node);
}

/* dispose of the node */
static int disposenode(TreeType tree, TreeNode node)
{
	if (tgetdisposekey(tree))
		disposekey(tree, tkey(node));
	if (tgetdisposedatum(tree))
		disposedatum(tree, tdatum(node));
	hfree(node);
	return(0); /* for benefit of tree traversal functions */
}

TreeType tnew(FPTR order, FPTR compare)
{
	TreeType	tree;
	
	assert(order != NULL);
	assert(compare != NULL);
	tree = (TreeType) halloc(sizeof(TreeStruct));
	if (tree) {
		root(tree) = NULL;
		ordr(tree) = order;
		cmpr(tree) = compare;
		(**tree).key_dispose = NULL;
		(**tree).datum_dispose = NULL;
		(**tree).delflag = false;
		assert(validtree(tree));
	}
	return(tree);
}

/*----------------------------------------------------------------*/
/* Disposing of tree */
/*----------------------------------------------------------------*/

static int tdispose_action(TREE_ACTION_PROTOTYPE);
										
/* dispose of the whole tree (returns value of tpostorder) */
int tdispose(TreeType tree)
{
	int val;
	
	assert(validtree(tree));
	val = tpostorder(tree, root(tree), tdispose_action, NULL);
	if (val == 0) hfree(tree);
	return(val);
}

static int tdispose_action(TREE_ACTION_PROTOTYPE)
{
	return(disposenode(tree, node));
}

/*----------------------------------------------------------------*/
/* Finding an item in the tree */
/*----------------------------------------------------------------*/

/* find key in tree and return node, parent, and last compare;
	if first is true then function stops at first match, otherwise
	it continues until it reaches a leaf.
*/
static TreeNode find(TreeType tree,
							TreeKey key,			/* key to find */
							FPTR cmpfunc,			/* function for comparing */
							Boolean first,			/* stop at first match */
							TreeNode *parentptr, /* parent node */
							int *cmpptr)			/* result of last comparison */
{
	TreeNode node, parent;
	int cmp, parentcmp;
	
	assert(validtree(tree));
	assert(validkey(key));
	parentcmp = 0;
	parent = node = root(tree);
	while (node) {
		cmp = (*cmpfunc)(key, getkey(node));
		if (cmp < 0 ) {
			parent = node;
			parentcmp = cmp;
			node = lchild(node);
		}
		else if (cmp > 0 || ! first) {
			parent = node;
			parentcmp = cmp;
			node = rchild(node);
		}
		else
			break;
	}
	if (parentptr) *parentptr = parent;
	if (cmpptr) *cmpptr = parentcmp;
	return(node);
}

/* find key in tree, return node and parent if found */
TreeNode tfind(TreeType tree, TreeKey key, TreeNode *parent)
{
	assert(validtree(tree));
	assert(validkey(key));
	return(find(tree, key, cmpr(tree), true, parent, NULL));
}

/* insert key into tree and return new node; if an error occurs then
	returns NULL.
*/
TreeNode tinsert(TreeType tree, TreeKey key)
{
	TreeNode	node, parent;
	int cmp;
	
	assert(validtree(tree));
	assert(validkey(key));
	node = find(tree, key, ordr(tree), false, &parent, &cmp);
	assert(! node);
	node = newnode(key);
	if (! node) return(node);
	if (! root(tree))
		root(tree) = node;
	else if (cmp < 0)
		lchild(parent) = node;
	else
		rchild(parent) = node;
	return(node);
}

/* delete key from tree */
void tdelete(TreeType tree, TreeKey key)
{
	TreeNode	parent, del, node;

	del = find(tree, key, cmpr(tree), true, &parent, NULL);
	if (! del)
		return;
	if (! rchild(del))
		node = lchild(del);
	else if (! lchild(del))
		node = rchild(del);
	else if ((**tree).delflag) {
		for (node = lchild(del); rchild(node); node = rchild(node))
			;
		rchild(node) = rchild(del);
		node = lchild(del);
	}
	else {
		for (node = rchild(del); lchild(node); node = lchild(node))
			;
		lchild(node) = lchild(del);
		node = rchild(del);
	}
	if (parent == del)
		root(tree) = node;
	else if (lchild(parent) == del)
		lchild(parent) = node;
	else if (rchild(parent) == del)
		rchild(parent) = node;
	else
		assert(false);
	(void) disposenode(tree, del);
	(**tree).delflag = ! (**tree).delflag;
}

/* dispose of old contents of datum and then set to new contents */
void tsetdatum(TreeType tree, TreeNode node, TreeDatum datum)
{
	assert(validtree(tree));
	assert(validnode(node));
	if (tgetdisposedatum(tree))
		(void) disposedatum(tree, getdatum(node));
	getdatum(node) = datum;
}

/*----------------------------------------------------------------*/
/* Disposing of the key and the datum */
/*----------------------------------------------------------------*/

/* dispose of the given key */
static int disposekey(TreeType tree, TreeKey key)
{
	assert(tgetdisposekey(tree) != NULL);
	return((*tgetdisposekey(tree))(key));
}

/* dispose of the given datum */
static int disposedatum(TreeType tree, TreeDatum datum)
{
	assert(tgetdisposedatum(tree) != NULL);
	return((*tgetdisposedatum(tree))(datum));
}

/*----------------------------------------------------------------*/
/* Getting height of tree */
/*----------------------------------------------------------------*/

static TreeIndex max(TreeIndex a, TreeIndex b);

static TreeIndex max(TreeIndex a, TreeIndex b)
{
	return(a > b ? a : b);
}

TreeIndex theight(TreeNode node, TreeIndex height)
{
	assert(validnode(node));
	for (;;) {
		if (lchild(node) && rchild(node))
			return(max(theight(lchild(node), height+1),
							theight(rchild(node), height+1)));
		else if (lchild(node)) {
			node = lchild(node);
			height++;
		}
		else if (rchild(node)) {
			node = rchild(node);
			height++;
		}
		else
			return(height);
	}
}
							
/*----------------------------------------------------------------*/
/*
	All of the following functions were included for use only during
	debugging of code that uses this library. Equivalent (but
	faster) macro's are provided.
*/
/*----------------------------------------------------------------*/

#if ASSERTIONS

/*----------------------------------------------------------------*/
/* Comparing items */
/*----------------------------------------------------------------*/

int tcmp(TreeType tree, TreeKey key1, TreeKey key2)
{
	assert(validtree(tree));
	assert(validkey(key1));
	assert(validkey(key2));
	assert(tgetcmp(tree) != NULL);
	return((*tgetcmp(tree))(key1, key2));
}
	
int torder(TreeType tree, TreeKey key1, TreeKey key2)
{
	assert(validtree(tree));
	assert(validkey(key1));
	assert(validkey(key2));
	assert(tgetorder(tree) != NULL);
	return((*tgetorder(tree))(key1, key2));
}
	
/*----------------------------------------------------------------*/
/* Getting items in node and tree */
/*----------------------------------------------------------------*/

TreeNode troot(TreeType tree)
{
	return(root(tree));
}

TreeNode tleft(TreeNode node)
{
	return(lchild(node));
}

TreeNode tright(TreeNode node)
{
	assert(validnode(node));
	return(rchild(node));
}

TreeKey tkey(TreeNode node)
{
	assert(validnode(node));
	return(getkey(node));
}

TreeDatum tdatum(TreeNode node)
{
	assert(validnode(node));
	return(getdatum(node));
}

/*----------------------------------------------------------------*/
/* Getting the tree's function pointers */
/*----------------------------------------------------------------*/

FPTR tgetorder(TreeType tree)
{
	assert(validtree(tree));
	return(ordr(tree));
}

FPTR tgetcmp(TreeType tree)
{
	assert(validtree(tree));
	return(cmpr(tree));
}

FPTR tgetdisposekey(TreeType tree)
{
	assert(validtree(tree));
	return((**tree).key_dispose);
}

FPTR tgetdisposedatum(TreeType tree)
{
	assert(validtree(tree));
	return((**tree).datum_dispose);
}

/*----------------------------------------------------------------*/
/* Setting tree's function pointers */
/*----------------------------------------------------------------*/

void tsetorder(TreeType tree, FPTR order)
{
	assert(validtree(tree));
	assert(order != NULL);
	ordr(tree) = order;
}

void tsetcmp(TreeType tree, FPTR compare)
{
	assert(validtree(tree));
	assert(compare != NULL);
	cmpr(tree) = compare;
}

void tsetdisposekey(TreeType tree, FPTR key_dispose)
{
	assert(validtree(tree));
	(**tree).key_dispose = key_dispose;
}

void tsetdisposedatum(TreeType tree, FPTR datum_dispose)
{
	assert(validtree(tree));
	(**tree).datum_dispose = datum_dispose;
}

void *tdata(TreeType tree)
{
	assert(validtree(tree));
	return((**tree).data);
}

void tsetdata(TreeType tree, void *data)
{
	assert(validtree(tree));
	(**tree).data = data;
}

#endif /* ASSERTIONS */