/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

#include <MyHeaders>
#include "treelib_private.h"

/*----------------------------------------------------------------*/
/* Tree traversal */
/*----------------------------------------------------------------*/

/*
	These tree traversal functions recursively traverse the tree,
	and execute the specified action function at each node visited.
	If the action function returns a non-zero value then the
	traversal functions stop and return the same value. This is
	useful if the action function detects an error.
	
	Whatever data and parameters don't change between recursive
	calls are placed into a TraverseData structure, which is on the
	stack, so that these traversal functions may be called recursively
	from within the action function (if the data were stored in global
	variables this would not be possible). Then, the actual recursive
	function is called.
*/

/* constant data for traversal */
typedef struct {
	TreeType	tree;
	FPTR		action;
	void		*data;
	int		stop;
} TraverseData;

static int	execute(VFPTR traverse, TreeType tree, TreeNode node,
						FPTR action, void *data);
static void inorder(TraverseData *td, TreeNode node, TreeIndex index);
static void preorder(TraverseData *td, TreeNode node, TreeIndex index);
static void postorder(TraverseData *td, TreeNode node, TreeIndex index);
static void levelorder(TraverseData *td, TreeNode node, TreeIndex index);

/*--------------------------------------------------------------------------*/
/* These are the external functions */
/*--------------------------------------------------------------------------*/

int tinorder(TreeType tree, TreeNode node, FPTR action, void *data)
{
	return(execute(inorder, tree, node, action, data));
}

int tpreorder(TreeType tree, TreeNode node, FPTR action, void *data)
{
	return(execute(preorder, tree, node, action, data));
}

int tpostorder(TreeType tree, TreeNode node, FPTR action, void *data)
{
	return(execute(postorder, tree, node, action, data));
}

int tlevelorder(TreeType tree, TreeNode node, FPTR action, void *data)
{
	return(execute(levelorder, tree, node, action, data));
}

/*--------------------------------------------------------------------------*/
/* These are the private functions */
/*--------------------------------------------------------------------------*/

static int execute(VFPTR traverse, TreeType tree, TreeNode node,
						FPTR action, void *data)
{
	TraverseData td;
	assert(validtree(tree));
	assert(action != NULL);
	td.tree = tree;
	td.action = action;
	td.data = data;
	td.stop = 0;
	if (node)
		traverse(&td, node, 1L);
	return(td.stop);
}

static void inorder(TraverseData *td, TreeNode node, TreeIndex index)
{
	while (node) {
		if (lchild(node)) {
			inorder(td, lchild(node), lindex(index));
			if (td->stop) break;
		}
		td->stop = td->action(td->tree, node, index, td->data);
		if (td->stop) break;
		node = rchild(node);
		index = rindex(index);
	}
}

static void preorder(TraverseData *td, TreeNode node, TreeIndex index)
{
	while (node) {
		td->stop = td->action(td->tree, node, index, td->data);
		if (td->stop) break;
		if (lchild(node)) {
			preorder(td, lchild(node), lindex(index));
			if (td->stop) break;
		}
		node = rchild(node);
		index = rindex(index);
	}
}

static void postorder(TraverseData *td, TreeNode node, TreeIndex index)
{
	if (lchild(node)) {
		postorder(td, lchild(node), lindex(index));
		if (td->stop) return;
	}
	if (rchild(node)) {
		postorder(td, rchild(node), rindex(index));
		if (td->stop) return;
	}
	td->stop = td->action(td->tree, node, index, td->data);
}

/*--------------------------------------------------------------------------*/
/*
	Determining size of queue:
	The level order traversal function places a maximum of
	two items into the queue for each node that is not a leaf,
	but one of these items is removed almost immediately, so that
	the queue may grow by a maximum of one for each node encountered
	(though space for the one extra item that is placed into the
	queue and then immediately removed must be available). The number
	of internal nodes (i.e., nodes that are not leaves) in a tree
	is given by
		2^height-1
	Thus, the maximum number of items in the queue at one time is
		= internal nodes + space for one extra item
		= 2^height-1 + 1
		= 2^height
	The queue, which is stored as a circular array, may contain
	a number of items equal to one less than the size of the array,
	or (Qmax-1) items. Thus, the maximum height tree that may
	be traversed for a given queue is
		height = floor(lg(Qmax-1))
	and the size of a queue for some height is
		Qmax = 2^height+1
*/
/*--------------------------------------------------------------------------*/

#define QUEUEMAX	(1025)

/* this is just a simple queue of nodes and node indexes */
typedef struct {
	size_t head;
	size_t tail;
	size_t max;
	struct {
		TreeNode node;
		TreeIndex index;
	} array[QUEUEMAX];
} QueueType;

/* the queue functions */
static int	qinit(QueueType *q);
static void	qend(QueueType *q);
static int	qput(QueueType *q, TreeNode node, TreeIndex index);
static void	qget(QueueType *q, TreeNode *node, TreeIndex *index);

#define qempty(q)	((q)->head == (q)->tail)
#define qfull(q)	((((q)->tail+1) % (q)->max) == (q)->head)

/* traverse tree in level-order */
static void levelorder(TraverseData *td, TreeNode node, TreeIndex index)
{
	QueueType queue, *q = &queue;
	
	td->stop = qinit(q);
	if (td->stop) return;

	td->stop = qput(q, node, index);
	if (td->stop) return;
	
	while (! qempty(q)) {
		qget(q, &node, &index);
		td->stop = td->action(td->tree, node, index, td->data);
		if (td->stop) break;
		if (lchild(node))
			if ((td->stop = qput(q, lchild(node), lindex(index))) != 0)
				break;
		if (rchild(node))
			if ((td->stop = qput(q, rchild(node), rindex(index))) != 0)
				break;
	}
	qend(q);
}

/*--------------------------------------------------------------------------*/
/* Queue functions for level order traversal */
/*--------------------------------------------------------------------------*/

static int qinit(QueueType *q)
{
	q->head = q->tail = 0;
	q->max = QUEUEMAX;
	return(0);
}

static void qend(QueueType *q)
{
	/* may do something in the future */
}

static int qput(QueueType *q, TreeNode node, TreeIndex index)
{
	if (qfull(q)) return(memFullErr);
	q->array[q->tail].node = node;
	q->array[q->tail].index = index;
	q->tail = (q->tail+1) % q->max;
	return(0);
}

static void qget(QueueType *q, TreeNode *node, TreeIndex *index)
{
	*node = q->array[q->head].node;
	*index = q->array[q->head].index;
	q->head = (q->head+1) % q->max;
}
