/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
A general purpose implementation of a standard binary tree data structure.
Nodes in the tree must contain a key, and may contain an optional
datum field. The key and datum field are both void pointers, so
they can be set to anything desired. Each tree has associated with
it several extra pieces of information. These are comparison, order,
dispose key, and dispose datum function pointers, as well as
a special data field which is attached to the whole tree and
may be used for any purpose desired by the application. Each of
these items is discussed below.

This implementation is approriate for the Macintosh since it
uses handles to its nodes and to the tree itself.

_________________________________
The Node

Each node contains a key field, a datum field, and a left
and right child pointer.

 The Key Field

Each node must have a non-null pointer to the node's key.
The key field is used for determining the relative value of a node,
and is therefore the field used while searching the tree.

 The Datum Field

This field is optional, and may contain any extra data associated
with a node.

_________________________________
The Tree

Each tree contains a pointer to the root node of the tree, as well
as several function pointers. Nodes in the tree are compared
using two functions: the compare and order functions.

Having two different functions for ordering keys is useful
for trees that must ignore case of keys but must still place
upper case letters before lower case letters. For simpler trees the
comparison and ordering operations may use the same function.
In general, the ordering rules should be either as permissive as the
comparison rules or more restrictive, but never more lenient.

As an example of differing order and comparison rules, suppose a tree
contains text that is to be sorted alphabetically. The ordering rules
might specify that upper case letters come before lower case versions
of the same letter, while the comparison rules could specify that
upper and lower case letters are to be considered equivalent. This is
an example where the ordering rules are more restrictive than the
comparison rules, that is, fewer instances of similar items are
considered equal. It would be an error to reverse the order and
comparison rules in this example.

 The Compare Function Pointer

This field must contain a pointer to a function used for comparing
two keys when a key is being search for in the tree.
The function should be declared as follows:
	int compare(TreeKey key1, TreeKey key2);
The key1 and key2 parameters are the keys to compare. The function
should return a value less than zero, equal to zero, or greater
than zero, according respectively, to whether key1 is smaller than,
equal to, or greater than key2.

 The Order Function Pointer

This field must contain a pointer to a function used for comparing
two keys while they are being inserted into the tree.
The function should be declared and should return the same types
as values as the compare function described above.

 The Dispose Key Function Pointer

This field may either be left empty (i.e., NULL), or may contain
a pointer to a function used to dispose of a key. It should be
declared as follows:
	int dispose(TreeKey key);
The key parameter is the key to dispose of. The function should return
zero if it was succesful, or any other value if it was unsuccesful.

The dispose key function is called when a node is disposed of, which
happens either when a node is deleted or when the tree is disposed
of.

 The Dispose Datum Function Pointer

This field may either be left empty, or may contain a pointer to
a function used to dispose of a datum. It should be declared as
follows:
	int dispose(TreeDatum datum);
The function is similar to the dispose key function which
was described above, but is also called when the datum field
of a node is changed.

 The Extra Data Field

This field may be set to any addition information the application
needs to attach to the tree.

____________________________________
The TreeIndex Data Type

While this implementation stores trees as pointers to nodes, it is
sometimes useful to be able to refer to a node in a tree
as if the tree were stored in an array. According to this scheme,
the root of the tree is stored at index 1, the first left child at
index 2, and the first right child at index 3. In general, the
left child of a node is at 2*index and the right child is at
2*index+1. While a node in a tree gives only information about
the location of its children, an index to a node specifies the
location of the node in relation to the remainder of the tree.

The nodes' index is calculated by the tree traversal functions
(tinorder, tpreorder, tpostorder, and trange), which pass this
information along to the action function [see description
of tinorder]. No use is actually made of the index, since it is
provided purely for the information of the action function.

____________________________________
Summary of Function Pointer Declarations

compare key		int compare(TreeKey key1, TreeKey key2)
order key		int order(TreeKey key1, TreeKey key2)
dispose key		int disposekey(TreeKey key)
dispose datum	int disposedatum(TreeDatum datum)
action			int action(TreeType tree,TreeNode node,TreeIndex index,void *data)

 used by tinorder, tpreorder, tpostorder, and trange
____________________________________
Brief Descriptions of the Functions

tnew		Allocates a new tree. Both the order and compare function
			pointers must be supplied (they may be the same if desired).
tdispose	Disposes of the tree and of all of its nodes. If the dispose
			key and/or the dispose datum functions are given then the
			key and/or datum fields of each node are disposed of. If either
			function returns a non-zero value then tdispose halts.

tcmp		Calls the compare function.
torder	Calls the order function

tfind		Searches the tree for the node with the given key using the
			compare function
tinsert	Creates a new node containing the given key; the node's position
			is determined using the order function. More than one occurence
			of a specific key is possible.
tdelete	Deletes the node with the given key from the tree.
			If the dispose key and/or the dispose datum functions
			are given then the key and/or datum fields of each node
			are disposed of.

tinorder	Traverses the tree in in-order, starting from the given node, or
			from the root if the node is NULL, and executes a user supplied
			action function at each node. The function should be declared
			as follows:
				int action(TreeType tree,TreeNode node,TreeIndex index,void *data);
			The parameters are:
				tree	The tree being traversed
				node	The node being visited
				index	The index to the node being visited
				data	A pointer to any user-supplied data
			The action function should return zero if it was succesful,
			otherwise a non-zero value. If a non-zero value is returned
			then tinorder halts and returns the non-zero value, otherwise
			tinorder returns 0.
tpreorder Just like tinorder except tree is traversed in pre-order.
tpostorder Just like tinorder except tree is traversed in post-order.

trange	Does a one dimensional range search for nodes whose keys
			fall between a low and high key. A user supplied action
			function is called for each node visited  nodes are visited
			in in-order  and should be declared in the same way as
			the action function for tinorder.

troot		Returns the root node of the tree, or NULL if tree is empty.
tleft		Returns the left child of the given node, or NULL if there
			is no left child.
tright	Returns the right child of the given node, or NULL if there
			is no right child.

theight	Returns height of tree starting from the given node

tkey		Returns the key field of the given node.
tdatum	Returns the datum field of the given node.
tsetdatum Sets the datum field of the given node.

tdata		Returns the data field of the tree.
tsetdata	Sets the data field of the tree.

tgetorder	Returns the tree's order function pointer.
tgetcmp		Returns the tree's compare function pointer.
tgetdisposekey		Returns the tree's dispose key function pointer.
tgetdisposedatum	Returns the tree's dispose datum function pointer.

tsetorder	Sets the tree's order function pointer.
tsetcmp		Sets the tree's compare function pointer.
tsetdisposekey		Sets the tree's dispose key function pointer.
tsetdisposedatum	Sets the tree's dispose datum function pointer.
*/

#ifndef _H_treelib
#define _H_treelib

#ifndef _H_mystdlib
#include "mystdlib.h"
#endif

/* prototypes for functions called indirectly by tree library */
#define TREE_ACTION_PROTOTYPE\
	TreeType tree, TreeNode node, TreeIndex index, void *data
#define TREE_DISPOSEKEY_PROTOTYPE	TreeKey key
#define TREE_DISPOSEDATUM_PROTOTYPE	TreeDatum datum
#define TREE_COMPARE_PROTOTYPE		TreeKey key1, TreeKey key2
#define TREE_ORDER_PROTOTYPE			TreeKey key1, TreeKey key2

/* data structures for a binary tree */
typedef size_t	TreeIndex;
typedef void	*TreeKey;
typedef void	*TreeDatum;
typedef struct TreeNode	TreeNodeStruct, **TreeNode;
struct TreeNode {
	TreeKey		key;
	TreeDatum	datum;
	TreeNode		left;
	TreeNode		right;
};

typedef struct {
	TreeNode	root;				/* root of tree */
	
	Boolean	delflag;			/* flag used for alternating side chosen for
										deletion, resulting in a slightly more
										balanced tree */
	
	/* key comparison functions; order is called when inserting, cmp
		is called when searching */
	FPTR		order;			/* ordering function */
	FPTR		cmp;				/* comparison function */

	/* disposal functions, called when a node is deleted and when
		the value of a datum changes */
	FPTR		key_dispose;	/* key dispose function */
	FPTR		datum_dispose;	/* datum dispose function */
	
	void		*data;			/* may be set to any used defined info */
} TreeStruct, **TreeType;

/* tree creation and disposal */
TreeType	tnew(FPTR order, FPTR compare);
int		tdispose(TreeType tree);

/* compare and order functions */
int		tcmp(TreeType tree, TreeKey key1, TreeKey key2);
int		torder(TreeType tree, TreeKey key1, TreeKey key2);

/* find, insert, and delete */
TreeNode	tfind(TreeType tree, TreeKey key, TreeNode *parent);
TreeNode	tinsert(TreeType tree, TreeKey key);
void		tdelete(TreeType tree, TreeKey key);

/* tree traversal */
int		tinorder(TreeType tree, TreeNode root, FPTR action, void *data);
int		tpreorder(TreeType tree, TreeNode root, FPTR action, void *data);
int		tpostorder(TreeType tree, TreeNode root, FPTR action, void *data);
int		tlevelorder(TreeType tree, TreeNode root, FPTR action, void *data);

/* range searching */
int		trange(TreeType tree, TreeNode node, TreeKey low, TreeKey high,
							FPTR action, void *data);

/* getting items in a node and relative to a node */
TreeNode		troot(TreeType tree);
TreeNode		tleft(TreeNode node);
TreeNode		tright(TreeNode node);
TreeKey		tkey(TreeNode node);
TreeDatum	tdatum(TreeNode node);

/* getting height of a tree */
TreeIndex theight(TreeNode node, TreeIndex height);

/* getting and setting tree data field */
void	*tdata(TreeType tree);
void	tsetdata(TreeType tree, void *data);

/* setting items in the node */
void	tsetdatum(TreeType tree, TreeNode node, TreeDatum datum);

/* getting tree's function pointers */
FPTR	tgetorder(TreeType tree);
FPTR	tgetcmp(TreeType tree);
FPTR	tgetdisposekey(TreeType tree);
FPTR	tgetdisposedatum(TreeType tree);

/* setting tree's function pointers */
void	tsetorder(TreeType tree, FPTR order);
void	tsetcmp(TreeType tree, FPTR compare);
void	tsetdisposekey(TreeType tree, FPTR key_dispose);
void	tsetdisposedatum(TreeType tree, FPTR datum_dispose);

/* macro equivalents for many of the above functions */
#if ! ASSERTIONS
#define tleft(n)					((**(n)).left)
#define tright(n)					((**(n)).right)
#define tkey(n)					((**(n)).key)
#define tdatum(n)					((**(n)).datum)

#define troot(t)					((**(t)).root)

#define tdata(t)					((**(t)).data)
#define tsetdata(t, d)			((void) (tdata(t) = (d)))

#define tgetunique(t)			((**(t)).unique)
#define tsetunique(t)			((void) (tgetunique(t) = (u)))

#define tgetorder(t)				((**(t)).order)
#define tgetcmp(t)				((**(t)).cmp)
#define tgetdisposekey(t)		((**(t)).key_dispose)
#define tgetdisposedatum(t)	((**(t)).datum_dispose)

#define tsetorder(t, o)			((void) (tgetorder(t) = (o)))
#define tsetcmp(t, c)			((void) (tgetcmp(t) = (c)))
#define tsetdisposekey(t,f)	((void) (tgetdisposekey(t) = (f)))
#define tsetdisposedatum(t,f)	((void) (tgetdisposedatum(t) = (f)))

#define torder(t, k1, k2)		((*tgetorder(t))((k1), (k2)))
#define tcmp(t, k1, k2)			((*tgetcmp(t))((k1), (k2)))

#endif /* ASSERTIONS */

#endif