/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
	See header ("buflib.h") for more details.
*/

#include <MyHeaders>
#include "buflib.h"

#define ALLOC_CHUNK	(512L)	/* bytes to allocate at once */

static long		buf_freespace(OutputBuffer buf);
static OSErr	buf_increase(OutputBuffer buf);
static OSErr	buf_addptr(OutputBuffer buf, Ptr ptr, long count);
static OSErr	buf_addhndl(OutputBuffer buf, Handle hndl, long offset, long count);

static Boolean	validbuffer(OutputBuffer buf);

#if ASSERTIONS
/* perform a whole bunch of tests on the buffer */
static Boolean	validbuffer(buf)
OutputBuffer	buf;
{
	if (! validhandle((Handle) buf)) return(false);
	if (! validhandle((**buf).buf)) return(false);
	if ((**buf).len < 0) return(false);
	if ((**buf).size < 0) return(false);
	if ((**buf).len >= (**buf).size) return(false);
	if ((**buf).size != hsize((**buf).buf)) return(false);
	return(true);
}
#endif

/*----------------------------------------------------------------*/
/* Buffer initialization */
/*----------------------------------------------------------------*/

/* create a new buffer */
OSErr buf_new(OutputBuffer *bufp, HCField *field)
{
	OSErr	err = noErr;
	Handle buf;
	
	*bufp = (OutputBuffer) halloc(sizeof(OutputBufferStruct));
	if ((err=MemError()) == noErr) {
		buf = halloc(ALLOC_CHUNK);
		if ((err=MemError()) == noErr) {
			(***bufp).buf = buf;
			(***bufp).field = field;
			(***bufp).size = ALLOC_CHUNK;
			(***bufp).len = 0;
			(***bufp).flush = false;
			assert(validbuffer(*bufp));
			return(err);
		}
		hfree(*bufp);
	}
	return(err);
}

/* detach buffer's handle and then dispose of buffer */
Handle buf_detach(OutputBuffer buf)
{
	Handle	hndl;
	
	assert(validbuffer(buf));
	hndl = (**buf).buf;
	hrealloc(hndl, (**buf).len+1);
	(*hndl)[(**buf).len] = 0;
	hfree(buf);
	return(hndl);
}

/*----------------------------------------------------------------*/
/* Functions for outputing a character, pointer, and handle */
/*----------------------------------------------------------------*/

/* output a single character */
OSErr buf_outputchar(OutputBuffer buf, char c)
{
	OSErr	err = noErr;
	
	assert(validbuffer(buf));
	if (buf_freespace(buf) < 1)
		err = buf_flush(buf);
	if (! err)
		(*(**buf).buf)[(**buf).len++] = c;
	return(err);
}

/* output count characters of string */
OSErr buf_outputptr(OutputBuffer buf, Ptr ptr, long count)
{
	OSErr	err = noErr;
	long	n;
	
	assert(validbuffer(buf));
	assert(ptr != NULL);
	assert(count >= 0);
	while (count > 0) {
		n = MIN(count, buf_freespace(buf));
		err = buf_addptr(buf, ptr, n);
		if (err) break;
		count -= n;
		ptr += n;
	}
	return(err);
}

/* output count characters of hndl starting from offset */
OSErr buf_outputhndl(OutputBuffer buf, Handle hndl, long offset, long count)
{
	OSErr	err = noErr;
	long	n;
	
	assert(validbuffer(buf));
	assert(validhandle(hndl));
	assert(offset >= 0);
	assert(count >= 0);
	while (count > 0) {
		n = MIN(count, buf_freespace(buf));
		err = buf_addhndl(buf, hndl, offset, n);
		if (err) break;
		count -= n;
		offset += n;
	}
	return(err);
}

/*----------------------------------------------------------------*/
/* Functions for flushing buffer */
/*----------------------------------------------------------------*/

/* flush output buffer */
OSErr buf_flush(OutputBuffer buf)
{
	OSErr	err = noErr;
	
	assert(validbuffer(buf));
	return(err);
}


/*----------------------------------------------------------------*/
/* Functions for getting info about buffer */
/*----------------------------------------------------------------*/

/* return free space in buffer */
static long buf_freespace(OutputBuffer buf)
{
	assert(validbuffer(buf));
	return((**buf).size - (**buf).len - 1);
}

/*----------------------------------------------------------------*/
/* Internal functions for adding things to buffer */
/*----------------------------------------------------------------*/

/* make space for another chunk of characters */
static OSErr buf_increase(OutputBuffer buf)
{
	OSErr	err = noErr;
	
	assert(validbuffer(buf));
	assert(buf_freespace(buf) == 0);
	hrealloc((**buf).buf, (**buf).size + ALLOC_CHUNK);
	if ((err = MemError()) == noErr)
		(**buf).size += ALLOC_CHUNK;
	return(err);
}

/* append count bytes from pointer to buffer */
static OSErr buf_addptr(OutputBuffer buf, Ptr ptr, long count)
{
	OSErr	err = noErr;
	
	assert(validbuffer(buf));
	assert(ptr != NULL);
	assert(0 <= count && count <= buf_freespace(buf));
	
	BlockMove(ptr, *(**buf).buf + (**buf).len, count);
	(**buf).len += count;
	
	if (buf_freespace(buf) == 0) {
		if ((**buf).flush)
			err = buf_flush(buf);
		if (!err && buf_freespace(buf) == 0)
			err = buf_increase(buf);
	}
	return(err);
}

/* append count bytes from handle to buffer */
static OSErr buf_addhndl(OutputBuffer buf, Handle hndl, long offset, long count)
{	
	assert(validbuffer(buf));
	assert(validhandle(hndl));
	assert(offset >= 0);
	assert(0 <= count && count <= buf_freespace(buf));
	
	return(buf_addptr(buf, *hndl + offset, count));
}
