/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
A library for simplifying buffered output from programs. Unlike
the standard IO package, the output does not go into a file
when the buffer is full or the buffer is closed; rather, the
output as appended to a handle which keeps on growing as
more output is added. When the program is finished with the
buffer it detaches the handle to the output from the buffer.
This is most useful for XCMD's and XFCN's, for instance
when building output which must replace the contents of
a field.

Future versions may implement temporary buffering of output
to disk, or even different output devices (i.e., a handle
in memory, or an output file). These features will be added
as the need arises.

Brief descriptions of the functions follow:

buf_new		Opens an output buffer.
buf_detach	Ends use of the output buffer, and returns the handle to
				the output.
buf_end		Flushes and then closes the output buffer. Not implemented.
buf_clear	Clears the contents of the buffer. Not implemented.
buf_kill		Closes the buffer without flushing its contents (any contents
				in the buffer are lost). Not implemented.
buf_outputchar	Outputs a single character to the buffer
buf_outputptr	Outputs count bytes from the pointer
buf_outputhndl	Outputs count bytes starting from the byte given by the
					offset into the handle.
buf_flush	Flushes the contents of the buffer. Has no effect at the moment.
*/

#ifndef _H_buflib
#define _H_buflib

#ifndef _H_mystdlib
#include "mystdlib.h"
#endif

#ifndef _H_hyperlib
#include "hyperlib.h"
#endif

/* structure describing a buffer */
typedef struct {
	HCField	*field;	/* the output field (NULL if none) */
	Handle	buf;		/* handle to buffer */
	Boolean	flush;	/* if true buffer is flushed when it fills up */
	long		size;		/* size of buffer (size of handle) */
	long		len;		/* length of buffer (length of data in handle) */
} OutputBufferStruct, **OutputBuffer;

OSErr	buf_new(OutputBuffer *bufp, HCField *field);
Handle buf_detach(OutputBuffer buf);
OSErr	buf_end(OutputBuffer buf);
OSErr	buf_clear(OutputBuffer buf);
void	buf_kill(OutputBuffer buf);

OSErr	buf_outputchar(OutputBuffer buf, char c);
OSErr	buf_outputptr(OutputBuffer buf, Ptr ptr, long count);
OSErr	buf_outputhndl(OutputBuffer buf, Handle hndl, long offset, long count);
OSErr	buf_flush(OutputBuffer buf);

#endif
