
/*
	These are error codes which are in use in some of my HyperCard
	XCMD's and XFCN's. To distinguish these error codes from the
	Macintosh's these error codes are all positive. Use the
	function GetErrString to translate an error code into a string.

Assigning Error Codes
	Every library can have its own set of unique error codes. To distinguish
	these codes from the others, each set is preceded with a "first" error
	code; this is one less than the first error code used by that library.
	Following the set are two more error codes, the first one giving
	the code of the last error code currently used by the library (called
	"last used"), and the second (called "last") gives the absolute last
	error code which can be used by the library without having to recompile
	everything else.
	
	The above scheme allows a library to reserve space for extra error
	codes, by leaving a gap between the "last used" and "last" error
	codes. Libraries are then chained together by setting the
	"first" error code of the next library equal to the "last" error
	code of the previous library. The specific order of libraries does
	not matter, but new libraries are usually added to the end of the
	existing list, so that other libraries don't have to be recompiled.

Converting Error Codes to Strings
	This makes use of the lookupTabl library, which looks in a special
	type of resource that stores string numbers along with the strings.
	
Problems
	There are several main compatability problems with this method of
	representing error codes.
	
		1. Compatability with the Macintosh. Nowhere does Apple state
			that all of its error codes will always be negative. In
			fact, I think some are positive (but they don't occur often).
		2. Compatability with software written by other people (or even
			by me for other people), since there is no way to enforce a
			specific numbering; in fact, the programming community would
			likely run out of error codes if we had to conform to a standard
			registration policy (unless we switched to long integers). Of
			course, it is possible to write a simple function which maps
			error codes from other software into compatible error codes.
	
		3. Updating this list for new software. If error codes are added
			or deleted, existing software will have to recompiled. I tried
			circumventing this problem by allowing libraries to reserve a
			dead-zone between the "last used" and "last" error codes.
*/

#ifndef _H_errorlib
#define _H_errorlib

#ifndef _H_mystdlib
#include "mystdlib.h"
#endif

char *ErrNumToStr(CStr255 str, ResID id, OSErr errnum);
OSErr ErrStrToNum(char *str, ResID id);

enum {
	/* basic errors (may be used by all programs) */
	ERR_BASIC_FIRST,
	ERR_BASIC_UNKNOWN=1,		/* Unknown error */
	ERR_BASIC_PROGRAM,		/* Program error (indicates a bug) */
	ERR_BASIC_ROMVER,			/* Wrong ROM version */
	ERR_BASIC_SYSVER,			/* Wrong system version */
	ERR_BASIC_HCVER,			/* Wrong version of HyperCard */
	ERR_BASIC_HYPERCARD,		/* Hypercard caused an error */
	ERR_BASIC_ARGCNT,			/* Wrong number of arguments */
	ERR_BASIC_USAGE,			/* Illegal parameter combination */
	ERR_BASIC_COMMAND,		/* Unknown command */
	ERR_BASIC_NOTIMPLEMENTED,/* Not yet implemented */
	ERR_BASIC_CANCELED,		/* Operation canceled */
	ERR_BASIC_LASTUSED =		ERR_BASIC_CANCELED,
	ERR_BASIC_LAST =			ERR_BASIC_LASTUSED + 9,

	/* strlistlib */
	ERR_STRLIST_FIRST,
	ERR_STRLIST_INDEX,		/* Illegal index */
	ERR_STRLIST_NEGCNT,		/* Negative count */
	ERR_STRLIST_LIST,			/* Bad list */
	ERR_STRLIST_EMPTY,		/* Empty list */
	ERR_STRLIST_ATTR,			/* Unknown attribute */
	ERR_STRLIST_LASTUSED =	ERR_STRLIST_ATTR,
	ERR_STRLIST_LAST =		ERR_STRLIST_LASTUSED + 4,

	/* list table */
	ERR_LTABLE_FIRST,
	ERR_LTABLE_EXISTS,		/* list already exists */
	ERR_LTABLE_NOTFOUND, 	/* list not found */
	ERR_LTABLE_LASTUSED =	ERR_LTABLE_NOTFOUND,
	ERR_LTABLE_LAST =			ERR_LTABLE_LASTUSED + 4,

	/* arraylib */
	ERR_ARRAY_FIRST,
	ERR_ARRAY_SUBSCRIPT,		/* error in subscript */
	ERR_ARRAY_LASTUSED =		ERR_ARRAY_SUBSCRIPT,
	ERR_ARRAY_LAST =			ERR_ARRAY_LASTUSED + 4,
	
	/* ArrayList XFCN */
	ERR_ALST_FIRST,
	ERR_ALST_CHKSUM,			/* incorrect check sum for global data */
	ERR_ALST_LASTUSED =		ERR_ALST_CHKSUM,
	ERR_ALST_LAST =			ERR_ALST_LASTUSED + 4,

	/* patlib */
	ERR_PAT_FIRST,
	ERR_PAT_EOF,				/* unexpected end of file */
	ERR_PAT_LITERAL,			/* illegal literal */
	ERR_PAT_DIGIT,				/* illegal digit */
	ERR_PAT_ENDTAG,			/* missing end tag marker */
	ERR_PAT_ECCL,				/* missing end character class marker */
	ERR_PAT_EGROUP,			/* missing end group marker */
	ERR_PAT_MAXTAG,			/* too many tags */
	ERR_PAT_SIZE,				/* pattern too big */
	ERR_PAT_ENDCLOSURE,		/* missing end closure marker */
	ERR_PAT_MAXMIN,			/* closure min is greater than closure max */
	ERR_PAT_MATCH,				/* error while matching pattern */
	ERR_PAT_LASTCOMPILE =	ERR_PAT_MAXMIN, /* last pattern compiling error */
	ERR_PAT_LASTUSED =		ERR_PAT_MATCH,
	ERR_PAT_LAST = 			ERR_PAT_LASTUSED + 4,

	/* Research XFCN */
	ERR_SRCH_FIRST,
	ERR_SRCH_OPTION,			/* illegal option */
	ERR_SRCH_CHKSUM,			/* damaged global data */
	ERR_SRCH_LASTUSED =		ERR_SRCH_CHKSUM,
	ERR_SRCH_LAST =			ERR_SRCH_LASTUSED + 4,

	/* string tree */
	ERR_STRTREE_FIRST,
	ERR_STRTREE_EXISTS,		/* key already exists */
	ERR_STRTREE_NOTFOUND,	/* key not found */
	ERR_STRTREE_ATTR,			/* unknown attribute */
	ERR_STRTREE_EMPTYKEY,	/* empty key */
	ERR_STRTREE_LASTUSED =	ERR_STRTREE_EMPTYKEY,
	ERR_STRTREE_LAST =		ERR_STRTREE_LASTUSED + 4,

	/* strforest */
	ERR_FOREST_FIRST,
	ERR_FOREST_EXISTS,		/* tree already exists */
	ERR_FOREST_NOTFOUND,		/* tree not found */
	ERR_FOREST_LASTUSED =	ERR_FOREST_NOTFOUND,
	ERR_FOREST_LAST =			ERR_FOREST_LASTUSED + 4,

	/* BinaryTree XFCN */
	ERR_TREE_FIRST,
	ERR_TREE_CHKSUM,			/* incorrect check sum for global data */
	ERR_TREE_LASTUSED =		ERR_TREE_CHKSUM,
	ERR_TREE_LAST =			ERR_TREE_LASTUSED + 4,
	
	ERR_ABSOLUTE_LAST			/* last error ever used (may grow or shrink) */
};

#endif