/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
Functions for simulating part of the HyperCard environment (by
implementing global variables and several callbacks) thus greatly
simplifying debugging. No longer must I worry about all the
things that can go wrong in HyperCard: I can verify that my code is
correct and then copy it into HyperCard, confident it will work.
To use, rename the main function of the XFCN or XCMD to
"hypersim_main" and make sure it is type "void" and not
"pascal void".

The function EntryPoint has a switch statement which executes
the callbacks. Look through that function to see what call backs
are implemented.

Warning:
		Function pointers may move to different addresses at each
		invocation of an XFCN. Unfortunately, this behvior won't occur
		in an application, unless your code is in segments other than
		the main segment and you unload those segments. This behaviour
		of XFCN's can make them very difficult to debug, especially
		if you store addresses of functions in data that are not
		reinitialized every time the XFCN is called. 
*/

#include <MyHeaders>
#include <console.h>
#include "hyperlib.h"
#include "hypersimlib.h"
#include "listlib.h"
#include "strlib.h"

/*-------------------------------------------------------------------------*/

int main(void);

static void hypersim_init(void);

static int	simcompare(ListItem item1, ListItem item2);
static int	simorder(ListItem item1, ListItem item2);

static void	EntryPoint(void);

static void		SetTheGlobal(String globalName, String newvalue);
static String	GetTheGlobal(String globalName);
static ListIndex	NewGlobal(String globalName);

/*-------------------------------------------------------------------------*/

static ListType	globalIndex; /* names of simulated global variables */
static ListType	globalValues;/* handles to simulated global variables */
static XCmdBlock	theParams;	/* for access to parameter block from EntryPoint */

/*-------------------------------------------------------------------------*/

int main(void)
{
	/* Setup console library to use a small window, which will fall
		above the debugger window (the values are for a small
		Macintosh screen). */
	console_options.top = 40;
	console_options.left = 4;
	console_options.nrows = 10;
	console_options.pause_atexit = true;

 	/* initialize ANSI library */
 	printf("");

 	/* restrict application size so that there will be some stack space */
	SetApplLimit(((Ptr) ApplicZone()) + 200000L);

	/* initialize simulation */
	hypersim_init();

	/* run simulation */
	hypersim_main(&theParams);

	return(0);
}

/* initialize the HyperCard simulator */
static void hypersim_init(void)
{
	/* setup parameter block to hold our entry point */
	theParams.entryPoint = EntryPoint;
	theParams.returnValue = NULL;

	/* create list to hold global variable names */
	if (globalIndex == NULL) {
		globalIndex = lnew();
		assert(! MemError());
		lsetcmp(globalIndex, simcompare);
		lsetorder(globalIndex, simorder);
	}

	/* create list to hold handles to global variables */
	if (globalValues == NULL) {
		globalValues = lnew();
		assert(! MemError());
	}
}

/* item comparison functions for the list of names of global variables */
static int simcompare(ListItem item1, ListItem item2)
{
	return(substricmp(*item1, *item2, 0));
}

static int simorder(ListItem item1, ListItem item2)
{
	return(substriorder(*item1, *item2, 0));
}

/* simulate a few of the HyperCard callbacks */
static void EntryPoint(void)
{
	char		*str;
	String	globalName;
	Handle	global;
	Handle	value;
	OSErr		err;
	XCmdBlockPtr paramPtr;
	
	paramPtr = &theParams;
	
	paramPtr->result = xresSucc;
	switch (paramPtr->request) {

	case xreqStrToBool:
		str = (char *) paramPtr->inArgs[0];
		paramPtr->outArgs[0] = EqualString(str, "\ptrue", false, true);
		break;

	case xreqBoolToStr:
		str = (char *) paramPtr->inArgs[1];
		if (paramPtr->inArgs[0])
			strcpy(str, "true");
		else
			strcpy(str, "false");
		break;
		
	case xreqGetGlobal:
		err = PtrToHand((Ptr)paramPtr->inArgs[0], &globalName,
			*((Ptr)paramPtr->inArgs[0])+1);
		assert(! err);
		PtoCstr(*globalName);
		paramPtr->outArgs[0] = (long) GetTheGlobal(globalName);
		DisposHandle(globalName);
		break;
		
	case xreqSetGlobal:
		err = PtrToHand((Ptr)paramPtr->inArgs[0], &globalName,
			*((Ptr)paramPtr->inArgs[0])+1);
		PtoCstr(*globalName);
		SetTheGlobal(globalName, (Handle) paramPtr->inArgs[1]);
		DisposHandle(globalName);
		break;

	default:
		paramPtr->result = xresNotImp;
	}
}

/*------------------------------------------------------------------------*/
/* Functions for simulating HyperCard globals */
/*------------------------------------------------------------------------*/

/* create a new global variable with the specified name */
static ListIndex NewGlobal(String globalName)
{
	ListIndex	index;
	Boolean		found;
	String		empty;
	OSErr			err;
	
	assert(lvalidlist(globalIndex));
	assert(lvalidlist(globalValues));
	assert(lvaliditem(globalName));
	
	/* locate the correct location for the global */
	index = lfirst(globalIndex);
	if (lsize(globalIndex) > 0) {
		index = llocate(globalIndex, lfirst(globalIndex),
				lsize(globalIndex), globalName, &found);
		assert(! found);
	}
		
	/* make a copy of the name of the global */
	err = HandToHand(&globalName);
	assert(! err);
	
	/* insert the copy of the name of the global into the index list */
	linsert(globalIndex, index, globalName);
	assert(MemError() == noErr);
	
	/* make an empty string */
	empty = stralloc("");
	
	/* insert the empty string into the values list */
	linsert(globalValues, index, empty);
	assert(! MemError());
	
	/* these are some postconditions */
	assert(lsize(globalIndex) == lsize(globalValues));
	assert(lvalidindex(globalIndex, index));
	assert(lvalidindex(globalValues, index));
	
	/* return the index to the new global */
	return(index);
}

/* function implementing the GetGlobal callback */
static String GetTheGlobal(String globalName)
{
	ListIndex	index;
	String		value;
	Boolean		found;
	OSErr			err;
	
	assert(lvalidlist(globalIndex));
	assert(lvalidlist(globalValues));
	assert(lvaliditem(globalName));

	/* search for global in index */
	found = false;
	if (lsize(globalIndex) > 0)
		index = llocate(globalIndex, lfirst(globalIndex),
									lsize(globalIndex), globalName, &found);

	/* if global wasn't found then create it and get its index */
	if (! found)
		index = NewGlobal(globalName);
	
	assert(lsize(globalIndex) == lsize(globalValues));
	assert(lvalidindex(globalIndex, index));
	assert(lvalidindex(globalValues, index));
	
	/* get the value of the global */
	value = lget(globalValues, index);
	assert(lvaliditem(value));
	
	/* make a copy of the value of the global */
	err = HandToHand(&value);
	assert(! err);
	
	/* return the copy */
	return(value);
}

/* function implementing the SetGlobal callback */
static void SetTheGlobal(String globalName, String newvalue)
{
	ListIndex	index;
	Handle		oldvalue;
	Boolean		found;
	OSErr			err;
	
	assert(lvalidlist(globalIndex));
	assert(lvalidlist(globalValues));
	assert(lvaliditem(globalName));
	assert(lvaliditem(newvalue));
	
	/* search for global in index */
	found = false;
	if (lsize(globalIndex) > 0) {
		index = llocate(globalIndex, lfirst(globalIndex),
								lsize(globalIndex), globalName, &found);
	}
	
	/* if global wasn't found then create it and get its index */
	if (! found)
		index = NewGlobal(globalName);
		
	assert(lsize(globalIndex) == lsize(globalValues));
	assert(lvalidindex(globalIndex, index));
	assert(lvalidindex(globalValues, index));
	
	/* get the old value of the global */
	oldvalue = lget(globalValues, index);
	assert(lvaliditem(oldvalue));
	
	/* dispose of the old value of the global */
	hfree(oldvalue);
	
	/* make a copy of the new value */
	err = HandToHand(&newvalue);
	assert(! err);
	
	/* set the value of the global to the copy of the new value */
	lset(globalValues, index, newvalue);
}
