/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
	Generic test program for XFCN's. Enter a line to execute, ommitting
	the parenthesis around the parameters and the XFCN's name. For
	instance, to enter the "insert" command to the BinaryTree XFCN use:
		insert tree key1,key2,key3 datum1,datum2,datum3 ,
	Entering a single character on a line is used to control this
	test program. The letter 't' will turn on the profiler's tracing
	of function calls, and the letter 'q' will quit the interactive
	test.
*/

#include <MyHeaders>
#include <console.h>
#include <profile.h>
#include "strlib.h"
#include "codelib.h"
#include "hypersimlib.h"

static char *progname = "BinaryTree XFCN";
static char *echofile = "output"; /* output file to echo console to */

char	*getline(char *str);
void	simulate(XCmdBlockPtr param);
void	prstr(String str);

static Boolean hide;
static Boolean trace;

static void preprogrammed(XCmdBlockPtr param);
static void interactive(XCmdBlockPtr param);
static void docommand(XCmdBlockPtr param, char *command);
static void profile(XCmdBlockPtr param);
static void	pre_exec(void);
static void	post_exec(void);

/* Preprogrammed commands to execute (you could arrange to
	have the commands read from a file, and then execute them
	line by line).  */
static char *program[16] = {
	"new tree",
	"insert tree 9;2;1;7;5;3;9;8;6 a:b:c:d:e:f:g:h:i ; :",
	"levelorder tree keys",
	"levelorder tree data",
	"levelorder tree indexes",
	"inorder tree",
	"preorder tree",
	"postorder tree",
	"height tree",
	"dispose tree",
	NULL
};
	
void hypersim_main(XCmdBlockPtr param)
{
	short	i;
	
	MaxApplZone();
	for (i = 0; i < 15; i++)
		MoreMasters();
	cecho2file(echofile, false, stdout);
	cecho2file(echofile, true, stderr);

	printf("This is a test of the %s\n", progname);
	printf("Debugging is turned %s\n", DEBUG ? "on" : "off");
	printf("Assembly language is turned %s\n", ASSEMBLY ? "on" : "off");
	printf("Assertions are turned %s\n", ASSERTIONS ? "on" : "off");
	printf("Using %s for some basic operations\n",
			ASSERTIONS ? "function calls" : "macros");
	
	InitProfile(200, 200);

	for (;;)
		simulate(param);
}

void simulate(XCmdBlockPtr param)
{
	CStr255	str;
	
	param->returnValue = NULL;
	hide = trace = _profile = _trace = false;
	
	printf("Select test type (1=preprogrammed; 2=interactive; 3=profile) ");
	switch (*getline(str)) {
	case '1':
		preprogrammed(param);
		break;
	case '2':
		interactive(param);
		break;
	case '3':
		if (! ASSERTIONS)
			profile(param);
		else
			printf("\nThe profile test would take too long when compiled with assertions.\n\n");
		break;
	}
}

static void preprogrammed(XCmdBlockPtr param)
{
	short i;
	
	printf("This is a preprogrammed test.\n\n");
	for (i = 0; program[i]; i++) {
		printf("%d: '%s'\n", i+1, program[i]);
		docommand(param, program[i]);
	}
}

static void profile(XCmdBlockPtr param)
{	
	printf("This test isn't implemented, but you can write your own.\n");
}

static void interactive(XCmdBlockPtr param)
{
	int c;
	CStr255 str;

	printf("This is an interactive test (q=quit;t=trace).\n");
	for (;;) {
		printf("> ");
		getline(str);
		if (strlen(str) == 1 && *str == 'q') break;
		docommand(param, str);
	}
}

#define MAXARG	(16)

static void docommand(XCmdBlockPtr param, char *command)
{
	short	arg;
	short i;
	char *str;
	char delim;
	
	if (strlen(command) == 1) {
		switch (*command) {
		case 't':
			trace = ! trace;
			printf("Tracing is now %s\n", trace ? "on" : "off");
			break;
		}
		return;
	}
	
	/* parse command into individual strings */
	for (arg = 0; arg < MAXARG && *command; arg++) {
		/* skip spaces */
		while (isspace(*command)) ++command;
		/* remember start of parameter */
		str = command;
		/* skip to end of parameter */
		while (*command && ! isspace(*command))
			++command;
		/* remember delimeter and terminate parameter */
		delim = *command;
		*command = 0;
		/* allocate and set the parameter */
		if (*str)
			param->params[arg] = stralloc(str);
		/* restore delimiter */
		*command = delim;
		if (*command)
			++command;
	}
	param->paramCount = arg;
	
	/* execute command */
	pre_exec();
	app_main(param);
	post_exec();
	
	/* dispose of parameters */
	while (--arg > 0)
		strfree(param->params[arg]);
		
	/* print result */
	if (param->returnValue) {
		printf("");
		prstr(param->returnValue);
		putchar('\n');
		hfree(param->returnValue);
		param->returnValue = NULL;
	}
}

/* set up environment before executing program being tested */
static void pre_exec(void)
{
	_trace = trace;
	_profile = true;
	if (hide) chide(stdin);
}

/* restore environment after executing program being tested */
static void post_exec(void)
{
	_trace = false;
	_profile = false;
	if (hide) cshow(stdin);
}

/* print a handle to a C string */
void prstr(String str)
{
	SignedByte state;
	
	if (str) {
		state = HGetState(str); HLock(str);
		printf("'%s' ", *str);
		HSetState(str, state);
	}
}
