/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

#include <MyHeaders>
#include "buildtabl.h"
#include "lookuptabl.h"
#include "strlib.h"	

/*---------------------------------------------------------------------------*/

/* structure describing a single entry in a table */
typedef struct {
	short	num;		/* number of string */
	char	str;		/* pointer to this is start of string */
} TableEntry;

/* structure describing header of a table */
typedef struct {
	short	count;	/* number of entries in table */
	short	width;	/* width of each string */
	short 	srtfld;	/* flags */
} TableHeader;
	
/* structure describing the resource */
typedef struct {
	TableHeader hdr;	/* the table's header */
	char base;			/* pointer to this is base of entries */
} TableResource;

/*	returns width of entire entry (including the string's number). */
#define entrywidth(tbl)	(((tbl)->hdr.width)+sizeof(short))

/* returns pointer to first entry in table */
#define entrybase(tbl)	(&(tbl)->base)

/* returns number of entries in table */
#define entrycount(tbl)	((tbl)->hdr.count)

/* retrieve the i'th entry in the table */
#define getentry(tbl, i) ((TableEntry*)(entrybase(tbl) + entrywidth(tbl) * (i)))

/*---------------------------------------------------------------------------*/

static OSErr readtable(ResID id, TableResource **table);
static TableEntry *binsearch_num(TableResource *table, short num);
static TableEntry *seqsearch_num(TableResource *table, short num);
static TableEntry *binsearch_str(TableResource *table, char *str);
static TableEntry *seqsearch_str(TableResource *table, char *str);

static short comparenum(void *num, void *entry);
static short comparestr(void *str, void *entry);

/*---------------------------------------------------------------------------*/

/* lookup handle to string in table */
OSErr hlookupTabl(ResID id, String str, short *number)
{
	OSErr err;
	HandleKey key;
	
	key = hlock(str, false);
	err = lookupTablStr(id, *str, number);
	hunlock(str, key);
	return(err);
}

/*---------------------------------------------------------------------------*/

/* lookup the string in the TABL resource, using either binary
	search or sequential search, depending on sort field of table. */
OSErr lookupTablStr(ResID id, char *str, short *number)
{
	TableResource *table;
	TableEntry *entry;
	OSErr err;
		
	*number = 0;
	err = readtable(id, &table);
	if (err) return(err);
	if (table->hdr.srtfld & STRFLD)
		entry = binsearch_str(table, str);
	else
		entry = seqsearch_str(table, str);
	if (entry)
		*number = entry->num;
	return(err);
}

/* lookup the number in the TABL resource, using either binary
	search or sequential search, depending on sort field of table. */
OSErr lookupTablNum(ResID id, short num, char *str)
{
	TableResource *table;
	TableEntry *entry;
	OSErr err;
		
	*str = 0;
	err = readtable(id, &table);
	if (err) return(err);
	if (table->hdr.srtfld & NUMFLD)
		entry = binsearch_num(table, num);
	else
		entry = seqsearch_num(table, num);
	if (entry)
		strcpy(str, &entry->str);
	return(err);
}

/*---------------------------------------------------------------------------*/

/*	Read the table resource. Notice that the handle to the resource
	is dereferenced but not locked, so you must not call any routines
	that may move or purge memory. */
static OSErr readtable(ResID id, TableResource **table)
{
	Handle rsrc;
	
	rsrc = GetResource(RSRC_TYPE, id);	
	if (ResError())
		return(ResError());
	if (! rsrc)
		return(resNotFound);
	
	*table = (TableResource *) *rsrc;
	
	assert((*table)->hdr.width % 2 == 0);
	
	return(noErr);
}

/*	Binary search for entry with specified number. The table
	must be sorted on the number field for this function to work. */
static TableEntry *binsearch_num(TableResource *table, short num)
{
	return(bsearch((void *) &num,
			entrybase(table),
			entrycount(table),
			entrywidth(table),
			comparenum));
}

/*	Binary search for entry with specified string. The table
	must be sorted on the string field for this function to work. */
static TableEntry *binsearch_str(TableResource *table, char *str)
{
	return(bsearch((void *) str,
			entrybase(table),
			entrycount(table),
			entrywidth(table),
			comparestr));
}

/* compare functions for binary search */

static int comparenum(void *nump, void *entryp)
{
	int *num = (int *) nump;
	TableEntry *entry = (TableEntry *) entryp;
	return(*num - entry->num);
}

static int comparestr(void *strp, void *entryp)
{
	char *str = (char *) strp;
	TableEntry *entry = (TableEntry *) entryp;
	return(stricmp(str, &entry->str));
}

/* Sequential search for entry with specified number */
static TableEntry *seqsearch_num(TableResource *table, short num)
{
	short i;
	TableEntry *entry;
	
	for (i = 0; i < entrycount(table); i++) {
		entry = getentry(table, i);
		if (num - entry->num == 0)
			return(entry);
	}
	return(NULL);
}

/* Sequential search for entry with specified string */
static TableEntry *seqsearch_str(TableResource *table, char *str)
{
	short i;
	TableEntry *entry;
	
	for (i = 0; i < entrycount(table); i++) {
		entry = getentry(table, i);
		if (stricmp(str, &entry->str) == 0)
			return(entry);
	}
	return(NULL);
}
