/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
	This implementation of the quick sort algorithm incorporates
	several improvements over a simple recursive version. These
	improvements should reduce execution time by around 20%-30%.
	Call qksort in precisely the same way as the standard qsort;
	qksort is identical to qsort except that it should be faster
	than many implementations of qsort.
	
	Qksort is based on information in Algorithms, by Robert
	Sedgewick, 2nd ed., Addison-Wesley, 1988, p. 116-126.
	
	List of improvements to naive quick sort algorithm:
		1. Tail recursion eliminated.
		2. Pushdown stack is used instead of recursive calls.
		3. Sorts smaller subfile first (larger is placed on stack).
		4. Improvements 1, 2, and 3 ensure that stack will
			never be larger than lgN.
		5. Insertion sort on small subfiles (this alone speeds
			execution by around 20%).
		6. Partitioning is done within this function, instead of
			calling a separate partition function. This saves about
			1/3 of the time when run on large random files (4096
			items, regular Mac Plus).
		7. Median of three. Makes worst case of O(n^2) very unlikely.
			Also eliminates need for sentinels in inner loops, thus
			speeding program.
		8. Uses a swap function which is optimized for the given size
			of data (i.e., char, short, long, etc.). The swap
			functions are in "swap.c" and "swapasm.c".

TEST RESULTS

	The test results (following this preface) show the profiler's
	output for the qsort function from THINK C 4.0, versus profiler's
	output for qksort run on the same data. An extra run on presorted
	data was done with qksort, in order to demonstrate that behavior
	does not deteriorate to O(n^2). The swap functions were
	compiled in C, so as to get the profiler's output for them.
	The results of each sort were verified to be correct by
	a function which scanned through the sorted lists (this scan is
	not included in the profiler's time). The profiler used as
	time units ticks of the VIA chip, which are 1.2766 seconds each.
	
THINK C's IMPLEMENTATION

	THINK C's implementation of the quick sort algorithm's only real
	improvement is the elimination of tail recursion from a naive
	implementation. Two versions of THINK C's qsort function were
	tested: qsort slow and qsort fast. The first, qsort slow, is the
	version of quick sort supplied with the compiler; this version
	makes an unnecessary call to a function called "std_compare".
	To get a fairer comparison of the algorithms, I removed
	the extra function call. I retained the extra function call in
	qsort slow, and you can see how this degraded performance. 
	
TEST RESULTS

	 Summary of tests 
	
	qksort		10.106 (random data 4096 integers)
	qsort fast	17.505 (random data 4096 integers)
	qsort slow	40.928 (random data 4096 integers)
	qksort		22.774 (sorted data 1024 integers)
	
	Qksort clearly runs much faster than either qsort fast or qsort slow.
	 
	 Detailed test results 
	
	 Qksort, random data, 4096 integers
	Function          Minimum  Maximum  Total   Average    %   Entries
	med3                   0      609   323698      473     4      683
	swap_short             0      438  1702595      117    21    14447
	compare                0      645  5888801       96    74    61286
	Total time = 7916222 (10.105849 seconds)
	
	 THINK C, qsort fast, random data, 4096 integers
	Function          Minimum  Maximum  Total   Average    %   Entries
	std_swap               0      596  2909684      261    21    11114
	qsort1                 0   488823  5170516     2475    37     2089
	compare                0      482  5631349       95    41    58989
	Total time = 13712229 (17.505032 seconds)

	 THINK C, qsort slow, random data, 4096 integers
	Function          Minimum  Maximum  Total   Average    %   Entries
	std_compare            0      897 19792887      335    61    58989
	std_swap               0      678  2890005      260     9    11114
	qsort1                 0   164693  3734897     1787    11     2089
	compare                0      458  5641777       95    17    58989
	Total time = 32060241 (40.928104 seconds)

	 Qksort, sorted data, 1024 integers
	Function          Minimum  Maximum  Total   Average    %   Entries
	med3                   0      841   437641      468     2      934
	swap_short             0      434   221828      117     1     1883
	compare                0      511 17124219       96    95   178033
	Total time = 17839816 (22.774309 seconds)
*/

#include <MyHeaders>
#include "sortlib.h"

#define SMALL		(10)	/* size of subfiles skipped by quick sort */
#define MAXSTACK	(32)	/* size of quick sort's stack (max >= lg(n)) */

/* macros for handling quicksort's pushdown stack */
#define push(stk, n)				((*(stk)++ = (n)))
#define pop(stk)					(*--(stk))
#define push2(stk, a, b)		((void) (push((stk), (a)), push((stk), (b))))
#define pop2(stk, a, b)			((void) (((a) = pop(stk)), ((b) = pop(stk))))
#define stkempty(base, stk)	((base) >= (stk))

static void _qksort(void *base, size_t hi, size_t size,
							FPTR compar, VFPTR swap);
static void *med3(void *a, void *b, void *c, size_t size,
							FPTR compar);

/* You call this function. First, we decide which swap function
	to use for the data; each swap function is optimized for
	different data sizes. Then we call the function to do
	the actual sorting. */
void qksort(void *base, size_t nmemb, size_t size, FPTR compar)
{
	_qksort(base, nmemb, size, compar, getswap(size));
}

/* Get the i'th item from the list to sort (base and size are
	local variables). */
#define get(i)		 ((void*) ((char*)base + size * (i)))

/* execute the quick sort algorithm */
static void _qksort(register void *base,
						register size_t hi,
						register size_t size,
						register FPTR compar,
						VFPTR swap)
{
	register size_t split;	/* Partition element. This serves as the
										left index to the array. When we've
										finished partitioning, it will be equal
										to the item to split the array on. This
										item is sometimes called the "pivot" item. */
	register size_t right;	/* Right edge of array */
	register void *last;		/* Pointer to last item in array. We stash
										this item at the top of the array, and
										compare all the other items to it until
										we've partitioned the array. Then we
										swap this item with the partition item
										(the partition item's index is stored
										in the SPLIT variable). */
	register size_t lo;		/* The bottom of our current subfile. */
	size_t *top;				/* Top of our stack of subfiles */
	size_t stack[MAXSTACK];	/* Storage for our stack of subfiles */
	void *median;				/* The median element as determined by
										the median of three swap done before
										the partitioning. */
	
	size_t nmemb = hi;
	
	lo = 0;
	top = stack;
	push2(top, lo, hi);
	do {
		/* skip subfiles smaller than some limit */
		if (hi - lo < SMALL) {
			pop2(top, hi, lo);
			continue;
		}
		
		/* exchange median of left, right, and middle items with last item */
		last = get(hi-1);
		if (hi - lo >= 3) {
			median = med3(get(lo), get((hi-lo) >> 1), last, size, compar);
			if (median != last)
				swap(size, median, last);
		}

		/* partition */
		split = lo-1; right = hi-1;
		for (;;) {
			while (compar(get(++split), last) < 0) ;
			while (compar(get(--right), last) > 0) ;
			if (split >= right) break;
			swap(size, get(split), get(right));
		}
		swap(size, get(split), last);
					
		/* assert([lo...split-1] < split < [split+1...hi]); */
	
		/* put larger subfile on stack and sort smaller subfile first */
		if (split - lo > hi - (split+1)) {
			if (split - lo > SMALL)
				push2(top, lo, split);	/* qksort(lo, split) */
			lo = split+1;					/* qksort(split+1, hi) */
		}
		else {
			if (hi - (split+1) > SMALL)
				push2(top, split+1, hi);/* qksort(split+1, hi) */
			hi = split;						/* qksort(lo, split) */
		}
	} while (! stkempty(stack, top));
	
	/* All items in the array are close to their final position.
		Insertion sort is therefore the best way to sort the array.
		(The items are within SMALL items of their final position,
		 where SMALL is usually between 5 and 25). */
	insort(base, hi, size, compar);
}
		
/* return median of three items */
static void *med3(void *a, void *b, void *c, size_t size,
						FPTR compar)
{
	void *tmp;
	if (compar(a, b) > 0) { tmp = a; a = b; b = tmp; }
	if (compar(a, c) > 0) { tmp = a; a = c; c = tmp; }
	if (compar(b, c) > 0) return(c);
	return(b);
}


