/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
	The function "validhandle" is probably the most useful single
	test function I have ever written for the Macintosh. It does
	a set of fairly complete and rigorous tests to determine whether
	a handle is legal or not. The function checkhandle does the
	actual checking, and returns an integer indicating what was
	wrong with the handle. I advise placing assertions with calls
	to validhandle at the start of most functions.
	
	The tests in this and the following functions are based on
	information in IM-II, p. 24-25. I don't expect anything here
	to be valid for some of the newer Mac's (I've only used this code
	on a Macintosh Plus).
*/

#include <MyHeaders>
#include "mystdlib.h"

#if DEBUG /* these functions are only useful while debugging */

/* The size of a handle's header is 8 bytes. The header immediately
	precedes the handle, so to get a pointer to the header we just
	subtract 8 bytes from the pointer to the handle. */
#define HEADER_SIZE	(8L)

/* the header of a relocatable block */
typedef struct {
	long	tag;			/* Contains flags, such as whether the handle is relocatable */
	long	correction; /* Size correction: handles are at least 12 bytes
								long, even if only a few bytes are actually used.
								Also, handles are always padded to an even size.
								This variable gives the amount of extra unused
								bytes at the end of the handle. */
	long	size;			/* Physical block size */
	long	relative;	/* A pointer to the handle's master pointer, expressed
								as an offset from the start of the heap zone. */
} HandleHeader;

/* summary of heap zone info */
typedef struct {
	long	start;
	long	end;
} ZoneInfo;

static void	HGetHeader(Handle hndl, HandleHeader *headerPtr);
static void	GetZoneInfo(ZoneInfo *zoneInfoPtr);

Boolean validhandle(Handle hndl)
{
	return(checkhandle(hndl) == noErr);
}

/* Verify that a handle is valid using various sanity checks.
	To aid in debugging the function returns a number indicating
	what check failed or 0 if the handle is ok. The handle must
	exist in the current heap zone. */
short checkhandle(Handle hndl)
{
	HandleHeader	header;	/* handle's header */
	ZoneInfo		zone;		/* summary of current heap zone information */
	Handle		master;	/* block's master pointer */
	long			deref;	/* dereferenced handle */
	int			err;		/* error indicating which test found the problem */
	
	/* get the heap zone information */
	GetZoneInfo(&zone);
	
	err = 1;
	/* 1. Handle must not be null or odd */
	if (! hndl || ((long) hndl & 1L))
		return(err);

	err++;
	/* 2. The handle must be within the current heap zone */
	if ((long)hndl < zone.start || (long)hndl > zone.end)
		return(err);
		
	err++;
	/* 3. Pointer to handle must not be null or odd */
	if (! *hndl || ((long) *hndl & 1L))
		return(err);
	
	err++;
	deref = ((long)*hndl & 0x00FFFFFFL);
	/* 4. The dereferenced handle must be within the current heap zone */
	if (deref < zone.start || deref > zone.end)
		return(err);

	/* extract header data from handle */
	HGetHeader(hndl, &header);

	err++;
	/* 5. The end of the dereferenced handle must be within the heap zone */
	if (deref + header.size - HEADER_SIZE > zone.end)
		return(err);
		
	err++;
	/* 6. Handle must be a relocatable block */
	if (! (header.tag & 0x00000080L))
		return(err);
	
	err++;
	/* 7. The physical size must not be negative, null, or odd */
	if (header.size <= 0 || (header.size & 1L))
		return(err);
	
	err++;
	/* 8. The relative handle must not be null or odd */
	if (! header.relative || (header.relative & 1L))
		return(err);
		
	/* get the master pointer */
	master = (Handle) (header.relative + (long) GetZone());
	master = (Handle) ((long) master & 0x00FFFFFFL);
	
	err++;
	/* 9. Handle should be same as master pointer */
	if (hndl != master)
		return(err);
	
	return(0);
}

/* get the relocatable block's header */
static void	HGetHeader(Handle hndl, HandleHeader *headerPtr)
{
	long	*header;

	header = (long *) (*hndl - HEADER_SIZE);
	headerPtr->tag =  (unsigned long) *header >> (unsigned long) 24;
	headerPtr->correction = headerPtr->tag & 4L;
	headerPtr->size = *header & 0x00FFFFFFL;
	headerPtr->relative = *(header + 1L);
}

/* get summary of current heap zone information */
static void	GetZoneInfo(ZoneInfo *zoneInfoPtr)
{
	THz	zone;		/* current heap zone */

	zone = GetZone();
	zoneInfoPtr->start = (long) & zone->heapData;
	zoneInfoPtr->end = (long) zone->bkLim;
}

#endif /* DEBUG */