/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
	These are assembly language (for MC68000) versions of C functions
	that can be found in "strlib.c". Following are some benchmarks
	comparing these functions with their C conterparts. Each function
	was executed 1000 times on a string with 254 characters. A timing
	for strcmp was included to indicate the lower bound on execution
	time. All times are in ticks.
		Assembly language:
			strcmp			110
			substrcmp		168
			substricmp		382
			substriorder	332
		Not assembly language:
			substrcmp		260
			substricmp		894
			substriorder	822
*/

#include <MyHeaders>
#include "mystdlib.h"

#if ASSEMBLY

size_t	substrlen(void);
int		substrcmp(void);
int		substricmp(void);
int		substriorder(void);

/*----------------------------------------------------------------*/
/* substrlen */
/*	Return length of substring of s that ends with character delim */
/*----------------------------------------------------------------*/

size_t substrlen(/*char *str1, char *str2, char delim*/)
{
	asm {
				move.l	4(a7), a0	; get string
				move.b	9(a7), d1	; get delimiter
				move.l	a0, d2		; remember start of string
	@next:	move.b	(a0)+, d0	; get characters until null or delimiter
				beq.s		@brk			;
				cmp.b		d0, d1		;
				bne.s		@next			;
	@brk:		move.l	a0, d0		; subtract start of string from end
				sub.l		d2, d0
				subq.l	#1, d0
				RTS
	}
}
	
/*----------------------------------------------------------------*/
/* convert register to uppercase or jump to label if not lowercase letter */
/*----------------------------------------------------------------*/
#define TOUPPER(reg, label)	cmpi.b	#97, reg\
										blt.s		label\
										cmpi.b	#122, reg\
										bgt.s 	label\
										subi.b 	#32, reg

/*----------------------------------------------------------------*/
/* convert register to lowercase or jump to label if not uppercase letter */
/*----------------------------------------------------------------*/
#define TOLOWER(reg, label)	cmpi.b	#90, reg\
										bgt.s 	label\
										cmpi.b	#65, reg\
										blt.s		label\
										addi.b 	#32, reg

/*----------------------------------------------------------------*/
/* This macro is useful when the delimiter in a string is reached,
	since the character in the register must be cleared before
	it can be compared to the other character to yield the difference
	between the strings being compared.
*/
/*----------------------------------------------------------------*/
#define CLRREG(reg, delim, label)\
			cmp.b	reg, delim\
			bne.s	label\
			clr.b	reg

/*----------------------------------------------------------------*/
/* This macro is used as the basic looping instructions for fetching
	the next characters from two strings being compared. The label
	in endloop is jumped to if the second character is null or
	if either of the characters match the delimiter. The last branch
	to endloop doesn't have a size specified (i.e., it isn't beq.s)
	since if the endloop label is the next instruction it would cause
	the assembler to produce the wrong instruction.
*/
/*----------------------------------------------------------------*/
#define NEXTCHAR(str1, str2, c1, c2, delim, endloop)\
			move.b	(str1)+, c1\
			move.b	(str2)+, c2\
			beq.s		endloop\
			cmp.b		c1, delim\
			beq.s		endloop\
			cmp.b		c2, delim\
			beq		endloop
			

/*----------------------------------------------------------------*/
/* substrcmp */
/* compare two substrings ending with delim; similar to strcmp.
/*----------------------------------------------------------------*/

int substrcmp(/*char *str1, char *str2, char delim*/)
{
	asm {
				movem.l	4(a7), a0/a1	; get strings
				move.b	12(a7), d2		; get delimiter 
	@next:	NEXTCHAR(a0, a1, d0, d1, d2, @brk) ; get next characters
				cmp.b		d0, d1			; while equal continue
				beq.s		@next
	@brk:		CLRREG(d0, d2, @1)		; clear characters if equal to delim
	@1:		CLRREG(d1, d2, @2)
	@2:		sub.b		d1, d0			; return difference
				ext.w		d0
				RTS
	}
}

/*----------------------------------------------------------------*/
/* substricmp */
/* ignore upper/lower case while comparing two substrings ending
/* with delim.
/*----------------------------------------------------------------*/

int substricmp(/*char *str1, char *str2, char delim*/)
{
	asm {
				movem.l	4(a7), a0/a1	; get strings
				move.b	12(a7), d2		; get delimiter 
	@next:	NEXTCHAR(a0, a1, d0, d1, d2, @brk) ; get next characters
				TOLOWER(d0, @1)			; convert to lowercase
	@1:		TOLOWER(d1, @2)
	@2:		cmp.b		d0, d1			; while equal continue
				beq.s		@next
	@brk:		CLRREG(d0, d2, @3)		; clear characters if equal to delim
	@3:		CLRREG(d1, d2, @4)
	@4:		sub.b		d1, d0			; return difference
				ext.w		d0
				RTS
	}
}

/* Compares strings for sorting. This function works by
	comparing the two strings without ignoring case,
	and remembering the first difference found. It then
	compares the remainder of the string while ignoring
	the case of the characters. If a difference is found
	in this second scan, then it is returned, otherwise
	the first difference is returned. This basically involves
	calls to substrcmp() and then to substricmp(), and I have
	simply copied those functions into this one.
*/
int substriorder(/*char *str1, char *str2, char delim*/)
{
	asm {
				; get strings and delimeter
				movem.l	4(a7), a0/a1	; get strings
				move.b	12(a7), d2		; get delimiter
				move.l	d3, -(a7)		; save d3
	
				; substrcmp: compare without ignoring case to get first difference

	@next1:	NEXTCHAR(a0, a1, d0, d1, d2, @brk1) ; get next characters
				cmp.b		d0, d1			; while equal continue
				beq.s		@next1
	@brk1:	CLRREG(d0, d2, @1)		; clear characters if equal to delim
	@1:		CLRREG(d1, d2, @2)
	@2:		sub.b		d1, d0			; get difference
				ext.w		d0
				beq.s		@done				; done if strings identical
				move.w	d0, d3			; else remember first difference

				; substricmp: compare while ignoring case to get next difference
				subq.l	#1, a0
				subq.l	#1, a1
	@next2:	NEXTCHAR(a0, a1, d0, d1, d2, @brk2) ; get next characters
				TOLOWER(d0, @3)			; convert to lowercase
	@3:		TOLOWER(d1, @4)
	@4:		cmp.b		d0, d1			; while equal continue
				beq.s		@next2
	@brk2:	CLRREG(d0, d2, @5)		; clear characters if equal to delim
	@5:		CLRREG(d1, d2, @6)
	@6:		sub.b		d1, d0			; get difference
				ext.w		d0
				
				bne.s		@done				; return this difference 
				move.w	d3, d0			; else return first difference
	
	@done:	move.l	(a7)+, d3		; restore d3
				RTS
	}
}
		
#endif /* ASSEMBLY */