/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
	Basic string functions, several of which operate on substrings,
	that is, a single character delimiter in addition to the
	regular NULL character can be used to delimit the strings. For
	instance, the function substrcmp compares two strings ending
	either with a NULL character or with a delimiter character.
	Several of these functions have equivalent functions written
	in assembly language (in the file "strasm.c"). There are
	other "miscelaneous" string functions, such as functions
	for string allocation.
	
	There are also functions for comparing strings using rules
	appropriate for alphabetic and international sorting and
	searching. An alphabetic comparison requires that the
	case of letters be ignored, but while comparing for
	sorting upper case letters should come just before lower
	case letters. For instance, when comparing, the strings "UPPER"
	and "upper" are considered the same, but when sorting, the former
	is considered smaller than the latter.
*/

#include <MyHeaders>
#include "strlib.h"

/*----------------------------------------------------------------*/
/* substring comparison and length */
/*----------------------------------------------------------------*/

#if ! (ASSEMBLY)

/* Functions equivalent to these have been written in assembly
	language.
*/

/*	Return length of substring of s that ends with character delim */
size_t substrlen(str, delim)
register char *str, delim;
{
	register char c;
	register char *p = str;
	
	while ((c=*str++) && c != delim)
		;
	return(str-p-1);
}

/* compare two substrings ending with delim; similar to strcmp */
int substrcmp(str1, str2, delim)
register char *str1, *str2, delim;
{
	register unsigned char c1, c2;
	
	do {
		c2 = *str2++;
	} while ((c1=*str1++) && c1 != delim && c1 == c2);
	if (c1 == delim) c1 = 0;
	if (c2 == delim) c2 = 0;
	return(c1 - c2);
} 

/* ignore upper/lower case while comparing two substrings ending with delim */
int substricmp(str1, str2, delim)
register char *str1, *str2, delim;
{
	register unsigned char c1, c2;
	
	do {
		c2 = *str2++;
		c1 = *str1++;
		if (! c1) break;
		if (c1 == delim) break; /* must check before tolower... */
		if (c2 == delim) break; /* ...otherwise may miss capital delim's */
		c1 = _isupper(c1) ? _tolower(c1) : c1;
		c2 = _isupper(c2) ? _tolower(c2) : c2;
	} while (c1 == c2);
	if (c1 == delim) c1 = 0;
	if (c2 == delim) c2 = 0;
	return(c1 - c2);
}	

/* compare strings for sorting */
int substriorder(str1, str2, delim)
register char *str1, *str2, delim;
{
	register unsigned char	c1, c2;
	register short	diff1, diff2;
	
	/* compare without ignoring case to get first difference */
	do {
		c2 = *str2++;
	} while ((c1=*str1++) && c1 != delim && c1 == c2);
	if (c1 == delim) c1 = 0;
	if (c2 == delim) c2 = 0;
	/* remember first difference */
	diff1 = c1 - c2;
	if (diff1) {
		/* compare remaining string while ignoring case */
		diff2 = substricmp(str1-1, str2-1, delim);
		/* if remainder is identical then return first difference, 
			else return difference while ignoring case.
		*/
		return(diff2 ? diff2 : diff1);
	}
	return(0);
}

#endif /* !(ASSEMBLY) */

/* count number of substrings */
size_t substrcnt(register char *str, char separator)
{
	register	size_t count;
	
	count = (*str != 0);
	while(*str)
		count += (*str++ == separator);
	return(count);
}

#if NOT_USED /* note: this function hasn't been tested */
/* stristr is just like the standard strstr function, except that
	it ignores the case of letters when comparing */
char *stristr(register char *s1, register char *s2)
{
	register unsigned char c1, c2;
	
	do {
		do {
			c1 = *s1++;
			c2 = *s2++;
			c1 = _isupper(c1) ? _tolower(c1) : c1;
			c2 = _isupper(c2) ? _tolower(c2) : c2;
		} while (c1 && c1 == c2);
		if (! c2)
			return(s1-1);
	} while (c1);
	return(NULL);
}
#endif /* NOT_USED */

/*----------------------------------------------------------------*/
/* International utilities string comparison */
/*----------------------------------------------------------------*/

IUstrcmp(char *str1, char *str2)
{
	register	size_t len1, len2;
	
	len1 = strlen(str1);
	len2 = strlen(str2);
	len1 = MIN(len1, SHRT_MAX);
	len2 = MIN(len2, SHRT_MAX);
	if (IUMagIDString(str1, str2, (short)len1, (short)len2) == 0)
		return(0);
	return(IUMagString(str1, str2, (short)len1, (short)len2));
}

int IUstrorder(char *str1, char *str2)
{
	register	size_t len1, len2;
	
	len1 = strlen(str1);
	len2 = strlen(str2);
	len1 = MIN(len1, SHRT_MAX);
	len2 = MIN(len2, SHRT_MAX);
	return(IUMagString(str1, str2, (short)len1, (short)len2));
}

/*----------------------------------------------------------------*/
/* Numeric string comparison */
/*----------------------------------------------------------------*/

/* This function compares two strings containing floating point
	numbers. The numbers may be preceded with a '+' or '-'. If the
	numbers are equal but a followed by non-numeric characters,
	then the strings are compared based on those characters. The
	order flag determines which rules to use when comparing
	any non-numeric characters in the strings.
	For instance:
		1234		<	123
		123.34	<	123.12
		123.abc	<	123.xyz
		1234A		>	123Z
		-23		<	+23
	With order=false:
		123.A		== 123.a
	with order=true:
		123.A		<	123.a
*/
int floatcmp(register char *s1, register char *s2, Boolean order)
{
	int sign1, sign2;	/* the sign (+/-) of each number */
	char *p1, *p2;		/* pointers to start of strings */
	int diff;			/* difference in strings */
	
	/* skip white space */
	while (_isspace(*s1)) ++s1;
	while (_isspace(*s2)) ++s2;
	
	/* compare first part, which consists of: "(+|-)?[0-9]*" */

	/* compare signs */
	sign1 = sign2 = 1;
	if (*s1 == '-' || *s1 == '+') sign1 = (*s1++ == '-') ? -1 : 1;
	if (*s2 == '-' || *s2 == '+') sign2 = (*s2++ == '-') ? -1 : 1;
	if (sign1 < sign2) return(-1);
	if (sign1 > sign2) return(1);
	
	/* remember start of strings */
	p1 = s1; p2 = s2;
	
	/* see which string has more digits */
	while (_isdigit(*s1)) ++s1;
	while (_isdigit(*s2)) ++s2;
	if (s1-p1 < s2-p2) return(-1);
	if (s1-p1 > s2-p2) return(1);
	
	/* same number of digits, so compare the digits */
	diff = strncmp(p1, p2, s1 - p1);
	if (diff)
		return(diff);
	
	/* compare second part, which consists of "\.?[0-9]*.*" */

	/* skip decimal point */
	if (*s1 == '.') ++s1;
	if (*s2 == '.') ++s2;
	
	/* compare digits */
	while (_isdigit(*s1) && *s1 == *s2)
		++s1, ++s2;
		
	/* if one of the strings has more digits than the other then it is
		larger than the other */
	if (_isdigit(*s1) && ! _isdigit(*s2))
		return(1);
	if (! _isdigit(*s1) && _isdigit(*s2))
		return(-1);
	
	/* the numbers either ended with differing digits, or they both
		ended with non-digits, so compare the remaining part as
		strings */
	if (order)
		return(striorder(s1, s2));
	return(stricmp(s1, s2));
}

/*----------------------------------------------------------------*/
/* String to number and vice-versa */
/*----------------------------------------------------------------*/

char *ltoa(long num, CStr255 str)
{
	NumToString(num, str);
	return(PtoCstr(str));
}

#if ! ANSI
long atol(char *s)
{
	long num = 0;
	int sign = 1;
	
	if (*s == '-' || *s == '+')
		sign = (*s++ == '-') ? -1 : 1;
	while (_isdigit(*s))
		num = num * 10 + sign * (*s++ - '0');
	return(num);
}
#endif /* ANSI */

#if THIS_FUNCTION_ISNT_NEEDED && ! (ANSI)
double atof(char *s) /* this function works, but I don't need it */
{
	double num = 0.0;
	int sign = 1;
	
	/* sign */
	if (*s == '-' || *s == '+')
		sign = (*s++ == '-') ? -1 : 1;
	/* digits */
	while (_isdigit(*s))
		num = num * 10.0 + sign * (*s++ - '0');
	/* period */
	if (*s++ == '.') {
		double i;
		/* digits */
		for (i = 10.0; _isdigit(*s); i *= 10.0)
			num += (sign * (*s++ - '0')) / i;
	}
	return(num);
}
#endif /* ANSI */

/*----------------------------------------------------------------*/
/* String allocation */
/*----------------------------------------------------------------*/

String stralloc(char *str)
{
	Handle hndl;
	if (PtrToHand(str, &hndl, strlen(str)+1) != noErr)
		return(NULL);
	return(hndl);
}

String strdup(String str)
{
	assert(validhandle(str));
	if (HandToHand(&str) != noErr)
		return(NULL);
	return(str);
}

