/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
DESCRIPTION
Implementation of Boyer-Moore algorithm for locating a fixed pattern
in a string. Based on description by Robert Sedgewick in Algorithms,
2nd ed., p. 286-289, (Addison-Wesley, 1988). See also R. S. Boyer
and J. S. Moore, A fast string searching algorithm, Communications
of the ACM, 20, 10 (October, 1977).

Boyer-Moore never uses more than M+N character comparisons (M is
length of pattern, N is length of string), and uses about N/M steps
if the alphabet is not small and the pattern is not long. Searching
for a word or two in English text should therefore have the
average performance of N/M steps.

This implementation may search for a pattern of any length. However,
for patterns whose lengths exceed the size of the next table, only
one of the hueristics suggested by Boyer and Moore is used. Therefore,
the number of steps will not be linear in the worst case.

You must call the function bminit before calling bmsearch. This will
initialize the tables used for skipping over the text.
*/

#include <MyHeaders>
#include "bmslib.h"

static void		initnext(uchar *pat, StrIndex len,
									NextTable next, Boolean ignrcase);
static void		initskip(uchar *pat, StrIndex len,
									SkipTable skip, Boolean ignrcase);
static Boolean	compare(uchar *pat, uchar *copy,
									StrIndex num, Boolean ignrcase);

/* initialize tables for Boyer-Moore search */
void bmsinit(bmdata, pat, ignrcase)
BoyerMoore	bmdata;		/* for search data */
uchar			*pat;			/* the pattern */
Boolean		ignrcase;	/* ignore case flag */
{
	bmdata->pat = pat;
	bmdata->len = strlen((char*)pat);
	bmdata->ignrcase = ignrcase;
	initskip(pat, bmdata->len, bmdata->skip, ignrcase);
	if (bmdata->len < MaxNextTable)
		initnext(pat, bmdata->len, bmdata->next, ignrcase);
}

/* initialize next table */
static void initnext(uchar *pat, StrIndex len, NextTable next, Boolean ignrcase)
{
	StrIndex	i, j;
	Boolean	cmp;
	uchar		p, c;
	
	next[0] = 1;
	for (j = 1; j < len; j++) {
		for (i = 1, cmp = false; i < len && ! cmp; i++) {
			StrIndex num = MIN(len-i, j); /* number of chars to compare */
			cmp = compare(pat+len-1, pat+len-1-i, num, ignrcase);
			if (cmp && len-1-i-num >= 0) {
				p = pat[len-1-num];
				c = pat[len-1-i-num];
				if (ignrcase) {
					if (_isupper(p)) p = _tolower(p);
					if (_isupper(c)) c = _tolower(c);
				}
				cmp = (p != c);
			}
		}
		next[j] = i-1;
	}
}

/* compare last num characters of pat and copy */
static Boolean compare(uchar *pat, uchar *copy, StrIndex num, Boolean ignrcase)
{
	StrIndex	i;
	uchar		p, c;
	
	while (num > 0) {
		p = *pat--; c = *copy--;
		if (ignrcase) {
			if (_isupper(p)) p = _tolower(p);
			if (_isupper(c)) c = _tolower(c);
		}
		if (p != c)
			break;
		--num;
	}
	return(num == 0);
}

/* initialize skip table using first occurences of characters from right */
static void initskip(uchar *pat, StrIndex len, SkipTable skip, Boolean ignrcase)
{
	StrIndex	i;
	uchar		c;
	
	for (i = 0; i < 256; i++)
		skip[i] = len;
	for (i = 0; i < len; i++) {
		c = pat[i];
		if (ignrcase && _isupper(c))
			c = _tolower(c);
		skip[c] = MIN(skip[c], len - i - 1);
	}
}

/* Search for pattern in string */
StrIndex bmsmatch(BoyerMoore bmdata,	/* pattern, skip, and next tables */
						register uchar *str,	/* string to find pattern in */
						register StrIndex len)	/* length of string */
{
	register StrIndex	i, j;		/* indices to string and pattern */
	register	uchar		c, p;		/* current chars from string and pattern */
	register uchar		*pat;		/* pattern */
	register StrIndex	plen;		/* length of pattern */
	register StrIndex	start;	/* index into string of start of last match attempt */
	register uchar		*next;	/* next table */
	register uchar		*skip;	/* skip table */
	register Boolean	ignrcase; /* ignore case flag */
	
	if (bmdata->len < 0) return(-1);
	if (bmdata->len == 0) return(0);
	pat = bmdata->pat;
	plen = bmdata->len;
	next = bmdata->next;
	skip = bmdata->skip;
	ignrcase = bmdata->ignrcase;
	i = j = start = bmdata->len-1;
	do {
		c = str[i]; p = pat[j];
		if (ignrcase) {
			if (_isupper(c)) c = _tolower(c);
			if (_isupper(p)) p = _tolower(p);
		}
		if (c == p)
			--i, --j;
		else {
			start += next[plen - j - 1];
			if (i + plen - j > start)
				start = i + plen - j;
			if (i + skip[c] > start)
				start = i + skip[c];
			i = start;
			j = plen - 1;
		}
	} while (j >= 0 && i < len);
	return(j < 0 ? i + 1 : -1);
}
