/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
__________________________________________________________________________
DESCRIPTION

This file implements a compiler for regular expressions.
The regular expressions are based on a merged form derived from
the UNIX utilities grep and egrep (this new form follows a
recommendation in The UNIX Programming Environment, p. 105).
Expressions are parsed using a simple recursive descent parser,
and compiled into a non-deterministic finite state machine
(an NFA) which is executed using an interpreter function.
To improve efficiency some recursion has been removed from the parser.

Notes:
	This compiler and the NFA's interpreter are preliminary and
	are likely to contain some bugs. Furthermore, they are many
	ways by which their efficiency could be improved. I will
	probably use a different algorithm and program in a future
	release (though I hope to support a similar set of regular
	expressions).

__________________________________________________________________________
A summary of the regular expressions recognized follows (from information
in The UNIX Programming Environment):
___________________________________________________________________________
c			any non-special character c matches itself
\c			turn off any special meaning of character c 
^			beginning of line
$			end of line
.			any single character
[]		any one of characters in ; ranges like a-z are legal
[^]		any single character not in ; ranges like a-z are legal
\(r\)		tagged regular expression r; can be nested
\n			what the n'th \(\) matched (n is a digit from 1 through 9)
r{n,m}	at least n and at most m occurences of regular expression r
r*			zero or more occurrences of regular expression r
r+			one or more occurrences of regular expression r
r?			zero or one occurrences of regular expression r
r1r2		regular expression r1 followed by regular expression r2
r1|r2		regular expression r1 or regular expression r2
(r)		regular expression r; can be nested
___________________________________________________________________________

Following is an informal grammar that I used while writing the parser.
Notice that this grammar has many flaws.

	<expr> ::=
		  <term>
		| <term> OR <expr>
		;
	<term> ::=
		| <factor>
		| <factor><term>
		;
	<factor> ::=
		  ANY
		| BOL
		| EOL
		| SCCL <ccl> ECCL
		| SCCL NCCL <ccl> ECCL
		| SGROUP <expr> EGROUP
		| ESCAPE TAGSTART <expr> ESCAPE TAGEND
		| ESCAPE <digit>
		| <literal>
		| <factor> <closure>
		;
	<closure> ::=
		| CLOSURE0
		| CLOSURE1
		| CLOSURE2
		| STARTCLOSURE <number> ENDCLOSURE
		| STARTCLOSURE <number> COMMA ENDCLOSURE
		| STARTCLOSURE COMMA <number> ENDCLOSURE
		| STARTCLOSURE <number> COMMA <number> ENDCLOSURE
		;
	<ccl> ::=
		  <cclfactor>
		| <cclfactor><ccl>
		;
	<cclfactor> ::=
		  <literal>
		| <literal> RANGE <literal>
		;
	<literal> ::=
		  ESCAPE letter
		| letter
		;
*/
#include <MyHeaders>
#include "patprivate.h"

/* The meta-characters in a regular expression */
#define ESCAPE		'\\'	/* Escape character */
#define BOL			'^'	/* Beginning Of Line */
#define EOL			'$'	/* End Of Line */
#define ANY			'.'	/* Match any character */
#define SCCL		'['	/* Start Character Class */
#define NCCL		'^'	/* Negate Character Class (preceded with SCCL) */
#define RANGE		'-'	/* Specifies a char range when in a char class */
#define ECCL		']'	/* End Character Class */
#define CLOSURE0	'*'	/* Match 0 or more times */
#define CLOSURE1	'?'	/* Match 0 or 1 time  */
#define CLOSURE2	'+'	/* Match 1 or more times */
#define STARTCLOSURE '{' /* Start a closure */
#define ENDCLOSURE	'}'/* End a closure */
#define COMMA		','	/* Comma separator in a closure */
#define TAGSTART	'('	/* Start Tag (preceded with ESCAPE) */
#define TAGEND		')'	/* End Tag (preceded with ESCAPE) */
#define SGROUP		'('	/* Start grouped expression */
#define EGROUP		')'	/* End grouped expression */
#define OR			'|'	/* Or two expressions */

/* Input types and structure used for compiling a regular expression */
typedef uchar		InputItem;	/* an item in the input string */
typedef size_t		InputIndex;	/* index to an input string */
typedef short		InputType;	/* data describing the type of an input item */
typedef struct {
	InputItem	*str;			/* input string */
	InputIndex	index;		/* index to string */
	short			tagnest;		/* counts nesting of tags */
	short			groupnest;	/* counts nesting of groups */
	Boolean		ccl;			/* true while in a character class */
	PatType		pat;			/* compiled pattern */
	PatState		state;		/* index to current state */
	short			ntags;		/* number of tags in pattern */
} InputRec, *Input;

/* Functions for doing any post-processing to the compiled pattern. */
static void postcompile(PatType pat);
static void visit(PatType pat, char visited[], PatState state, Boolean isnull);

/* Parsing functions. Each function adds new states as necessary
	onto the end of the compiled pattern, and returns the index of
	the initial state of the machine created (the final state
	is always given by the final entry in the arrays).
	The heirarchy of these functions roughly corresponds to the
	nonterminal symbols given in my informal grammar for regular
	expressions.
*/
static PatState	expression(Input in, OSErr *err);
static PatState	term(Input in, OSErr *err);
static PatState	factor(Input in, OSErr *err);
static PatState	closure(Input in, PatState initial, PatState oldstate, OSErr *err);
static PatState	ccl(Input in, OSErr *err);
static PatItem		digit(Input in, OSErr *err);
static PatItem		literal(Input in, OSErr *err);
static long			number(Input in, OSErr *err);

/* Input functions. Centralizing access to the input string
	simplifies the program. Determining the type of an input
	item is usually done with the typeinput function, though
	some input is identified within the parsing functions.
	Skipinput will advance to the next input item, sometimes
	skipping more than one character depending on the type
	of the current input item.
*/
static InputItem	peekinput(Input in);
static InputItem	peekpeekinput(Input in);
static InputType	typeinput(Input in);
static void			nextinput(Input in);
static void			skipinput(Input in);
static void			nextstate(Input in, OSErr *err);

/* Functions for setting the state transitions for the machine. */
static void	pat_setnext_gen(PatType pat, PatState state, PatState next);
static void	pat_setnull(PatType pat, PatState state, PatState next1, PatState next2);
static void	pat_setornull(PatType pat, PatState state, PatState next1, PatState next2);
static void	pat_setliteral(PatType pat, PatState state, PatItem item, PatState next);
static void	pat_setcliter(PatType pat, PatState state, PatState next1, PatState next2);
static void	pat_setclreps(PatType pat, PatState state, PatReps min, PatReps max);

/* Set literal was included since literal states are the states
	most often encountered in the parser. */
void	setliteral(Input in, PatItem item, OSErr *err);
static PatState getstate(Input in);

/* Macros for some simple operations */
#define getstate(in)		((in)->state)
#define nextinput(in)	((in)->index++)
#define peekinput(in)	((in)->str[(in)->index])
#define peekpeekinput(in) (peekinput(in) ? (in)->str[(in)->index+1] : 0)
#define CHECKERROR(t, v, e) ((!(e) || (e) == ERR_PAT_EOF) && (t) ? (v) : (e))

/* Compile the regular expression contained in the input string */
OSErr patcompile(PatType pat,			/* compiled pattern */
						uchar *str,			/* input string */
						Boolean ignrcase,	/* if true then ignore character case */
						size_t *where)		/* set to last index reached in input string */
{
	PatState	state;/* index for initial state */
	InputRec	in;	/* input string and parse data */
	OSErr		err;
	
	/* initialize */
	err = noErr;
	pat->ignrcase = ignrcase;
	in.str = (InputItem *) str;
	in.index = 0;
	in.tagnest = 0;
	in.groupnest = 0;
	in.ccl = false;
	in.pat = pat;
	in.state = 1;
	in.ntags = 0;
	pat_setnull(in.pat, 0, 0, 0);

	/* compile */
	state = expression(&in, &err);
	if (! err) {
		pat_setnull(in.pat, 0, state, state);
		pat_setnull(in.pat, getstate(&in), 0, 0);
		postcompile(pat);
	}

	*where = in.index;
	return(err);
}

/*----------------------------------------------------------------------*/
/* do post-processing to the compiled pattern */
/*----------------------------------------------------------------------*/

static void postcompile(pat)
PatType	pat;
{
	Boolean	visited[MAXPAT];
	PatIndex	i;
	
	for (i = 0; i < MAXPAT; i++)
		visited[i] = 0;
	visit(pat, visited, pat_next1(pat, 0), true);
}

/* recursively visit every state in the pattern */
static void visit(pat, visited, state, isnull)
PatType	pat;			/* the pattern */
char		visited[];	/* set for each state visited */
PatState	state;		/* current state */
Boolean	isnull;		/* true if all states were null states */
{
	if (state == 0)
		;
	else if (pat_isliteral(pat, state)) {
		switch (pat_item(pat, state)) {
		case P_BOL: /* convert P_BOL into BOL */
			if (visited[state] || ! isnull)
				pat_setitem(pat, state, BOL);
			break;
		case P_EOL: /* convert P_EOL into EOL */
			if	(	! pat_isfinal(pat, pat_next(pat, state)) &&
					! pat_isornull(pat, pat_next(pat, state)))
			{
				pat_setitem(pat, state, EOL);
			}
			break;
		}
		if (visited[state] < 2) {
			++visited[state];
			visit(pat, visited, pat_next(pat, state), false);
		}
	}
	else if (pat_isnull(pat, state) || pat_iscliter(pat, state)) {
		if (visited[state] < 2) {
			++visited[state];
			visit(pat, visited, pat_next1(pat, state), isnull);
			visit(pat, visited, pat_next2(pat, state), isnull);
		}
	}
	else {
		if (visited[state] < 2) {
			++visited[state];
			visit(pat, visited, state+1, false);
		}
	}
}

/*----------------------------------------------------------------------*/
/*	Functions for compiling a regular expression (the functions roughly
	correspond to informal grammar given at head of this file).
*/
/*----------------------------------------------------------------------*/

/*	Compile the regular expression. */
static PatState expression(in, err)
Input	in;
OSErr	*err;
{
	PatState	initial;	/* initial state of machine create by this function */
	
	initial = term(in, err);
	if (*err) goto error;
	
	if (typeinput(in) == P_OR) {
		PatState previnitial;	/* previous machines initial state */
		PatState	prevfinal;		/* null state used for terminating OR */
		PatState nextinitial;	/* initial state of next machine */

		skipinput(in);				/* skip the OR symbol */
		
		previnitial = initial;		/* remember initial state of previous machine */
		
		prevfinal = getstate(in);/* reserve a null state to point to exit of OR */
		nextstate(in, err);
		if (*err) goto error;

		initial = getstate(in);	/* get entry point for OR machine */
		nextstate(in, err);		/* and reserve this state */
		if (*err) goto error;
		
		nextinitial = expression(in, err);/* evaluate second expression */
		if (*err) goto error;
		
		/* fill in OR machine's states */
		pat_setnull(in->pat, initial, nextinitial, previnitial);
		pat_setornull(in->pat, prevfinal, getstate(in), getstate(in));
		
	}
error:
	return(initial);
}

/*	Compiles the terms of an expression. A while loop eliminates
	tail recursion. The loop repeats until the current input
	item is something that isn't handled by the factor function.
*/
static PatState term(in, err)
Input	in;
OSErr	*err;
{
	Boolean	done;		/* true when finished */
	PatState	initial;	/* initial state of machine create by this function */

	initial = factor(in, err);
	done = false;
	do {
		if (*err)
			done = true;
		else switch(typeinput(in)) {
			case P_TAGSTART:
			case P_TAGREF:
			case P_ANY:
			case P_SCCL:
			case P_NCCL:
			case P_SGROUP:
			case P_ESCAPE:
			case P_BOL:
			case P_EOL:
			case P_LETTER:
				(void) factor(in, err);
				break;
			default:
				done = true;
		}
	} while (! done);
	return(initial);
}

/*	Compiles various parts of a regular expression (not a very useful comment) */
static PatState factor(in, err)
Input	in;
OSErr	*err;
{
	PatItem		item;		/* input item if letter or digit */
	InputType	type;		/* type of current input */
	PatState		oldinitial;/* for remembering original state index */
	PatState		initial;	/* initial state of machine create by this function */
	InputIndex	start;	/* remembers start of (, [, and \( for errors */
	
	initial = oldinitial = getstate(in);
	type = typeinput(in);
	switch (type) {
	case P_SGROUP: /* parse '(' <expr> ')' */
		in->groupnest++;
		start = in->index;
		skipinput(in);

		initial = expression(in, err);
		pat_setnext_gen(in->pat, oldinitial-1, initial);
		
		*err = CHECKERROR(typeinput(in) != P_EGROUP, ERR_PAT_EGROUP, *err);
		if (*err) {
			in->index = start;
			goto error;
		}
		skipinput(in);
		in->groupnest--;
		
		break;
		
	case P_TAGSTART:		/* parse '\' '(' <expr> '\' ')' */
		
		*err = CHECKERROR(in->ntags >= MAXTAG, ERR_PAT_MAXTAG, *err);
		if (*err) goto error;

		start = in->index;
		skipinput(in);
		in->tagnest++;
		
		setliteral(in, pat_num2tagstart(in->ntags++), err);
		if (*err) goto error;

		(void) expression(in, err);

		*err = CHECKERROR(typeinput(in) != P_TAGEND, ERR_PAT_ENDTAG, *err);
		if (*err) {
			in->index = start;
			goto error;
		}
		
		skipinput(in);
		in->tagnest--;
		
		setliteral(in, pat_num2tagend(in->ntags-1), err);
		if (*err) goto error;
		
		break;

	case P_SCCL:	/* parse '[' <ccl> ']' */
	case P_NCCL:	/* parse '[' '^' <ccl> ']' */
		/* skip SCCL and set flag that we're inside a ccl  */
		start = in->index;
		skipinput(in);
		in->ccl = true;
		
		/* reserve space for ccl marker */
		nextstate(in, err);
		if (*err) goto error;
		
		/* parse the ccl */
		(void) ccl(in, err);
		*err = CHECKERROR(peekinput(in) != ECCL, ERR_PAT_ECCL, *err);
		if (*err) {
			in->index = start;
			goto error;
		}

		/* skip the ECCL */
		skipinput(in);
		in->ccl = false;
		
		/* set the ccl marker in the pattern, with the next state
			pointing to the last item in the ccl. */
		pat_setliteral(in->pat, initial, type, getstate(in)-1);
		break;
		
	case P_TAGREF:		/* parse '\' <digit> */
		/* get the tagreference number and set it in the pattern */
		skipinput(in);
		item = digit(in, err);
		if (*err) goto error;
		setliteral(in, pat_num2tagref(item-'1'), err);
		if (*err) goto error;
		break;
		
	case P_ANY:
	case P_BOL:
	case P_EOL:
		skipinput(in);
		setliteral(in, type, err);
		if (*err) goto error;
		if (in->index == 0 && type == P_BOL)
			type = 0; /* so closure won't apply */
		break;
		
	default:
		/* set the literal character */
		item = literal(in, err);
		if (*err) goto error;
		setliteral(in, item, err);
		if (*err) goto error;
	}

	if (type != P_TAGSTART && type)
		initial = closure(in, initial, oldinitial, err);

error:
	return(initial);
}

/* compile a closure */
PatState closure(in, initial, oldstate, err)
Input		in;
PatState	initial;
PatState	oldstate;
OSErr		*err;
{
	PatReps	min, max;	/* min and max repetitions for closure */
	
	switch(peekinput(in)) {
	case CLOSURE0: /* zero or more */
		pat_setnull(in->pat, getstate(in), getstate(in)+1, initial);
		pat_setnext_gen(in->pat, oldstate-1, getstate(in));
		initial = getstate(in);
		nextstate(in, err);
		if (*err) goto error;
		nextinput(in);
		break;
		
	case CLOSURE1: /* zero or one */
		pat_setnull(in->pat, getstate(in), getstate(in)+1, initial);
		pat_setnext_gen(in->pat, initial, getstate(in)+1);
		pat_setnext_gen(in->pat, oldstate-1, getstate(in));
		initial = getstate(in);
		nextstate(in, err);
		if (*err) goto error;
		nextinput(in);
		break;
		
	case CLOSURE2: /* one or more */
		pat_setnull(in->pat, getstate(in), getstate(in)+1, initial);
		pat_setnext_gen(in->pat, oldstate-1, initial);
		nextstate(in, err);
		if (*err) goto error;
		nextinput(in);
		break;
	
	case STARTCLOSURE: /* a {n,m} closure */
		/* parse the closure to get min and max repetitions */
		min = max = 1;
		nextinput(in);
		if (_isdigit(peekinput(in)))
			min = max = number(in, err);
		if (*err) goto error;
		if (peekinput(in) == COMMA) {
			max = USHRT_MAX;
			skipinput(in);
			if (_isdigit(peekinput(in)))
				max = number(in, err);
			if (*err) goto error;
		}
		if (max < min) {
			*err = ERR_PAT_MAXMIN;
			goto error;
		}
		if (peekinput(in) != ENDCLOSURE) {
			*err = ERR_PAT_ENDCLOSURE;
			goto error;
		}
		skipinput(in);
		
		/* set min and max */
		pat_setclreps(in->pat, getstate(in), min, max);
		nextstate(in, err);
		if (*err) goto error;
		
		/* link in closure */
		pat_setnext_gen(in->pat, getstate(in)-2, getstate(in));
		pat_setcliter(in->pat, getstate(in), getstate(in)+1, initial);
		pat_setnext_gen(in->pat, oldstate-1, getstate(in));
		initial = getstate(in) - 1;
		nextstate(in, err);
		if (*err) goto error;
		break;
	}

error:
	return(initial);
}

/*	Compile a character class. */
static PatState ccl(in, err)
Input	in;
OSErr	*err;
{
	InputItem	item1, item2;
	PatState		initial;	/* initial state of machine create by this function */
	
	initial = getstate(in);
	/* get items until an end character class marker is encountered */
	do {
		/* get first item */
		item1 = literal(in, err);
		/* check for a character range */
		if (!*err && peekinput(in) == RANGE && peekpeekinput(in) != ECCL) {
			/* skip range marker and get second item */
			nextinput(in);
			item2 = literal(in, err);
			/* make sure item1 is smaller than or equal to item2 */
			if (item1 > item2) {
				InputItem tmpitem = item1;
				item1 = item2;
				item2 = tmpitem;
			}
			/* set range */
			if (!*err) setliteral(in, P_RANGE, err);
			if (!*err) setliteral(in, item1, err);
			if (!*err) setliteral(in, item2, err);
		}
		else if (! *err)
			setliteral(in, item1, err);
	} while (! *err && peekinput(in) != ECCL);
	return(initial);
}

/* Parse a number from the input */
static long number(in, err)
Input	in;
OSErr	*err;
{
	long n = 0;
	while (_isdigit(peekinput(in))) {
		n = n * 10 + peekinput(in) - '0';
		nextinput(in);
	}
	return(n);
}

/* Parse a single digit from the input. */
static PatItem digit(in, err)
Input	in;
OSErr	*err;
{
	InputItem	item;
	
	if (_isdigit(peekinput(in))) {
		item = peekinput(in);
		nextinput(in);
	}
	else if (! typeinput(in))
		*err = ERR_PAT_EOF;
	else
		*err = ERR_PAT_DIGIT;
	return(item);
}

/* Parse a single literal character from the input. */
static PatItem literal(in, err)
Input			in;
OSErr			*err;
{
	InputItem	item;
	
	if (typeinput(in) == P_ESCAPE) {
		skipinput(in);
		item = peekinput(in);
		nextinput(in);
	}
	else if (typeinput(in) == P_LETTER) {
		item = peekinput(in);
		nextinput(in);
	}
	else if (! typeinput(in))
		*err = ERR_PAT_EOF;
	else
		*err = ERR_PAT_LITERAL;
	if (pat_ignrcase(in->pat) && _isupper(item))
		item = _tolower(item);
	return(item);
}

/*----------------------------------------------------------------------*/
/* Functions for handling the input string. */
/*----------------------------------------------------------------------*/

/*	Skip one or more input items, depending on the type of
	the current item.
*/
static void skipinput(in)
Input	in;
{
	switch(typeinput(in)) {
	case P_NCCL:
	case P_TAGSTART:
	case P_TAGEND:
		nextinput(in);
		break;
	}
	nextinput(in);
}

/*	Return the type of the current input item. May look ahead one
	or more characters.
*/
static InputType typeinput(Input in)
{
	InputItem	item;
	
	item = peekinput(in);
	if (in->ccl) {
		switch(item) {
		case 0:
			return(0);
		case ESCAPE:
			return(P_ESCAPE);
		default:
			return(P_LETTER);
		}
	}
	else switch(item) {
	case 0:
		return(0);
	case EOL:
		return(P_EOL);
	case BOL:
		return(P_BOL);
	case ANY:
		return(P_ANY);
	case SGROUP:
		return(P_SGROUP);
	case OR:
		return(P_OR);
	case SCCL:
		return((peekpeekinput(in) == NCCL) ? P_NCCL : P_SCCL);
	case EGROUP:
		return((in->groupnest > 0) ? P_EGROUP : P_LETTER);
	case ESCAPE:
		item = peekpeekinput(in);
		if (_isdigit(item) && item != '0')
			return(P_TAGREF);
		else switch(item) {
		case TAGEND:
			return((in->tagnest > 0) ? P_TAGEND : P_ESCAPE);
		case TAGSTART:
			return(P_TAGSTART);
		default:
			return(P_ESCAPE);
		}
		break;
	default:
		return(P_LETTER);
	}
}

/*----------------------------------------------------------------------*/
/* Functions for setting the state transitions.
/*----------------------------------------------------------------------*/

static void nextstate(Input in, OSErr *err)
{
	if (in->state >= MAXPAT)
		*err = ERR_PAT_SIZE;
	else
		in->state++;
}

static void setliteral(Input in, PatItem item, OSErr *err)
{
	pat_setliteral(in->pat, getstate(in), item, getstate(in)+1);
	nextstate(in, err);
}

static void pat_setnext_gen(PatType pat, PatState state, PatState next)
{
	switch (pat_type(pat, state)) {
	case PAT_NULL_STATE:
	case PAT_ORNULL_STATE:
	case PAT_CLITER_STATE:
		pat_setnext1(pat, state, next);
		break;
	case PAT_LITERAL_STATE:
		pat_setnext(pat, state, next);
		break;
	default:
		assert(false);
	}
}

static void pat_setnull(PatType pat, PatState state, PatState next1, PatState next2)
{
	pat_settype(pat, state, PAT_NULL_STATE);
	pat_setnext1(pat, state, next1);
	pat_setnext2(pat, state, next2);
}

static void pat_setornull(PatType pat, PatState state, PatState next1, PatState next2)
{
	pat_settype(pat, state, PAT_ORNULL_STATE);
	pat_setnext1(pat, state, next1);
	pat_setnext2(pat, state, next2);
}

static void pat_setliteral(PatType pat, PatState state, PatItem item, PatState next)
{
	pat_settype(pat, state, PAT_LITERAL_STATE);
	pat_setitem(pat, state, item);
	pat_setnext(pat, state, next);
}

static void pat_setcliter(PatType pat, PatState state, PatState next1, PatState next2)
{
	pat_settype(pat, state, PAT_CLITER_STATE);
	pat_setnext1(pat, state, next1);
	pat_setnext2(pat, state, next2);
}

static void pat_setclreps(PatType pat, PatState state, PatReps min, PatReps max)
{
	pat_settype(pat, state, PAT_CLREPS_STATE);
	pat_setmin(pat, state, min);
	pat_setmax(pat, state, max);
}
