/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
	Functions for interpreting the NFA produced by the regular
	expression compiler in "patcompile.c". These are preliminary
	functions which are likely to be significantly improved
	in the future.
*/

#include <MyHeaders>
#include "patprivate.h"

#define STR_STACKEMPTY	(-1L)	/* indicates an empty stack */

/* String index stack data types */
#define STR_MAXSTACK	(256)
typedef size_t	StrStackItem;	/* an item in the stack */
typedef short		StrStackIndex;	/* index to an item in the stack */
typedef struct {
	StrStackItem	array[STR_MAXSTACK];
	StrStackIndex	top;
} StrStack;

/* String index stack functions */
static void				str_stkinit(StrStack *stk);
static void				str_stkpush(StrStack *stk, StrStackItem item);
static StrStackItem	str_stkpop(StrStack *stk);
static Boolean			str_stkempty(StrStack *stk);
static Boolean			str_stkfull(StrStack *stk);

#if ! ASSERTIONS
#define str_stkinit(stk)			((stk)->top = 0)
#define str_stkempty(stk)			((stk)->top == 0)
#define str_stkfull(stk)			((stk)->top >= STR_MAXSTACK)
#define str_stkpush(stk, item)	((stk)->array[(stk)->top++] = (item))
#define str_stkpop(stk)				((stk)->array[--(stk)->top])
#endif

/* Function for handling character classes */
static Boolean	ccl(PatType pat, uchar c, PatState state);

/* Function for handling tag references */
static Boolean	tagref(uchar *str, size_t index, size_t length,
							PatMatchType match, short tagnum);

/* search for first occurrence of pattern in string, return true if found */
Boolean patsearch(PatType pat,			/* pattern to search for */
						uchar *str,				/* string to search in */
						size_t index,			/* index into string */
						size_t length,			/* length of string */
						PatMatchType match,	/* resulting match data */
						OSErr *err)
{
	short i; /* general purpose indexer */
	Boolean found; /* true if we've found a match */
	PatMatchStruct empty; /* for clearing the match data */
	
	*err = noErr;
	
	/* clear empty match record to be used for clearing actual match record */
	empty.hastag = false;
	empty.start = empty.end = 0;
	for (i = 0; i < MAXTAG; i++)
		empty.tags[i].start = empty.tags[i].end = 0;

	/* initialize */
	found = false;
	*match = empty;

	/* match at each character until a match or end of string */
	do {
		if (match->hastag)
			*match = empty;
		found = patmatch(pat, str, index, length, match, err);
	} while (! found && ! *err && str[index++]);
	
	return(found);
}

/* match the pattern to the string, return true if found */
Boolean patmatch(PatType pat,
						uchar *str,
						size_t index,
						size_t length,
						PatMatchType match,
						OSErr *err)
{
	register PatState	state;	/* current state */
	register PatItem	pitem;	/* current item in pattern */
	register uchar		sitem;	/* current item in string */
	size_t	start;		/* initial index into string */
	PatState	next;			/* for null states */
	PatReps	cnt;			/* closure counter */
	PatStack	_stk;			/* stack of alternate states to process */
	PatStack	_clstk;		/* stack of closure counters */
	StrStack	_strstk;		/* indices visited during evaluation of closures */
	register PatStack	*stk = &_stk;
	register PatStack	*clstk = &_clstk;
	register StrStack	*strstk = &_strstk;
	
	
	/* initialize */
	pat_stkinit(stk);
	pat_stkinit(clstk);
	str_stkinit(strstk);
	match->start = start = index;
	state = pat_next1(pat, 0);
	
	/* repeat until a final state is reached or all posibilities have been tried */
	while (state && state != PAT_STACKEMPTY) {

#if APPLICATION && ANSI && DEBUG
		/*fprintf(stderr, "index = %ld, state = %d\n", index, state);*/
#endif
		
		switch (pat_type(pat, state)) {
		case PAT_LITERAL_STATE:
			
			sitem = str[index];
			pitem = pat_item(pat, state);
			if (pat_ignrcase(pat) && _isupper(sitem))
				sitem = _tolower(sitem);
				
			if ((pitem == sitem || pitem == P_ANY) && index < length) {
				state = pat_next(pat, state);
				++index;
			}
			else if (pitem == P_BOL && index == 0)
				state = pat_next(pat, state);
			else if (pitem == P_EOL && sitem == 0)
				state = pat_next(pat, state);
			else if (pat_isccl(pitem) && index < length && ccl(pat, sitem, state)) {
				state = pat_next(pat, pat_next(pat, state));
				++index;
			}
			else if (pat_istagref(pitem) &&
						tagref(str, index, length, match, pat_tagref2num(pitem)))
			{
				state = pat_next(pat, state);
				index += pat_tagsize(match->tags, pat_tagref2num(pitem));
			}
			else if (pat_istagstart(pitem)) {
				match->tags[pat_tagstart2num(pitem)].start = index;
				state = pat_next(pat, state);
				match->hastag = true;
			}
			else if (pat_istagend(pitem)) {
				match->tags[pat_tagend2num(pitem)].end = index;
				state = pat_next(pat, state);
			}
			else {
				/* match failed, so try an alternative state, and backtrack
					in string to start of alternative */
				state = pat_stkpop(stk);
				if (! str_stkempty(strstk))
					index = str_stkpop(strstk);
			}
			break;

		case PAT_CLITER_STATE:
			/* match a counted closure */
			
			/* get remaining counts for this closure */
			cnt = pat_stkpop(clstk);
			
			if (cnt >= pat_max(pat, state-1)) {
				/* we've matched the specified number of times, so
					exit closure */
				state = pat_next1(pat, state);
				break;
			}
			else if (cnt < pat_min(pat, state-1)) {
				/* we still haven't matched the minimum number of times,
					so increment count and push back on to stack */
				state = pat_next2(pat, state);
				pat_stkpush(clstk, cnt+1);
				break;
			}
			
			/* we've matched at least the minimum but less than the
				maximum, so increment count, then push back onto stack,
				and then continue matching closure */
			pat_stkpush(clstk, cnt+1);
			
			/* no break */

		case PAT_NULL_STATE:
		case PAT_ORNULL_STATE:
			/*
				Handle a null state presenting at most two choices.
				If there's only one choice -- which would occur in the
				case of a "connective" null state -- we branch directly
				to it. If there are two choices, then we push one
				of them on to the stack so that we can execute it
				if the first one fails, and we remember our current
				position in the input string so that we can backtrack
				to it.
			*/
			
			next = pat_next1(pat, state);
			state = pat_next2(pat, state);
			if (next != state) {
				if (pat_stkfull(stk)) goto error;
				if (str_stkfull(strstk)) goto error;
				pat_stkpush(stk, next);
				str_stkpush(strstk, index);
			}
			break;

		case PAT_CLREPS_STATE:
			/*
				We've entered the initialization node of a bracketed
				closure, so we push a count of 0 onto the closure
				stack. The count is incremented every time we succesfully
				match the regular expression to which this closure applies.
			*/
			if (pat_stkfull(clstk)) goto error;
			pat_stkpush(clstk, 0);
			++state;
			break;
		
		default:
			assert(false);
		}

		if (index > start && str[index-1] == 0)
			break;
		else if (index > length)
			break;
	}
	match->end = index;
	return(state == 0);

error:
	*err = ERR_PAT_MATCH;
	return(false);
}

/* match the character to the current character class */
static Boolean ccl(PatType pat, uchar c, PatState state)
{
	PatState	next;		/* state at end of ccl */
	Boolean	invert;	/* true if in an NCCL */
	
	invert = (pat_item(pat, state) == P_NCCL);
	next = pat_next(pat, state);
	for (state += 1; state <= next; ++state) {
		if (pat_item(pat, state) == P_RANGE) {
			if (pat_item(pat, state+1) <= c && c <= pat_item(pat, state+2))
				break;
			state += 2;
		}
		else if (pat_item(pat, state) == c)
			break;
	}
	return(invert ? !(state <= next) : state <= next);
}

/* match tagged portion of string to current location in string */
static Boolean tagref(register uchar *str,
							size_t index,
							size_t length,
							PatMatchType match,
							short tagnum)
{
	register uchar	*start, *end;
				
	start = str + match->tags[tagnum].start;
	end = str + match->tags[tagnum].end;
	while (index < length && start < end && *start++ == str[index++])
		;
	return(start == end);
}
				
/*----------------------------------------------------------------------*/
/* String index stack functions. */
/*----------------------------------------------------------------------*/

#if ASSERTIONS

static void str_stkinit(StrStack *stk)
{
	stk->top = 0;
}

static void str_stkpush(StrStack *stk, StrStackItem item)
{
	assert(! str_stkfull(stk));
	stk->array[stk->top++] = item;
}

static StrStackItem str_stkpop(StrStack *stk)
{
	assert(! str_stkempty(stk));
	return(stk->array[--stk->top]);
}

static Boolean str_stkfull(StrStack *stk)
{
	return(stk->top >= STR_MAXSTACK);
}

static Boolean str_stkempty(StrStack *stk)
{
	return(stk->top == 0);
}
#endif /* ASSERTIONS */