/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
	Routines used by the functions in "patcompile.c" and "patmatch.c",
	which are used for compiling an NFA from a regular expression
	and then matching the NFA to an input string. Many of the
	following functions are used only during debugging, and are
	replaced by faster macros in the final program.
*/
	
#include <MyHeaders>
#include "patprivate.h"

static Boolean	validstate(PatState state);

/* return true if next state is the final state */
Boolean pat_isfinal(PatType pat, PatState state)
{
	switch (pat_type(pat, state)) {
	case PAT_NULL_STATE:
	case PAT_CLITER_STATE:
		return(pat_next1(pat, state) == 0 && pat_next2(pat, state) == 0);
	}
	return(false);
}

#if ASSERTIONS

static Boolean validstate(PatState state)
{
	return(0 <= state && state < MAXPAT);
}

/*----------------------------------------------------------------------*/
/* Functions for getting pattern types.
/*----------------------------------------------------------------------*/

PatStateType pat_type(PatType pat, PatState state)
{
	assert(validstate(state));
	return(pat->state[state].type);
}

Boolean pat_isnull(PatType pat, PatState state)
{
	return(pat_type(pat, state) == PAT_NULL_STATE || pat_isornull(pat, state));
}

Boolean pat_isornull(PatType pat, PatState state)
{
	return(pat_type(pat, state) == PAT_ORNULL_STATE);
}

Boolean pat_isliteral(PatType pat, PatState state)
{
	return(pat_type(pat, state) == PAT_LITERAL_STATE);
}

Boolean pat_iscliter(PatType pat, PatState state)
{
	return(pat_type(pat, state) == PAT_CLITER_STATE);
}

Boolean pat_isclreps(PatType pat, PatState state)
{
	return(pat_type(pat, state) == PAT_CLREPS_STATE);
}

/*----------------------------------------------------------------------*/
/* Functions for getting pattern items.
/*----------------------------------------------------------------------*/

Boolean pat_ignrcase(PatType pat)
{
	return(pat->ignrcase);
}

PatItem pat_item(PatType pat, PatState state)
{
	assert(pat_isliteral(pat, state));
	return(pat->state[state].a.item);
}

PatState pat_next(PatType pat, PatState state)
{
	assert(pat_isliteral(pat, state));
	return(pat->state[state].b.next);
}

PatState pat_next1(PatType pat, PatState state)
{
	assert(pat_isnull(pat, state) || pat_iscliter(pat, state));
	return(pat->state[state].a.next1);
}

PatState pat_next2(PatType pat, PatState state)
{
	assert(pat_isnull(pat, state) || pat_iscliter(pat, state));
	return(pat->state[state].b.next2);
}

PatReps pat_min(PatType pat, PatState state)
{
	assert(pat_isclreps(pat, state));
	return(pat->state[state].a.min);
}

PatReps pat_max(PatType pat, PatState state)
{
	assert(pat_isclreps(pat, state));
	return(pat->state[state].b.max);
}

/*----------------------------------------------------------------------*/
/* Functions for setting pattern items.
/*----------------------------------------------------------------------*/

void pat_settype(PatType pat, PatState state, PatStateType type)
{
	assert(validstate(state));
	pat->state[state].type = type;
}

void pat_setitem(PatType pat, PatState state, PatItem item)
{
	assert(pat_isliteral(pat, state));
	pat->state[state].a.item = item;
}

void pat_setnext(PatType pat, PatState state, PatState next)
{
	assert(pat_isliteral(pat, state));
	assert(validstate(next));
	pat->state[state].b.next = next;
}

void pat_setnext1(PatType pat, PatState state, PatState next1)
{
	assert(pat_isnull(pat, state) || pat_iscliter(pat, state));
	assert(validstate(next1));
	pat->state[state].a.next1 = next1;
}

void pat_setnext2(PatType pat, PatState state, PatState next2)
{
	assert(pat_isnull(pat, state) || pat_iscliter(pat, state));
	assert(validstate(next2));
	pat->state[state].b.next2 = next2;
}

void pat_setmin(PatType pat, PatState state, PatReps min)
{
	assert(pat_isclreps(pat, state));
	pat->state[state].a.min = min;
}

void pat_setmax(PatType pat, PatState state, PatReps max)
{
	assert(pat_isclreps(pat, state));
	pat->state[state].b.max = max;
}

/*----------------------------------------------------------------------*/
/* Pattern stack functions. */
/*----------------------------------------------------------------------*/

void pat_stkinit(PatStack *stk)
{
	stk->top = 0;
}

void pat_stkpush(PatStack *stk, PatStackItem item)
{
	assert(! pat_stkfull(stk));
	stk->array[stk->top++] = item;
}

PatStackItem pat_stkpop(PatStack *stk)
{
	return(pat_stkempty(stk) ? PAT_STACKEMPTY : stk->array[--stk->top]);
}

Boolean pat_stkfull(PatStack *stk)
{
	return(stk->top >= PAT_MAXSTACK);
}

Boolean pat_stkempty(PatStack *stk)
{
	return(stk->top == 0);
}

#endif /* ASSERTIONS */