/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
	Functions for executing a search for the Research XFCN. You should
	read the XFCN's manual for more complete descriptions of the
	input format, options, and output format of a search.
	
	The function ExecSearch is called once the search command is
	identified. It then initializes the data structures used while
	searching and parses the list of options to the search command.
	
	The search takes the input string (contained in a single handle),
	and searches each substring in it; the substrings are
	by default delimited by the newline character, though a different
	delimiter may be specified by one of the options. Output is,
	by default, the substrings that matched the search pattern, but can
	also be in the form of small chunk expressions giving the start
	and end indices of the matched portion of the input string.
	
	If the 'containers' option is specified, then the
	input string is interpreted as a list of HyperCard
	containers to search. And, if a list of cards to
	search is specified, then the contents of the
	containers in each of the listed cards will be searched.
*/

#include <MyHeaders>
#include "search_private.h"

static OSErr search_stack(SearchData *sd, String list);

static OSErr search_cards(SearchData *sd, String list,
								size_t first_card, size_t last_card,
								size_t first_container, size_t last_container,
								size_t first_char, size_t last_char);

static OSErr search_one_card(SearchData *sd, String list, size_t card,
									size_t first_container, size_t last_container,
									size_t first_char, size_t last_char);

static OSErr search_one_container(SearchData *sd, Handle name,
											size_t first_char, size_t last_char);

OSErr ExecSearch(GlobalData gdata, short arg)
{
	SearchData sd; /* options and other data */
	OSErr err;
	
	/* initialize */
	err = _search_init(&sd, gdata, arg);
	if (err) return(err);
	
	/* search */
	if (sd.opts.containers)
		err = search_stack(&sd, sd.input);
	else {
		err = _search_input(&sd, sd.opts.start_char, ULONG_MAX);
		if (! err && ! sd.stop)
			err = _search_input(&sd, 0, sd.opts.start_char);
	}
	
	/* output */
	if (! err && sd.opts.match) {
		char *s = GetMatched(sd.gdata) ? "true" : "false";
		err = buf_outputptr(sd.buf, s, strlen(s));
	}
	result2handle(GetParam(sd.gdata), buf_detach(sd.buf));
	
	return(err);
}

/* search all specified cards */
static OSErr search_stack(SearchData *sd, String list)
{
	OSErr err = noErr;
	
	
	/* search second part of start_card */
	
	/* search from start_char of start_container of start_card */
	err = search_cards(sd, list,
							sd->opts.start_card,
							sd->opts.start_card + 1,
							sd->opts.start_container, sd->opts.start_container + 1,
							sd->opts.start_char, ULONG_MAX);
	/* search from start_container+1 of start_card */
	if (! err && ! sd->stop) {
		err = search_cards(sd, list,
							sd->opts.start_card,
							sd->opts.start_card + 1,
							sd->opts.start_container + 1, ULONG_MAX,
							0, ULONG_MAX);
	}
	
	/* search second part of cards */
	if (sd->opts.many_cards) {
	
		/* search from start_card+1 to last_card */
		if (! err && ! sd->stop) {
			err = search_cards(sd, list,
									sd->opts.start_card + 1,
									sd->opts.last_card,
									0, ULONG_MAX,
									0, ULONG_MAX);
		}
		
		/* wrap around and search first part of cards */
		
		/* search from first_card to start_card */
		if (! err && ! sd->stop) {
			err = search_cards(sd, list,
									sd->opts.first_card,
									sd->opts.start_card,
									0, ULONG_MAX,
									0, ULONG_MAX);
		}
	}
	
	/* wrap around and search first part of start_card */
	
	/* search from container 0 to start_container of start_card */
	if (! err && ! sd->stop) {
		err = search_cards(sd, list,
							sd->opts.start_card,
							sd->opts.start_card + 1,
							0, sd->opts.start_container,
							0, ULONG_MAX);
	}
	/* search from char 0 to start_char of start_container of start_card */
	if (! err && ! sd->stop) {
		err = search_cards(sd, list,
							sd->opts.start_card,
							sd->opts.start_card + 1,
							sd->opts.start_container,
							sd->opts.start_container + 1,
							0, sd->opts.start_char);
	}
	
	return(err);
}

/* search the containers in all of the specified cards */
static OSErr search_cards(SearchData *sd, String list,
								size_t first_card, size_t last_card,
								size_t first_container, size_t last_container,
								size_t first_char, size_t last_char)
{
	size_t card;
	OSErr err = noErr;
	
	for (card = first_card; ! err && ! sd->stop && card < last_card; card++) {
		err = search_one_card(sd, list, card,
									first_container, last_container,
									first_char, last_char);
	}
	return(err);
}

/* search all the specified containers in the specified card */
static OSErr search_one_card(SearchData *sd, String list, size_t card,
									size_t first_container, size_t last_container,
									size_t first_char, size_t last_char)
{
	size_t	container;	/* container number */
	String	name;			/* name of container */
	CStr255	appstr;		/* string appended to end of name */
	CStr31	numstr;		/* for converting numbers to strings */
	size_t	i;				/* index into list */
	OSErr err = noErr;
	
	/* allocate handle to hold container's name */
	name = halloc(sizeof(Str255));
	if (! name) return(MemError());
	
	/* if searching many cards then append " of card " to container names */
	*appstr = 0;
	if (sd->opts.many_cards) {
		strcpy(appstr, " of card ");
		strcat(appstr, ltoa(card+1, numstr));
	}
		
	/* search every container in list */
	i = 0;
	for (container = 0;
			! err && ! sd->stop && (*list)[i] && container < last_container;
				container++)
	{
		size_t len;	/* length of container's name */
		char *str;	/* pointer to start of container's name */
		char delim;	/* delimiter of container's name */
		
		/* get container's name */
		str = *list + i;
		len = substrlen(str, COMMA);
		delim = str[len];
		str[len] = 0;
		
		/* remember that there is more than one container */
		if (delim)
			sd->opts.many_containers = true;
		
		/* copy container's name to name handle */
		**name = 0;
		strncat(*name, str, sizeof(Str255) - strlen(appstr) - 1);
		strcat(*name, appstr);

		/* advance to next container name */
		i += len;
		(*list)[i] = delim;
		if (delim) ++i;

		/* skip until we get to the first container */
		if (container < first_container)
			continue;
		
		/* fill in container info */
		sd->container.card = card;
		sd->container.name = name;
		sd->container.num = container;
		
		err = search_one_container(sd, name, first_char, last_char);
	}
	hfree(name);
	return(err);
}

/* search the named container */
static OSErr search_one_container(SearchData *sd, Handle name,
											size_t first_char, size_t last_char)
{
	OSErr err = noErr;
	CStr255 str;
	
	/* get contents of container */
	CtoPstr(strcpy(str, *name));
	sd->input = EvalExpr(GetParam(sd->gdata), (StrPtr) str);
	if (GetParam(sd->gdata)->result)
		err = ERR_BASIC_HYPERCARD;
	else if (sd->input) {
		/* search the contents */
		err = _search_input(sd, first_char, last_char);
		hfree(sd->input);
	}
	else
		err = ERR_BASIC_HYPERCARD;
	return(err);
}

