/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
	This file contains functions which are part of the set of
	functions used to search for strings in a HyperCard string
	or container.
	
	The function search_init initializes the data structures
	used, compiles the pattern being search for, and parses
	the user supplied options.
*/

#include <MyHeaders>
#include "search_private.h"

static OSErr sd_init(SearchData *sd, GlobalData gdata, short arg);
static OSErr pattern_init(SearchData *sd, short arg);
static void init_cursors(SearchData *sd);
static void init_options(SearchData *sd);
static OSErr parse_options(SearchData *sd, short arg);

/*----------------------------------------------------------------*/
/* Initializing search data structure and parsing search options */
/*----------------------------------------------------------------*/

/* initialize the search data */
OSErr _search_init(SearchData *sd, GlobalData gdata, short arg)
{
	OSErr err = noErr;

	SetMatched(gdata, false);
	err = sd_init(sd, gdata, arg + 1);
	if (! err) err = parse_options(sd, arg + 2);
	if (! err) err = pattern_init(sd, arg);
	if (! err) err = buf_new(&sd->buf, NULL);
	return(err);
}

/* initialize the search data */
static OSErr sd_init(SearchData *sd, GlobalData gdata, short arg)
{	
	sd->gdata = gdata;
	sd->stop = false;
	init_cursors(sd);
	init_options(sd);
	return(param2handle(GetParam(gdata), arg, &sd->input));
}

/* initialize the cursors */
static void init_cursors(SearchData *sd)
{
	CursHandle cursor;
	short i;
	short id;
	
	for (i = 0; i < NBUSY_CURS; i++) {
		id = GetRsrcID(sd->gdata, RSRC_BUSY_CURS_0 + i);
		cursor = (CursHandle) GetResource('CURS', id);
		if (cursor)
			BlockMove(*cursor, sd->opts.period.cursors + i, sizeof(Cursor));
	}
	sd->opts.period.last_cursor = 0;
}

/* compile pattern */
static OSErr pattern_init(SearchData *sd, short arg)
{
	String patstr;
	HandleKey key;
	OSErr err = noErr;

	err = param2handle(GetParam(sd->gdata), arg, &patstr);
	if (err) return(err);
	
	key = hlock(patstr, true);
	
	if (sd->opts.meta == META_NONE)
		bmsinit(&sd->u.bmdata, (uchar*)*patstr, sd->opts.ignrcase);
	else {
		size_t where; /* location of first error */
		err = patcompile(&sd->u.pat, (uchar*)*patstr, sd->opts.ignrcase,
								&where);
		if (err)
			SetErrorWhere(sd->gdata, where);
	}
	
	hunlock(patstr, key);
	return(err);
}

static void init_options(SearchData *sd)
{
	SearchOptions *opts;
	
	opts = &sd->opts;
	opts->invert	= false;
	opts->count		= false;
	opts->number	= false;
	opts->ignrcase = false;
	opts->chunk		= false;
	opts->global	= false;
	opts->first		= false;
	opts->match		= false;
	opts->containers = false;
	
	opts->periodic	= true;
	opts->period.evt_interval = 30; /* should be shorter */
	opts->period.spin_interval = 30;
	opts->period.evt_next = TickCount() + opts->period.evt_interval;
	opts->period.spin_next = TickCount() + opts->period.spin_interval;
	opts->period.cancel = true;
	opts->period.spin = true;
	
	opts->separator = NEWLINE;
	opts->meta = META_UNIX;

	opts->first_card	= 0;
	opts->last_card	= _cards_in_stack(GetParam(sd->gdata));
	opts->start_card	= _current_card_num(GetParam(sd->gdata));
	opts->start_container = 0;
	opts->start_char	= 0;
	
	opts->many_cards	= false;
	opts->many_containers = false;
}

/* convert the parameter into fields in the options structure */
static OSErr parse_options(SearchData *sd, short arg)
{
	SearchOptions *opts; /* pointer to the options field of the search data */
	CStr255		word;	/* current option word */
	OptionType	opt;	/* current option */
	short			i;		/* index to current option word */
	size_t		num;	/* temp number */
	OSErr			err;

	err = noErr;
	opts = &sd->opts;
	for (i = 0; param2word(GetParam(sd->gdata), arg, i, word) == noErr; i++) {
		err = str2option(sd->gdata, word, &opt);
		if (err) break;
		
		switch (opt) {
		case OPTION_INVERT:	opts->invert = true; break;
		case OPTION_COUNT:	opts->count = true; break;
		case OPTION_NUMBER:	opts->number = true; break;
		case OPTION_IGNRCASE:opts->ignrcase = true; break;
		case OPTION_GLOBAL:	opts->global = true; /* ... */
		case OPTION_CHUNK:	opts->chunk = true; break;
		case OPTION_MATCH:	opts->match = true; /* ... */
		case OPTION_FIRST:	opts->first = true; break;
		case OPTION_CONTAINERS: opts->containers = true; break;
		case OPTION_NOPERIODIC:	opts->periodic = false; break;
		case OPTION_NOCANCEL:	opts->period.cancel = false; break;
		case OPTION_NOSPIN:		opts->period.spin = false; break;
		case OPTION_NOMETA:	opts->meta = META_NONE; break;
		case OPTION_UNIXMETA:opts->meta = META_UNIX; break;
		case OPTION_SEPARATOR:
			(void) param2word(GetParam(sd->gdata), arg, ++i, word);
			opts->separator = *word;
			break;
		case OPTION_FIRST_CARD:
		case OPTION_LAST_CARD:
		case OPTION_START_CARD:
		case OPTION_START_CHAR:
		case OPTION_START_CONTAINER:
			(void) param2word(GetParam(sd->gdata), arg, ++i, word);
			num = labs(atol(word));
			switch (opt) {
			case OPTION_FIRST_CARD:
				sd->opts.many_cards = true;
				opts->first_card =  num > 0 ? num-1 : 0;
				break;
			case OPTION_LAST_CARD:
				sd->opts.many_cards = true;
				opts->last_card = num;
				break;
			case OPTION_START_CARD:
				sd->opts.many_cards = true;
				opts->start_card = num > 0 ? num-1 : 0;
				break;
			case OPTION_START_CHAR:
				opts->start_char =  num > 0 ? num-1 : 0; break;
			case OPTION_START_CONTAINER:
				opts->start_container =  num > 0 ? num-1 : 0; break;
			}
			break;
		default:
			err = ERR_SRCH_OPTION;
			break;
		}
	}
		
	if (sd->opts.many_cards)
		sd->opts.many_containers = true;

	return(err);
}
