/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
	This file contains functions which are part of the set of
	functions used to search for strings in a HyperCard string
	or container.
	
	The function search_input searches the string contained in the
	sd->input field for matches to the pattern being searched for.
	The input string is broken into substrings using the separator
	character (default NEWLINE) as a delimiter. The function
	search_output is called after each succesful search to output
	the results.
	
	The search terminates when either the entire input has been
	processed, or the 'first' option was specified and there has
	been at least one match (the sd->stop flag will be set to
	true), or an error occurred.
*/

#include <MyHeaders>
#include "search_private.h"

static OSErr search_line(SearchData *sd, MatchData *md);
static OSErr boyermoore(SearchData *sd, MatchData *md, size_t offset);
static OSErr regexp(SearchData *sd, MatchData *md, size_t offset);

/* search the input and output all matches (depending on options) */
OSErr _search_input(SearchData *sd, size_t first_char, size_t last_char)
{
	MatchData md;				/* data describing matched line */
	Boolean	search;			/* flag that we should search current line */
	PatMatchStruct	empty;	/* for clearing match data */
	OSErr err = noErr;
	short i;

	/* clear a match record to be used for clearing the match record */
	empty.hastag = false;
	empty.start = empty.end = 0;
	for (i = 0; i < MAXTAG; i++)
		empty.tags[i].start = empty.tags[i].end = 0;

	/* initialize match data */
	err = noErr;
	md.count = 0;
	md.index = 0;
	md.length = 0;
	md.first_index = 0;
	md.last_index = 0;
	md.lnum = 0;

	/* search every substring in input */
	while (! err && ! sd->stop && (*sd->input)[md.index] && md.index < last_char)
	{
		/* find end of current substring */
		md.length = substrlen(*sd->input + md.index, sd->opts.separator);

		if (md.index + md.length >= first_char) {

			/* determine start of range of characters in substring to search */
			if (md.index < first_char)
				md.first_index = first_char; /* search from start_char */
			else
				md.first_index = md.index; /* search from start of line */

			/* determine end of range of characters to match */
			if (md.index + md.length >= last_char)
				md.last_index = last_char; /* search up to last_char */
			else
				md.last_index = md.index + md.length; /* search up to end of line */
	
			/* clear match data */
			md.match = empty;
			md.found = false;
			
			/* search the current substring */
			err = search_line(sd, &md);
		}
		
		/* advance to next substring */
		md.index += md.length;
		if ((*sd->input)[md.index])
			++md.index;
		++md.lnum;
	}
	
	/* output any data at the end of the search */
	if (err == noErr)
		err = _search_output_end(sd, &md);
		
	return(err);
}
		
/* search current line for pattern */
static OSErr search_line(SearchData *sd, MatchData *md)
{
	size_t		index;	/* index to start of match */
	SignedByte	state;	/* old state of locked input handle */
	char			*str;		/* pointer to start of current substring */
	size_t		offset;	/* offset from start of current substring */
	OSErr			err = noErr;
	
	/* initialize */
	offset = md->first_index;
	state = HGetState(sd->input);

	/* search for each unique occurrence in line */
	do {
		/* search for string using appropriate algorithm */
		if (sd->opts.meta == META_NONE)
			err = boyermoore(sd, md, offset);
		else
			err = regexp(sd, md, offset);
		if (err) break;
		
		if (sd->opts.invert)
			md->found = ! md->found;
	
		if (md->found) {
		
			/* remember that there was a match */
			SetMatched(sd->gdata, true);
			
			/* output results */
			err = _search_output(sd, md);
		}
		
		/* advance past current match */
		offset = md->first_index + md->match.end;
		
		/* execute any periodic actions */
		if (! err && sd->opts.periodic)
			err = _search_periodic(sd);
			
		/* stop at first match if 'first' option was specified */
		sd->stop = sd->stop || (sd->opts.first && md->found);

	} while (sd->opts.global && md->found &&
				! sd->stop && ! err && offset < md->last_index);
	
	return(err);
}

/* use Boyer-Moore algorithm */
static OSErr boyermoore(SearchData *sd, MatchData *md, size_t offset)
{
	size_t index;
	HandleKey key;
	
	key = hlock(sd->input, false);

	/* search */
	index = bmsmatch(&sd->u.bmdata,
							(uchar*) *sd->input + offset,
							md->last_index - offset);
	/* see if we found something */
	if (index != -1L) {
		/* setup match data to indicate the matched portion of the line */
		md->match.start = index + offset - md->first_index;
		md->match.end = md->match.start + sd->u.bmdata.len;
		md->found = true;
	}
	else {
		/* setup match data to indicate that there was no match */
		md->match.start = md->match.end = md->last_index;
		md->found = false;
	}
	hunlock(sd->input, key);
	return(noErr);
}

/* use regular expression algorithm */
static OSErr regexp(SearchData *sd, MatchData *md, size_t offset)
{
	HandleKey key;
	OSErr err;
	
	key = hlock(sd->input, false);
	md->found = patsearch(&sd->u.pat,
								(uchar*) *sd->input + offset,
								offset - md->first_index,
								md->last_index - offset,
								&md->match,
								&err);
	hunlock(sd->input, key);
	return(err);
}
		
