/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
	This file contains functions which are part of the set of
	functions used to search for strings in a HyperCard string
	or container.
	
	The function output_search is called after each succesful
	search of an input string. It outputs data describing the
	match, in a format determined by the options.
	The default configuration simply outputs the matched
	substring. In the case of the 'chunk' option, a
	special expression is built (using the function build_chunk)
	describing the matched portion of the input string. Many
	other options alter the precise format of the output.
	
	The output is actually accumulated in a buffer using
	the buflib library. Eventually, when the search terminates,
	we return a handle to the buffer to HyperCard. In the
	future I may add an option to dump large outputs to
	a file, instead of storing them in RAM; this option
	is especially desirable since HyperCard can only cope
	with a maximum of 32000 characters in each container.
*/

#include <MyHeaders>
#include "search_private.h"

static OSErr	output_container_name(SearchData *sd);
static void		buildchunk(SearchData *sd, MatchData *md, CStr255 chunk);
static char		*makechunk(char *chunk, size_t start, size_t end);

/*----------------------------------------------------------------*/

/* output the results of the current search */
OSErr _search_output(SearchData *sd, MatchData *md)
{
	CStr255	str;
	short		len;
	OSErr		err;
	
	err = noErr;
	
	/* don't output anything if the report match flag is set;
		the search function will automatically stop when we encounter
		the first match, and then it will call output either "true" or
		"false", depending uppon whether we matched something. */
	if (sd->opts.match)
		return(noErr);
	
	/* output data in requested format */
	if (sd->opts.count)
		md->count++;
	else if (sd->opts.chunk) {
		/* output chunk expression */
		buildchunk(sd, md, str);
		err = buf_outputptr(sd->buf, str, strlen(str));
	}
	else {
		/* output container name if searching more than one container */
		err = output_container_name(sd);

		/* insert the line number followed by a comma before the output line */
		if (sd->opts.number) {
			ltoa(md->lnum + 1, str);
			err = buf_outputptr(sd->buf, str, strlen(str));
			if (!err)
				err = buf_outputchar(sd->buf, COMMA);
		}
		/* output the line followed by the separator */
		if (!err) err = buf_outputhndl(sd->buf, sd->input, md->index, md->length);
		if (!err) err = buf_outputchar(sd->buf, sd->opts.separator);
	}
	return(err);
}

/* insert the name of the input container followed by a comma before output line */
static OSErr output_container_name(SearchData *sd)
{
	OSErr err = noErr;
	
	if (sd->opts.containers && sd->opts.many_containers) {
		err = buf_outputhndl(sd->buf, sd->container.name, 0,
												strlen(*sd->container.name));
		if (!err)
			err = buf_outputchar(sd->buf, COMMA);
	}
	return(err);
}

/* build a chunk expression describing the matched string */
static void buildchunk(SearchData *sd, MatchData *md, CStr255 chunk)
{
	CStr31	numstr;	/* string used for converting numbers */
	size_t	offset;	/* offset from start of line or file */
	short		i;			/* loop index for outputing tags */
	Boolean	haschunk;
	
	/* initialize */
	*chunk = 0;
	offset = (sd->opts.separator == NEWLINE) ? 0 : md->index;
	
	/* build chunk expression for whole match of the form
		"start_index end_index line_number" */
	chunk = makechunk(chunk, offset + md->match.start, offset + md->match.end);
	*chunk++ = SPACE; *chunk = 0;
 	chunk += strlen(strcat(chunk, ltoa(md->lnum + 1, numstr)));
	
	/* If we're processing containers then append the number of the container
		and the number of the current card. */
	if (sd->opts.containers) {
		*chunk++ = SPACE; *chunk = 0; 
		chunk += strlen(strcat(chunk, ltoa(sd->container.num + 1, numstr)));
		if (sd->opts.many_cards) {
			*chunk++ = SPACE; *chunk = 0;
			chunk += strlen(strcat(chunk, ltoa(sd->container.card + 1, numstr)));
		}
	}
	
	/* If any tags matched then append chunk expressions for each tag */
	if (md->match.hastag) {
		for (i = 0; i < MAXTAG; i++) {
			*chunk++ = COMMA; *chunk = 0;
			if (md->match.tags[i].start < md->match.tags[i].end) {
				chunk = makechunk(chunk, offset + md->match.tags[i].start,
										offset + md->match.tags[i].end);
			}
		}
	}

	*chunk++ = NEWLINE; *chunk = 0;
}

/* make a single chunk expression of the form "start end"; returns pointer
	to end of new chunk expression. */
static char *makechunk(char *chunk, size_t start, size_t end)
{
	CStr31 numstr;
	chunk += strlen(strcat(chunk, ltoa(start + 1, numstr)));
	*chunk++ = SPACE; *chunk = 0;
	chunk += strlen(strcat(chunk, ltoa(end, numstr)));
	return(chunk);
}

/* output any data after all of input has been processed */
OSErr _search_output_end(SearchData *sd, MatchData *md)
{
	CStr31 str;
	OSErr err = noErr;
	
	if (sd->opts.count) {
		/* output container name followed by number of matches */
		output_container_name(sd);
		ltoa(md->count, str);
		err = buf_outputptr(sd->buf, str, strlen(str));
		if (! err) err = buf_outputchar(sd->buf, NEWLINE);
	}
	return(err);
}

