/*
Copyright  1990 Ari I. Halberstadt

This file is part of a free program. For precise terms of distribution
and use please see the files in the folder Legal accompanying this
software. If you make any modifications to this file you must add the
date, your name, and a description of what you have done to the
revision history below. This notice and the copyright notice must
remain intact in all copies of this file. The files in the folder
Legal must be available to anyone receiving copies of this file.

Add initials and name here:
Initials	Name
AIH		Ari I. Halberstadt

Revision History (date format is mm-dd-yy):
Date      Name  Comments
07-11-90	 AIH   Version 0.9
*/

/*
	Header file for functions executing a search for the Research XFCN.
	You should read the XFCN's manual for more complete descriptions
	of the input format, options, and output format of a search.
*/

#ifndef _H_search_private
#define _H_search_private

#ifndef _H_RsrchMain
#include "RsrchMain.h"
#endif

#ifndef _H_bmslib
#include "bmslib.h"
#endif

#ifndef _H_patlib
#include "patlib.h"
#endif

#ifndef _H_strlib
#include "strlib.h"
#endif

#ifndef _H_buflib
#include "buflib.h"
#endif


/* A few useful character constants. The application flag is useful
	while testing, since THINK C's implementation of the stdio library
	automatically converts a carriage return into a newline. The
	Macintosh normally uses a carriage return as a line terminator. */
#if APPLICATION
	#define NEWLINE	'\n'
#else
	#define NEWLINE	'\r'
#endif
#define COMMA		','
#define SPACE		' '
#define COLON		':'

/* Types of metacharacters. Only META_NONE and META_UNIX are
	currently supported. META_NONE results in use of the
	Boyer-Moore search algorithm, while META_UNIX uses
	UNIX style metacharacters. Eventually I plan to support
	the MPW metacharacter style as well. */
typedef enum { META_NONE, META_UNIX, META_MPW } MetaType;

/* Structure containing search options. The structure is initialized
	by the init_options function, and thereafter its fields are
	refered to by various functions. The flags in this structure
	correspond to the options available to the user. */
typedef struct {
	Boolean invert:1;			/* invert search */
	Boolean count:1;			/* count matches */
	Boolean number:1;			/* number output lines */
	Boolean ignrcase:1;		/* ignore case in character comparisons */
	Boolean chunk:1;			/* output in chunk form */
	Boolean global:1;			/* search for all occurrences on each line */
	Boolean first:1;			/* stops at first match */
	Boolean match:1;			/* returns true if there was at least one match */
	Boolean containers:1;	/* input specifies list of containers to search */

	Boolean	periodic:1;		/* execute periodic actions */

	/* structure describing periodic actions (all times are in ticks) */
	struct {
		size_t	evt_interval;	/* interval between calls to WaitNextEvent */
		size_t	spin_interval;	/* interval between periodic spins of cursor */
		size_t	evt_next;		/* time of next call to WaitNextEvent */
		size_t	spin_next;		/* time of next cursor spin */
		Boolean	cancel:1;		/* check for cancel events */
		Boolean	spin:1;			/* spin the busy cursor */
		short		last_cursor;	/* last busy cursor */
		Cursor	cursors[NBUSY_CURS]; /* busy cursors */
	} period;
	
	char	  separator;		/* character to use as input separator */
	MetaType	meta;				/* type of metacharacters to use */

	size_t first_card;		/* first card to search */
	size_t last_card;			/* last card to search */
	size_t start_card;		/* card to start searching from */
	size_t start_container;	/* container to start searching from */
	size_t start_char;		/* character to start searching from */

	/* extra flags set by program depending on options */
	size_t many_cards;		/* we're searching more than one card */
	size_t many_containers;	/* more than one container */

} SearchOptions;

/* Structure containing information needed while searching. This
	structure is passed around to all of the search routines, thus
	avoiding global variables and rediculously long lists of
	parameters. The structure is initialized by search_init(),
	and its fields are thereafter accessed in no particular
	order by the search functions. For instance, if the output
	function needs to determine what format to output in (chunk,
	number, etc.), it looks in the options structure, figures
	out what it needs to do, and then appends the output to
	the output buffer. */
typedef struct {
	GlobalData		gdata;	/* the xfcn's global data */
	SearchOptions	opts;		/* the search options */
	String			input;	/* handle to input string */
	OutputBuffer	buf;		/* buffered output containing results of search */
	
	Boolean stop;	/* search terminates when this is true */
	
	/* This structure is for outputing the results of a search when the
		containers option is specified. The structure contains the
		name of the current input container being searched, and
		the number of the input container in the list of containers
		(for instance, if the list of containers is "a,b,c", and we're
		processing container b, then num will be equal to 1). The
		card_num field contains the number of the current card being
		searched. If only one card is being searched then this
		field is ignored. */
	struct {
		Handle name;	/* name of current container */
		size_t num;		/* number of current container */
		size_t card;	/* current card number */
	} container;

	/* This union contains the compiled pattern data to use while
		searching. The pattern is compiled once when we startup,
		and this data is used thereafter, instead of referring to the
		actual pattern as entered by the user. */
	union {
		PatTypeStruct		pat;	/* compiled regular expression */
		BoyerMooreStruct	bmdata;/* compiled Boyer-Moore pattern */
	} u;
} SearchData;

/* Structure describing current input string and matched portion of string.
	We use this to keep track of where we are in the input, and other
	data relating to our search, such as how many matches were found. Like
	the SearchData structure, this structure is also passed around to
	several of the search functions. */
typedef struct {
	PatMatchStruct match;	/* data returned by the regular expression matcher
										describing the matched portion of the input */
	Boolean		found;		/* True if pattern was found. Notice that the xfcn's
										global data will also store a flag indicating
										whether there was at least one match; you can
										access this flag using the GetMatched function. */
	size_t		index;		/* index to start of current substring being searched */
	size_t		length;		/* length of current substring being searched */
	size_t		first_index;/* first character to search in substring */
	size_t		last_index;	/* last character to search in substring */
	size_t		lnum;			/* number of substring being searched */
	size_t		count;		/* count of matched lines */
} MatchData;

/* search and option initialization */
OSErr	_search_init(SearchData *sd, GlobalData gdata, short arg);
OSErr _search_input(SearchData *sd, size_t first_char, size_t last_char);

/* outputing results */
OSErr	_search_output(SearchData *sd, MatchData *md);
OSErr	_search_output_end(SearchData *sd, MatchData *md);

/* getting card & stack info */
size_t _current_card_num(XCmdBlockPtr param);
size_t _cards_in_stack(XCmdBlockPtr param);

/* executing periodic actions */
OSErr _search_periodic(SearchData *sd);

#endif